#ifndef GLOBALS_H
#define GLOBALS_H


#include <graphics/newci.h>
#include <graphics/mylib.h>
#include <utils/newasm.h>
#include "logdef.h"
#include "gate.h"

#ifdef GLOBALS_G
# define vextern
#else
# define vextern extern
#endif


/*T*/

#define AnaUsingAlarm   false

/*Floating point hardware flag*/

#define AnaFloat        false
#define AnaDebugflag    false

/*Things nobody would ever change*/


#define AnaLarge        1000000L
    /*Dummy size of dynamically allocatable arrays*/
#define AnaNotyet       (-1000000L)


/*Limits on simulation variables*/

#define AnaParamnum     16

#define AnaN_Vdd        1
#define AnaVddmax       20
#define AnaVddmin       0.1

#define AnaN_Vstep      2
#define AnaVstepmax     AnaVddmax
#define AnaVstepmin     0

#define AnaN_Highest    3
#define AnaHighestmax   50

#define AnaN_Lowest     4
#define AnaLowestmin    (-50)

#define AnaN_Dtmax      5

#define AnaDtmaxmax     1e20
#define AnaDtmaxmin     1e-300

#define AnaN_Timerelax  6

#define AnaTimerelaxmax  1e20
#define AnaTimerelaxmin  1e-300

#define AnaN_Timemin    7

#define AnaTimeminmin   1e-300
#define AnaTimeminmax   1e-10

#define AnaN_Tol        8
#define AnaTolmax       10

#define AnaTolmin       1e-30

#define AnaN_Eqn_Epsilon  9

#define AnaEqn_epsilonmax  1e-8
#define AnaEqn_epsilonmin  1e-300

#define AnaN_Vardone_Epsilon  10

#define AnaVardone_epsilonmax  1e-8
#define AnaVardone_epsilonmin  1e-300

#define AnaN_Timefactor  11
#define AnaTimefactormax  1000
#define AnaTimefactormin  2

#define AnaN_Timeincrease  12
#define AnaTimeincreasemax  10
#define AnaTimeincreasemin  1

#define AnaN_Iterfirstmax  13

#define AnaIterfirstmaxmax  100000L

#define AnaIterfirstmaxmin  2

#define AnaN_Srange     14

#define AnaSrangemax    1e8

#define AnaSrangemin    1

#define AnaN_Small      15

#define AnaSmallmax     0.5
#define AnaSmallmin     (-0.5)

#define AnaN_Nodecap    16

#define AnaNodecapmin   1e-100
#define AnaNodecapmax   1e-10


/*constants for equation solver*/

#define AnaRowbound     62
#define AnaColbound     63

/*constants for history mode attributes*/

#define AnaH_Color      1
#define AnaH_PhosVis    2
#define AnaH_Sigdig     3
#define AnaH_Method     4
#define AnaH_Scale      5
#define AnaH_LinOffset  6
#define AnaH_Vzero      7
#define AnaH_LogOffset  8
#define AnaH_Present    9

/*constants for log plotting*/

#define AnaLn10         2.302585093
#define AnaInvLn10      0.4342944819

#define AnaLinear       0
#define AnaLog          1

/*final stops for quantities*/

#define AnaCapTooSmall  1e-60
#define AnaCapTooBig    1e3
#define AnaResTooSmall  1e-15
#define AnaResTooBig    1e15
#define AnaITooBig      1e15
#define AnaITooSmall    1e-60
#define AnaVTooBig      1e3
#define AnaVTooSmall    1e-60
#define AnaPhaseTooBig  1e3
#define AnaFreqTooBig   1e10
#define AnaTimeTooBig   1e60
#define AnaTimeTooSmall  1e-60

/*Constants for probe mode*/

#define AnaMessFlags    0
#define AnaMessScoreboard  1
#define AnaMessNode     2
#define AnaMessGate1    3
#define AnaMessGate2    4
#define AnaMessGate3    5
#define AnaMessGate4    6
#define AnaMessGate5    7


/**
	This first block represents simulation and extraction information for
	each of the cells in the cell library.  The information is help in an
	array of all cells, indexed by identification number, which is initialized
	by program at runtime.
*/
typedef enum {
	ANALOG_ACT_NEWKIND,    ///< Initialize a new kind
	ANALOG_ACT_EX,         ///< Contribute to matrix
	ANALOG_ACT_PREUPDATE,  ///< Calculations before updating
	ANALOG_ACT_UPDATE,     ///< Update memory elements
	ANALOG_ACT_DISPLAY,    ///< Update display
	ANALOG_ACT_IIN,        ///< Relate current of a pin
	ANALOG_ACT_PASS1,      ///< First pass of guess
	ANALOG_ACT_PASS2,      ///< econd pass of guess
	ANALOG_ACT_ATTRCHANGE, ///< Checks attribute changes
	ANALOG_ACT_RESET,      ///< Resets memory elements
	ANALOG_ACT_SET,        ///< Sets state at reset
	ANALOG_ACT_CLEAR,      ///< Clears state at reset
    ANALOG_ACT_REFRESH,    ///< Refresh display
	ANALOG_ACT_ERASE,      ///< Erase display
	ANALOG_ACT_TOUCH,      ///< Touch a gate
	ANALOG_ACT_DISCONNECT, ///< Twiddle meters
	ANALOG_ACT_NEWGATE,    ///< Make a new instance
	ANALOG_ACT_COPYGATE,   ///< Copy an instance
	ANALOG_ACT_OPENCONFIG, ///< To load in config values
	ANALOG_ACT_READGATE,   ///< to read in a gate description
	ANALOG_ACT_WRITEGATE,  ///< to write out a gate description
	ANALOG_ACT_MESSAGE,    ///< display a message
	ANALOG_ACT_PROBE,      ///< contribute to probe display
	ANALOG_ACT_DISPOSEGATE ///< Dispose an old instance
} Analog_32_actionkinds;


typedef struct Analog_32_action {
  Analog_32_actionkinds action;   ///< Desired Action
  log_krec *kind;                 ///< Kind of interest
  log_grec *inst;                 ///< Gate of interest
  long pin;                       ///< Pin of interest
  long page;                      ///< Page of interest
  double Iin;                     ///< Current of pin of interest
  int ok;                         ///< An error flag
  int accepted;                   ///< Another error flag
  char message[256];              ///< A scoreboard message
  long messagenum;                ///< purpose of message
} Analog_32_action;


/**
 * Each pin of a standard cell will have pin-specific information which will
 * involve initital guessing, syntax checking, and simulation. The plist_cell
 * structure is an array, dynamically configured to be (# of pins) at
 * initialization, which will hold these values.
 */
typedef struct Anapinrec {
  int standalone; /**< @brief Says if a pin is allowed to be unconnected without
						underconstraining the system.

						For example, an inverter would have standalone=true, while a
						source grounded NMOS transistor's current input
						would have a value standalone=false. Used for
						syntax checking */
  int active; ///< Says if the pin counts as a connection for syntax checking of other gates. */
} Anapinrec;

typedef Anapinrec Anaparray[AnaLarge];




/** After a successful simulation, all cells that store energy must be allowed
	to update its state. Thus each cell has an procedure pointer for update,
	which may be set to nil for memoryless procedures. This is how cells like
	scopes will communicate its information back to LOG in the final version
	The parameters are the same as for ex_proc. Update phase takes into account
	the ordering of updates; first-phase updates are non-destructive, i.e current
	meters, while second-phase updates are destructive, like system memory.
*/
typedef enum {
  Anasimple, Anacomplex
} Anaupdate_phase;


/** @brief The complete cell descriptor

	A current meter arrow may be pointed at any pin of any gate. To enable
	the operation of this feature, each gate type must have a procedure that
	can return the current into any node a pin is connected to,
	given the nodemap aand attr fields.
*/
typedef struct AnaCell_rec {
  int simulatable;   ///< Can this gate be simulated?
  void (*proc_cell) (Analog_32_action *act);
      ///< Procedure for cell execution
  long intr_cell;   ///< Number of internal variables in cell
  Anaupdate_phase phase;   ///< Phase of update
  Anapinrec *plist_cell;   ///< Pin description array of length:pins_cell
  void *Ainfo;
} AnaCell_rec;

/******************************************************************************/
/*                    Procedure Type for initialization procedure             */
/******************************************************************************/

/*******************************************************************************/
/*In addition to the list of cell definitions, a list of cell instances
  is also required.                                                            */
/*******************************************************************************/

/*=============================================================================*/
/*  Log keeps a node structure which has an analog node anyptr. In addition,
    gates have an anyptr for analog simulation, a subptr of which will have
    internal node storage. The following ptrs fit in those pages               */
/*=============================================================================*/

/// Internal Nodes
typedef struct AnaInt_rec {
  double ltimestep;   ///< Value of last simulation
  double last;   ///< last value of a simulation
  double now;   ///< present value of a simulation
  long nodenum;   ///< Node number for array indexing
  struct AnaInt_rec *next;   ///< Next internal node
} AnaInt_rec;

/// External Nodes
typedef struct AnaExt_rec {
  long refcnt;   ///< Marker for syntax checking
  int Ok_Alone;   ///< Marker for syntax checking
  double ltimestep;   ///< Value of last simulation
  double last;   ///< last value of a simulation
  double now;   ///< present value of a simulation
  long nodenum;   ///< Node number for array indexing
  long col_now;   ///< color for glow mode
  long col_ltimestep;   ///< color for glow mode
  int changed_now;   ///< changed flag for glowmode
} AnaExt_rec;

/* Each gate also has analog-specific information. */

/// Internal nodelist for each gate
typedef struct AnaGate_intnoderec {
  AnaInt_rec *Int_Node;
  struct AnaGate_intnoderec *Next;
} AnaGate_intnoderec;


typedef struct AnaInstpinrec {
  double Cap;   ///< The capacitance of a pin
  double Tau;   ///< The last tau of the pin
  double Vmem;   ///< The voltage memory of a pin
  double Vreset;   ///< The new voltage for a reset
} AnaInstpinrec;

// Pin specific information is held in a variable length array.
typedef AnaInstpinrec Anapininstarray[AnaLarge];


/// Pointer to list of cells instances.
typedef struct Anainstlist {
  AnaGate_intnoderec *Gate_nodelist;
  void *InstVar;
  AnaInstpinrec *Pininfo;
  int NotReady;
} Anainstlist;

/** @brief To insure swift pen movement regardless of circuit size, I am breaking
			the execution up into units. The following variable deals with this */
typedef enum {
  AnaE_Evaluate, AnaE_Solve, AnaE_Calculate, AnaE_Refresh, AnaE_Adapt,
  AnaE_Memory1, AnaE_Memory2, AnaE_Update
} Anaepochs_type;


/// handles the handshaking for system clock
typedef enum {
  Anasettime, Anawait, Anacontinue
} Anatimetype;


/// The color map
typedef struct AnaAnalog_Colors {
  short C_Vmeter, C_Imeter, C_Switchon, C_Switchoff, C_Limiton, C_Probe,
	C_Glow1, C_Glow2, C_Glow3, C_Glow4, C_Glow5, C_Glow6;
} AnaAnalog_Colors;


/// Levels for glow mode
typedef struct AnaAnalog_Levels {
  long Level_a, Level_b, Level_c, Level_d, Level_e, Level_sum;
} AnaAnalog_Levels;

/// Structure for checking limits of simulation variables
typedef union AnaLimcheck {
  long i;
  double r;
} AnaLimcheck;

typedef enum {
  AnaMin, AnaMax
} AnaBounds;

typedef AnaLimcheck AnaCheckarray[AnaParamnum][2];

/// structures for handling commands - Clist is head of list
typedef struct AnaCommand_rec {
  char Command[17];
  char Arguments[256];
  struct AnaCommand_rec *Next;
} AnaCommand_rec;


/// Types for equation solvers
/** Array position 0 is a 'bit-bucket' for vdd & gnd only */
typedef double AnaMatArray[AnaRowbound + 1][AnaColbound + 1];
typedef int AnaMatFlags[AnaRowbound + 1][AnaColbound + 1];

typedef struct AnaMatRec {
  AnaMatArray Mat;
  AnaMatFlags MatNZ;
} AnaMatRec;

typedef enum {
  AnaS_Start, AnaS_Bestrow, AnaS_Divide, AnaS_Subtract, AnaS_Finish
} AnaSolve_Type;

/// Types for log.cnf interface
typedef struct AnaKlistnode {
  log_krec *NodeKind;
  struct AnaKlistnode *Nextnode;
} AnaKlistnode;




/*******************************************************************************/
/*System data structures*/

vextern log_action_t *AnaLogglobals;   ///< Log's global variables
vextern log_action_t *AnaLogIglobals;   ///< Log's global variables
vextern AnaInt_rec *AnaInt_nodelist;   ///< List of current internal pointers
vextern AnaAnalog_Colors AnaColormap;   ///< The color map
vextern AnaAnalog_Levels AnaLevelmap;   ///< The glow mode level map
vextern AnaCheckarray AnaLimits;   ///< Array for checking system parameters
vextern AnaCommand_rec *AnaClist;   ///< Structure for holding cnf commands
vextern AnaKlistnode *AnaKlist;   ///< Structure for tracking analog gates
vextern log_sigrec *AnaVddsig, *Anagndsig;
    ///< Signal pointers for Vdd and Gnd
vextern log_nrec *AnaVddnode, *Anagndnode;
    ///< Signal pointers for Vdd and Gnd



vextern char AnaCh;   ///< Used for 'press any key' debugging


/*******************************************************************************/
/*Variables that show the state of my simulation and state machine*/

vextern int AnaReady;   ///< Says That a circuit is ready to simulate
vextern int AnaChanged;   ///< Says That a change has been made

/*Holds state of simulation proper*/

vextern int AnaStart;   ///< Indicates whether this is initial timestep
vextern int AnaError;   ///< Simulation flag
vextern int AnaDone;    ///< Simulation flags
vextern int AnaOverflow;   ///< Simulation flags
vextern int AnaToobig;   ///< Simulation flag
vextern int AnaToosmall;   ///< Simulation flag
vextern int AnaSuccess;   ///< Simulation flag
vextern long AnaItercount; ///< Variables for adaptive iteration watchdog
vextern long AnaItermax; ///< Variables for adaptive iteration watchdog
vextern double AnaAccum;   ///< Total accumulated timed
vextern double AnaDt; ///< Present timestep of system
vextern double AnaDtlast;   ///< Last timestep of system
vextern int AnaChangedDt;   ///< A new dt has happened
vextern int AnaChangedP;   ///< A transistor parameter has changed
vextern int AnaInuse;   ///< At least one analog gate exists
vextern long AnaDtcount;   ///< Counting variable to prevent roundoff
vextern long AnaRelaxcount;
    ///< Comparison variable for Dtcount for relaxation
vextern int AnaExact;   ///< Flag for doing accuracy checking
vextern int AnaDifficult;   ///< Flag for a difficult simulation
vextern int AnaToomany;   ///< Flag for too many iterations
vextern int AnaErase;   ///< State of reset memory
vextern int AnaOldmodel;   ///< Model selection when loading in gates
vextern int AnaDebugset;   ///< Flag used in debugging mode
vextern long AnaDebugsize;   ///< Number of significant digits for debug mode
vextern int AnaTimermode;   ///< Mode selection for benchmarking
vextern int AnaTimerstate;   ///< State bit for timer state machine
vextern long AnaTimerval;   ///< Benchmark memory location
vextern double AnaVshift;   ///< Smallest voltage not requiring a gearshift
vextern int AnaDimgate;   ///< Switch for dimming unconnected gates

vextern int AnaProbeOn;   ///< State of probe mode
vextern char AnaLastnodeStr[256];   ///< last node voltage displayed
vextern char AnaLastGate1Str[256];   ///< last gate1 message displayed
vextern char AnaLastGate2Str[256];   ///< last gate2 message displayed
vextern char AnaLastGate3Str[256];   ///< last gate3 message displayed
vextern char AnaLastGate4Str[256];   ///< last gate3 message displayed
vextern char AnaLastGate5Str[256];   ///< last gate3 message displayed

vextern long AnaGateHystercount;   ///< Hysteresis counter for probemode
vextern long AnaNodeHystercount;   ///< Hysteresis counter for probemode

vextern Anaepochs_type AnaEpoch;   ///< Holds present epoch of simulator
vextern long AnaLastpage;   ///< Lastpage evaluated;  0->Clear array
vextern log_grec *AnaLastgate;   ///< Lastgate evaluated
vextern Anatimetype AnaTimestate;   ///< Present time state of system


/*******************************************************************************/
/*internal variable for doing watchdog timing*/

/*$if UsingAlarm$
    T1:Timers_desc;            {Variable for watchdog}
    Quit:Boolean;              {Variable for watchdog}
$end$*/

vextern long AnaStoptime;   ///< Variable for 'slow' watchdog
/*******************************************************************************/


/*******************************************************************************/
/*simulation variables open to the public*/

vextern double AnaVdd;   ///< System Power
vextern double AnaDtmax;   ///< largest allowable timestep
vextern double AnaVstep;   ///< Maximum voltage step allowed in simulation

vextern double AnaHighest;   ///< Highest voltage allowed in simulation
vextern double AnaLowest;   ///< Lowest voltage allowed in simulation
vextern double AnaSmall;   ///< Guess on a capacitor when Vc=0

vextern double AnaTol;   ///< Maximum percentage difference for convergance
vextern double AnaEqn_Epsilon;   ///< epsilon for equation solver
vextern double AnaVardone_Epsilon;   ///< epsilon for eqn zero test

vextern double AnaTimefactor;   ///< How much to scale timestep
vextern double AnaTimemin;   ///< lowest allowable timestep
vextern long AnaIterfirstmax;   ///< Number of iterations for first try
vextern long AnaTimeincrease;   ///< Factor to increase maximum interation by
vextern double AnaSrange;   ///< Range of optimal voltage steps in simulation

vextern double AnaTimerelax;   ///< Minimum time to relax accuracy checking
vextern double AnaNodecap;   ///< Capacitance on a node

vextern long AnaProbeprec;   ///< Precision of probe mode displays
vextern long AnaHysteresis;   ///< Amount of Hysteresis in probe mode

/*******************************************************************************/

/*******************************************************************************/
/*Variables for equation solver*/

vextern long AnaSysRow, AnaSysCol;
vextern AnaMatRec *AnaSystem;
vextern AnaSolve_Type AnaS_Section;
vextern long AnaLastJ;

/*******************************************************************************/



extern void AnaAddcommand (char *Keyword, char *Arg);
extern void AnaAnalog_ChangeReset (int Erase);
extern void AnaCapattrinsert (long base, long attrnum, log_grec *inst,
				 int *Ok);
extern void AnaCapattrload (long base, log_grec *inst, int first);
extern void AnaCapattrread (long base, log_grec *inst);
extern void AnaCapattrwrite (long base, log_grec *inst);
extern void AnaCapCopy (log_grec *inst);
extern void AnaCapDispose (log_grec *inst);
extern void AnaCapex (log_grec *inst);
extern void AnaCapIin (Analog_32_action *act);
extern void AnaCapInit (log_grec *inst);
extern void AnaCappass1 (log_grec *inst);
extern void AnaCappass2 (log_grec *inst);
extern void AnaCapReset (log_grec *inst);
extern void AnaCapUpdate (log_grec *inst);
extern void AnaDiffdump ();
extern void AnaDummyGate (Analog_32_action *act);
extern void AnaDummyInit (log_krec *dummykindptr);
extern void AnaGetcommand (char *Keyword, char *Arg, int *Found);
extern void AnaGetvar (char *Varname, double *Value);
extern void AnaMatdump ();
extern void AnaPopcommand ();
extern void AnaTimeinit ();
extern void AnaTimeReset ();
extern void AnaParamUpdate ();

extern void AnaScoreboard (char *Newmessage, long purpose);
extern int AnaCrange (log_grec *Inst, long Attrnum, int OkZero);
extern int AnaDisplayrange (log_grec *Inst, long Attrnum);
extern double AnaExpo (double V);
extern double AnaSqrt (double V);
extern int AnaGate_On_Display (log_grec *Inst);
extern long AnaGlowVal (long Intensity);
extern int AnaQuit ();
extern char *AnaProbeAmpStr (char *AnaResult, double Val);
extern char *AnaProbeVoltStr (char *AnaResult, double Val);
extern double AnaTanch (double V);


extern int AnaVrange (log_grec *Inst, long Attrnum);

/*=========================================================================*/
/* These structures, and variables, hold physical and fab constants for    */
/* NFET7/PFET7 series of gates, and related circuits.                      */
/*=========================================================================*/

/// Physical constants, independent of fabrication process
typedef struct AnaPhysicalConstants {

  double      T;   ///<  Temperature (degree K)
  double      k;   ///<  Boltzmann's Constant  (Joules/degree K)
  double      q;   ///<  Charge on an electron (Coulombs)
  double    e_v;   ///<  Permittivity of vacumn (F/m)
  double    e_s;   ///<  Permittivity of Si     (F/m)
  double   e_ox;   ///<  Permittivity of SiO2   (F/m)

  double     Ut;   ///<  k*T/q
} AnaPhysicalConstants;

/// Fabrication parameters, independent of device type.
typedef struct AnaFabParameters {
  double phi_ms;   ///<  silicon-oxide interface charge

  double     Eg;     ///<  Bandgap voltage, temperature-dependent
  double     ni;     ///<  ni, temperature-dependent
  double lambda;     ///<  lithography scaling factor
  char fabrun[256];  ///<  name of fabrication run
  char process[256]; ///<  name of fabrication process

}AnaFabParameters;

/// Fabrication parameters, different for N/P channel devices
typedef struct AnaChannelParameters {
  short     nsign;  ///<  1 for nchannel, -1 for pchannel

  double      Tox;  ///<  oxide thickness -- meters
  double      psi;  ///<  potential at depletion edge -- V
  double       Na;  ///<  bulk doping concentration -- 1/cm^3
  double      mu0;  ///<  carrier mobility -- cm^2/(V*s)
  double   deltaW;  ///<  W drawn-->effective correction -- um
  double   deltaL;  ///<  L drawn-->effective correction -- um
  double  del_NaS;  ///<  slope for Na offset dependence
  double  Early_s;  ///<  slope term for Early Effect
  double      L_0;  ///<  intercept term for Early Effect

  double      Qss;  ///<  fixed oxide charge -- C
  double  del_Na0;  ///<  intercept for Na offset dependence
  double   del_mu;  ///<  offset for mobility (unitless)

  double aCactive;  ///<  Capacitance/um^2, active
  double   aCwell;  ///<  Capacitance/um^2, well
  double   linCgs;  ///<  Capacitance/um, gate-source/drain
  double     aCgw;  ///<  Capacitance/um^2, equiv linear gate C

  double      Cox;  ///<  e_ox/Tox F/m^2
  double       mu;  ///<  temperature-dependent mobility
}AnaChannelParameters;


vextern AnaPhysicalConstants AnaPhysical;
vextern AnaPhysicalConstants AnaMaxPhysical;
vextern AnaPhysicalConstants AnaMinPhysical;
vextern AnaFabParameters AnaGenFab;
vextern AnaFabParameters AnaMaxGenFab;
vextern AnaFabParameters AnaMinGenFab;
vextern AnaChannelParameters AnaNFab;
vextern AnaChannelParameters AnaMaxNFab;
vextern AnaChannelParameters AnaMinNFab;
vextern AnaChannelParameters AnaPFab;
vextern AnaChannelParameters AnaMaxPFab;
vextern AnaChannelParameters AnaMinPFab;

#undef vextern

#endif /*GLOBALS_H*/
