#ifndef LOG_ACTION_H
#define LOG_ACTION_H

#include "logdef.h"
#include "gate.h"
#include "wire.h"
#include "page.h"
#include "label.h"
#include <graphics/mylib.h>
#include <limits.h>

typedef enum
{
	act_init,
	act_endinit,
	act_exit,
	act_clearmsg,
	act_status,
	act_cnf,
	act_immed,
	act_func,
	act_color,
	act_select,
	act_clear,
	act_edit,
	act_pass,
	act_tstep,
	act_erasegate,
	act_touchgate,
	act_history,
	act_histval,
	act_histstr,
	act_trigger,
	act_glowcol,
	act_newgate,
	act_disposegate,
	act_copygate,
	act_writegate,
	act_readgate,
	act_connectgate,
	act_disconnectgate,
	act_configgate,
	act_configchgate,
	act_configrelgate,
	act_confignode,
	act_configchnode,
	act_configrelnode,
	act_confighist,
	act_configchhist,
	act_configrelhist,
	act_newkind,
	act_disposekind,
	act_newnode,
	act_disposenode,
	act_copynode,
	act_combinenodes,
	act_combineintonode,
	act_writenode,
	act_readnode,
	act_refnodes,
	act_nodeval,
	act_general,
	act_gennode,
	act_genkind,
	act_gengate
} log_actionkinds;

typedef struct log_hooks_t
{
	void (*hidecursor) ();
	void (*hidecursorrect) (long x1, long y1, long x2, long y2);
	void (*unhidecursor) ();
	void (*nodeconflict) (log_nrec *n);
	void (*gateconflict) (log_grec *g);
	void (*clearconflicts) (struct log_tool *tool);
	void (*setdimgate) (log_grec *g, int dim);
	void (*xform) (log_grec *g, short *x, short *y);
	void (*getsig) (char *name, log_sigrec **sig);
	void (*report) (int num, char *name);
	void (*message) (char *msg);
	void (*vmessage) (char *msg);
	void (*vmessageflag) (char *msg, int b);
	void (*trace) (char *msg);
	void (*dump) (char *msg);
	void (*setdump) (char *name);
	void (*closefiles) ();
	void (*clearalpha) ();
	void (*drawstr) (int x, int y, char *s);
	void (*rightstr) (int x, int y, char *s);
	void (*centerstr) (int x, int y, char *s);
	void (*clip) ();
	void (*remcursor) ();
	void (*clearwindow) ();
	void (*setmode) (const char *s);
	void (*getword) (char *buf, char *wrd);
	void (*getint) (char *buf, long *i, long def);
	void (*getreal) (char *buf, double *r, double def);
	void (*getbool) (char *buf, int *b);
	void (*pen) ();
	void (*pass) ();
	void (*trykbd) ();
	void (*scroll) ();
	void (*trykbdscroll) ();
	void (*inkey) (char *ch);
	void (*udrawgate) (int x, int y, int g, int c);
	void (*gdrawgate) (int x, int y, int g, int c);
	void (*drawgate) (log_grec *g);
	void (*erasegate) (log_grec *g);
	void (*drawnode) (log_nrec *n, int c);
	void (*refresh) ();   /*quietly refresh drawing area*/
	void (*refreshsoon) ();
	void (*beginbottom) ();
	void (*endbottom) ();
	void (*beginerror) ();
	void (*enderror) ();
	void (*redraw) ();   /*redraw drawing area*/
	void (*redrawwindow) ();   /*redraw entire window*/
	void (*change) (log_page *page);
	void (*newgate) (log_grec **g, short gt);
	void (*copygate) (log_grec *old, log_grec **g);
	void (*disposegate) (log_grec **g);
	void (*newnode) (log_nrec **n, uchar st);
	void (*copynode) (log_nrec *old, log_nrec **n);
	void (*disposenode) (log_nrec **n);
	void (*switchnode) (log_nrec **n, log_nrec *n2);
	void (*dumpnodes) ();
	void (*readlnpass) (char *s, short mode);
	void (*clearfunc) ();
	void (*refrfunc) ();
	void (*assertfunc) (char *name);
	void (*findobject) (short x, short y);
	void (*findwire) (short x, short y);
	void (*deletehw) (log_hwrec *hw);
	void (*deletevw) (log_vwrec *vw);
	void (*disconnectgate) (log_grec *g);
	void (*deletegate) (log_grec *g);
	void (*addhw) (short x1, short x2, short y);
	void (*addvw) (short x, short y1, short y2);
	void (*connectgate) (log_grec *g, int *success);
	void (*waitnear) ();
	void (*addgate) (short x, short y, short gtype);
	void (*uaddgate) (short x, short y, short gtype);
	void (*getgate) (char *name, short *gtype);
	void (*realunit) (double r, short p, char *u, char *s);
	void (*realunit2) (double r, short p, char *u, char *s);
	void (*addlabel) (log_lrec **l, char *s);
	void (*getcolor) (char *name, short *val, short def);
	void (*setcmap) (short n, short r, short g, short b);
	void (*parselabel) (strlist_t **lbl, short *numattrs, log_kattrrec **attr);
	void (*editattrs) (log_gattrrec *gattr, short numattrs, log_kattrrec *kattr, strlist_t *lbl,
			char *name, void (*proc) (), void (*chproc) (), void (*relproc) ());
	void (*newattrs) (log_gattrrec **gattr, short numattrs, log_kattrrec *kattr);
	void (*disposeattrs) (log_gattrrec **gattr, short numattrs, log_kattrrec *kattr);
	void (*copyattrs) (log_gattrrec **gattr, log_gattrrec *oldattr, short numattrs, log_kattrrec *kattr);
	void (*initwindow) ();
	void (*clearshowalpha) ();
	void (*setupregion) (log_regrec **r, short pagenum);
	void (*stamp) (long *i);
	void (*noclip) ();
} log_hooks_t;

typedef struct log_hook2_t
{
	void (*getnodeval) (log_nrec *n, double *val, char *opts);
	void (*solderat) (short x, short y);
	void (*unsolderwires) (log_hwrec *hw, log_vwrec *vw);
	void (*settofrom) (log_grec **g, char *name);
	void (*findattrnum) (log_krec *k, char *name, char *kinds, short *num);
	void (*findattrname) (log_krec *k, short num, char *name);
	void (*findpinnum) (log_krec *k, char *name, short *num);
	void (*findpinname) (log_krec *k, short num, char *name);
	void (*setgattr) (log_grec *g, short num, char *value);
	void (*setnattr) (log_nrec *n, short num, char *value);
	void (*findpointmarker) (log_krec *k, short num, short *x, short *y);
	void (*findboxmarker) (log_krec *k, short num, short *x1, short *y1, short *x2, short *y2);
	void (*send_general) (struct log_tool *sim, char *act);
	void (*send_gennode) (log_nrec *n, char *act);
	void (*send_genkind) (log_krec *k, char *act);
	void (*send_gengate) (log_grec *g, char *act);
	void (*send_all) (char *act);
	void (*send_allnodes) (char *act);
	void (*send_allkinds) (char *act);
	void (*send_allgates) (char *act);
	void (*addlabelat) (short x, short y, char *s);
	void (*addboxat) (short x1, short y1, short x2, short y2);
	void (*plainxform) (log_grec *g, short *x, short *y);
	void (*findattrnum2) (short numattrs, log_kattrrec *kattr, strlist_t *lbl, char *name, char *kinds, short *num);
	void (*findattrname2) (short numattrs, log_kattrrec *kattr, strlist_t *lbl, short num, char *name);
	void (*showpinname) (log_grec *g, short i, short c, char *name);
} log_hook2_t;

typedef struct log_action_t
{
	log_actionkinds action;
	log_krec *actkind;
	log_grec *actgate;
	log_grec *actgate2;
	log_nrec *actnode;
	log_nrec *actnode2;
	long actx;
	long acty;
	int actflag;
	int actflag2;
	FILE **actfile;
	struct log_tool *acttool;
	char *actstr;
	double actval;
	double actval2;
	double actval3;
	log_gattrrec *actgattr;
	log_kattrrec *actkattr;
	unsigned char *actproc;

	log_page *pages[log_maxpages]; /** \TODO use a variable size array or a list */
	size_t numpages; ///< number of pages
	log_nrec *nbase;
	long curpage;  ///< current page number
	long showpage; ///< currenctly displayed page number


	log_sigrec *signaltab;
	short maxsignal;
	short lastsignal;

	short scale;
	short hscale;
	long xoff;	   ///< screen_x = grid_x * scale - xoff
	long yoff;

	char func[17];
	char *funcarg;

	log_grec *neargate;
	log_lrec *nearlabel;
	log_brec *nearbox;
	log_hwrec *nearhw;
	log_hwrec *movinghw;
	log_vwrec *nearvw;
	log_vwrec *movingvw;

	log_nrec *probenode;
	log_grec *probegate;
	short probepin;
	struct log_tool *probesimtype;
	short baselinecolor;

	log_hnrec *hnbase;
	log_htrec *htbase;
	log_htrec *htlast;

	long fastspeed;
	long fastmin;
	long fastmax;
	long fastrate;

	long singlestepcount;

	double time;
	double prevtimestep;

	log_simstatekinds simstate;
	log_simstatekinds oldsimstate;

	struct log_tool *simstatetool;
	struct log_tool *oldsimstatetool;

	int histactive;      /** Scope mode is triggered */
	int dumpopen;        /** Dump output file is open */
	int traceopen;       /** Trace output is open */
	int verbose;         /** Commands acknowledge with messages */
	int traceflag;       /** Debuggin/trace mode on */
	int quiet;           /** Suppress sound signals */
	int glowmode;        /** Glowing-wires mode */
	int probemode;       /** Probe-cursor mode */
	int markers;         /** Printing markers are defined */
	int invisible;       /** Gates and wires are invisible */
	int textinvisible;   /** Lablels and boxes are invisible */
	int pnuminvisible;   /** Pin numbers are invisible */
	int dotsvisible;      /** Are "red dots" on or off normally? */
	int showconflicts;   /** Show onlyconflict */
	int resetflag;       /** "Reset" command has been given */
	int pwrflag;         /** Simulator is turned on */
	int initdone;       /** Done with program initialization */
	int refrflag;        /** Refreshing all displays */

	int busyflag;        /** Flag for idling simulations */

	short markerx1;
	short markery1;
	short markerx2;
	short markery2;

	m_tablet_info t;
	m_tablet_info t0;
	int startpoint;
	int stillnear;
	int incircuit;

	short gridx;
	short gridy;
	short posx;
	short posy;
	short oldx;
	short oldy;
	short cx_min;
	short cy_min;
	short cx_max;
	short cy_max;
	/** Area covered bu cyrsor */

	FILE **tracefile;

	char *homedirname;

	struct log_tool *toolbase;

	log_hooks_t hook;
	log_colorrec color;

	long curstamp;
	long labelstamp;                /** Changes to labels */
	long boxstamp;                  /** Changes to boxes */
	long msgstamp;                  /** Clearing messages screen */
	long refrstamp;                 /** Clearing graphics screen */
	long nattrstamp;                /** Changes to note attributes */
	long gattrstamp;                /** Changes to gate attributes */
	long sigstamp;                  /** Changes to signal table */
	long loadstamp;                 /** A page was loaded */
	long colorstamp;                /** Color assignment have changes */
	long resetstamp;   /* Simulation has been reset */
	long markerstamp;               /** Page markers moved */

	log_krec *probekind;
	strlist_t *actstrlist;
	log_hook2_t *hook2;
	log_hwrec *probehwire;
	log_vwrec *probevwire;
	long actx2;
	long acty2;
	char genfunc [17];
} log_action_t;

#endif
