
#include <utils/p2c.h>
#include <graphics/newci.h>
#include "logfont.h"
#include "logglobals.h"
#include "logstructs.h"
#include "window.h"
#include "utils.h"
#include "node.h"
#include "log.h"
#include "gate.h"
#include "wire.h"
#include "label.h"
#include "page.h"
#include "tool.h"
#include "keyboard.h"
#include "box.h"
#include "pen.h"

/// Check if inside a gate's "yellow box."
/** new version by Tim Edwards, Dec 1996 */
int insidegate(log_grec *gate, short x, short y)
{
	short xx1, xx2, yy1, yy2, z;
	short loc_x1, loc_x2;
	log_krec *WITH;

	x = (x - gate->x) * log_scale0;
	y = (y - gate->y) * log_scale0;
	WITH = gate->kind;
	loc_x1 = WITH->x1;
	loc_x2 = WITH->x2;

	/* Modify x1 position for length of text on TO/FROM-type gates */

	if (WITH->flag.U3.named && gate->sig > 0)
	{
		char STR1[256];
		short xwid = m_strwidth(NULL, strrtrim(strcpy(STR1,
						gg.signaltab[gate->sig - 1].name))) - NAMEGAP;

		if (WITH->flag.U3.nright)
		{
			if (xwid > loc_x2) loc_x2 = xwid;
		}
		else
		{
			if (-xwid < loc_x1) loc_x1 = -xwid;
		}
	}

	if (gate->rot == 0)
	{
		return (loc_x1 <= x && x <= loc_x2 && WITH->y1 <= y && y <= WITH->y2);
	}
	else
	{
		z = gate->rot;
		xx1 = loc_x1 * log_rotxx[z] + WITH->y1 * log_rotyx[z];
		xx2 = loc_x2 * log_rotxx[z] + WITH->y2 * log_rotyx[z];
		sortshints(&xx1, &xx2);
		yy1 = loc_x1 * log_rotxy[z] + WITH->y1 * log_rotyy[z];
		yy2 = loc_x2 * log_rotxy[z] + WITH->y2 * log_rotyy[z];
		sortshints(&yy1, &yy2);
		return (xx1 <= x && x <= xx2 && yy1 <= y && y <= yy2);
	}
}

/// Erase a gate (including LED's and other "magic" stuff) using grid coordinates.
void eragate(log_grec *gate)
{
	log_grec *g1, *g2;

	if (gate->conflict || gate->oconflict)
	{
		g1 = gateconflictbase;
		g2 = NULL;
		while (g1 != NULL && g1 != gate)
		{
			g2 = g1;
			g1 = g1->confnext;
		}

		if (g1 == NULL)
			report(10, "ERAGATE");
		else if (g2 == NULL)
			gateconflictbase = gate->confnext;
		else
			g2->confnext = gate->confnext;

		gate->conflict = false;
		gate->conflict1 = false;
		gate->conflict2 = false;
		gate->oconflict = false;
	}

	drawgatec(gate->x, gate->y, gate->g, gg.color.backgr);
	gsignallabel(gate->x, gate->y, gate, gg.color.backgr);

	if (gate->kind->numpnums > 0)
		drawpnums(gate, gg.color.backgr);

	drawgatedotsc(gate->x, gate->y, gate->g, gg.color.backgr);
	calltoolgate(gate, act_erasegate);
}

/// Add a gate to the gate list.
void linkgate(log_grec **gate)
{
	(*gate)->next = gg.pages[gg.curpage - 1]->gbase;
	gg.pages[gg.curpage - 1]->gbase = *gate;
	chpageplace(gg.pages[gg.curpage - 1], (*gate)->x - (*gate)->kind->bbmax,
			(*gate)->y - (*gate)->kind->bbmax, (*gate)->x + (*gate)->kind->bbmax,
			(*gate)->y + (*gate)->kind->bbmax);
}

void newgptr(log_grec **gate)
{
	*gate = (log_grec *)Malloc(sizeof(log_grec));
}


void newgate2(log_grec **gate, short gt, short sig, log_gattrrec *attrs)
{
	char name[256];

	newgptr(gate);
	(*gate)->g = gt;
	(*gate)->rot = gt / log_kindoffset;
	(*gate)->kind = kind[(gt & (log_kindoffset - 1)) - 1];
	(*gate)->vars = (na_long)0;
	(*gate)->conflict = false;
	(*gate)->conflict1 = false;
	(*gate)->conflict2 = false;
	(*gate)->oconflict = false;
	(*gate)->dimcolor = false;
	(*gate)->pin = (log_nrec **)Malloc((*gate)->kind->numpins * sizeof(log_nrec *));
	(*gate)->pinpos = (log_pinposrec *)
		Malloc((*gate)->kind->numpins * sizeof(log_pinposrec));
	if (attrs != NULL)
		copyattrs(&(*gate)->attr, attrs, (*gate)->kind->numattrs, (*gate)->kind->attr);
	else
		newattrs(&(*gate)->attr, (*gate)->kind->numattrs, (*gate)->kind->attr);

	if (sig != 0)
	{
		(*gate)->sig = sig;
	}
	else if ((*gate)->kind->simtype == simtype_common)
	{
		*name = '\0';
		short i = 1;
		while ((*gate)->kind->proc[i - 1] != '\0')
		{
			sprintf(name + strlen(name), "%c", (*gate)->kind->proc[i - 1]);
			i++;
		}
		(*gate)->sig = getsignal(0, name);
	}
	else
	{
		(*gate)->sig = 0;
	}

	(*gate)->info = (na_long)0;
	(*gate)->temp = (na_long)0;
	gg.actflag = (attrs != NULL);
	calltoolgate(*gate, act_newgate);
}

/// Creates a gate.
void newgate(log_grec **gate, short gt)
{
	newgate2(gate, gt, 0, NULL);
}


/// Make a copy of a gate (unlinked).
void copygate(log_grec *old, log_grec **gate)
{
	newgptr(gate);
	**gate = *old;
	if ((*gate)->conflict || (*gate)->oconflict)
	{
		(*gate)->confnext = gateconflictbase;
		gateconflictbase = *gate;
	}
	(*gate)->pin = (log_nrec **)Malloc((*gate)->kind->numpins * sizeof(log_nrec *));

	long numpins = (*gate)->kind->numpins;
	for (long i = 0; i < numpins; i++)
		(*gate)->pin[i] = old->pin[i];

	(*gate)->pinpos = (log_pinposrec *)
		Malloc((*gate)->kind->numpins * sizeof(log_pinposrec));

	numpins = (*gate)->kind->numpins;
	for (long i = 0; i < numpins; i++)
		(*gate)->pinpos[i] = old->pinpos[i];

	copyattrs(&(*gate)->attr, old->attr, (*gate)->kind->numattrs, (*gate)->kind->attr);
	(*gate)->info = (na_long)0;
	gg.actgate2 = old;
	calltoolgate(*gate, act_copygate);
}


/// Remove a gate from gate list.
void unlkgate(log_grec **gate)
{
	log_grec *g1;

	g1 = gg.pages[gg.curpage - 1]->gbase;
	while (g1 != NULL && g1->next != *gate)
		g1 = g1->next;
	if (g1 == NULL)
		gg.pages[gg.curpage - 1]->gbase = (*gate)->next;
	else
		g1->next = (*gate)->next;
	chpageplace(gg.pages[gg.curpage - 1], (*gate)->x - (*gate)->kind->bbmax,
			(*gate)->y - (*gate)->kind->bbmax, (*gate)->x + (*gate)->kind->bbmax,
			(*gate)->y + (*gate)->kind->bbmax);
}


/// Dispose of a gate in memory.
void disposegptr(log_grec **gate)
{
	free(*gate);
	*gate = NULL;
}


void disposegate(log_grec **gate)
{
	log_grec *g1, *g2;

	if ((*gate)->conflict || (*gate)->oconflict)
	{
		g1 = gateconflictbase;
		g2 = NULL;
		while (g1 != NULL && g1 != *gate)
		{
			g2 = g1;
			g1 = g1->confnext;
		}
		if (g1 == NULL)
			report(10, "DISPOSEGATE");
		else if (g2 == NULL)
			gateconflictbase = (*gate)->confnext;
		else
			g2->confnext = (*gate)->confnext;
	}
	if (*gate == gg.probegate)
		gg.probegate = NULL;
	calltoolgate(*gate, act_disposegate);
	disposeattrs(&(*gate)->attr, (*gate)->kind->numattrs, (*gate)->kind->attr);
	Free((*gate)->pin);
	Free((*gate)->pinpos);
	disposegptr(gate);
}

/** @brief Find which gate (if any) is under cursor.
			Also checks for textual labels.
*/
void closergate(short x, short y)
{
	if (gg.textinvisible)
		gg.nearlabel = NULL;
	else {
		gg.nearlabel = gg.pages[gg.curpage - 1]->lbase;
		while (gg.nearlabel != NULL &&
				(x < gg.nearlabel->x ||
				 x > gg.nearlabel->x + m_strwidth(logfont_lfont,
					 gg.nearlabel->name) / gg.scale ||
				 y < gg.nearlabel->y || y > gg.nearlabel->y + 2)) {
			gg.nearlabel = gg.nearlabel->next;
		}
	}
	if (gg.nearlabel != NULL) {
		gg.nearbox = NULL;
		gg.neargate = NULL;
		return;
	}
	if (!gg.textinvisible) {
		gg.nearbox = gg.pages[gg.curpage - 1]->bbase;
		while (gg.nearbox != NULL &&
				((x != gg.nearbox->x1 && x != gg.nearbox->x2 &&
				  y != gg.nearbox->y1 &&
				  y != gg.nearbox->y2) || x < gg.nearbox->x1 ||
				 x > gg.nearbox->x2 || y < gg.nearbox->y1 || y > gg.nearbox->y2))
			gg.nearbox = gg.nearbox->next;
	} else
		gg.nearbox = NULL;
	if (gg.nearbox != NULL)
	{
		gg.neargate = NULL;
		return;
	}
	gg.neargate = gg.pages[gg.curpage - 1]->gbase;
	while (gg.neargate != NULL && !insidegate(gg.neargate, x, y))
		gg.neargate = gg.neargate->next;
}


void chggate(log_grec *gate, int i, log_nrec *oldnode, log_nrec *n)
{
	log_hwrec *hw;
	log_vwrec *vw;
	log_grec *g1;
	short j, k, savepg, pg, x, y, FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Change gate %p pin %d to node %p\n",
				gate, i, n);
	j = i;
	do
	{
		if (gate->pin[j - 1] == oldnode)
		{
			switchnode(&gate->pin[j - 1], n);
			x = gate->pinpos[j - 1].x;
			y = gate->pinpos[j - 1].y;
			hw = gg.pages[gg.curpage - 1]->hwbase;
			while (hw != NULL)
			{
				if (hw->x1 <= x && x <= hw->x2 && hw->y == y && hw->node == oldnode)
					chghw(hw, oldnode, n);
				hw = hw->next;
			}
			vw = gg.pages[gg.curpage - 1]->vwbase;
			while (vw != NULL)
			{
				if (vw->y1 <= y && y <= vw->y2 && vw->x == x && vw->node == oldnode)
					chgvw(vw, oldnode, n);
				vw = vw->next;
			}
			g1 = gg.pages[gg.curpage - 1]->gbase;
			while (g1 != NULL)
			{
				if (g1 != gate && P_imax2((long)abs(g1->x - x), (long)abs(g1->y - y)) <=
						g1->kind->bbmax)
				{
					FORLIM = g1->kind->numpins;
					for (k = 1; k <= FORLIM; k++)
					{
						if (x == g1->pinpos[k - 1].x && y == g1->pinpos[k - 1].y &&
								g1->pin[k - 1] == oldnode)
							chggate(g1, k, oldnode, n);
					}
				}
				g1 = g1->next;
			}
		}
		j = gate->kind->pin[j - 1].c;
	} while (j != i && j != 0);
	if (i != 1 || gate->kind->simtype != simtype_common || gate->sig == 0 ||
			gg.signaltab[gate->sig - 1].np != oldnode)
		return;
	switchnode(&gg.signaltab[gate->sig - 1].np, n);
	savepg = gg.curpage;
	FORLIM = gg.numpages;
	for (pg = 1; pg <= FORLIM; pg++)
	{
		gg.curpage = pg;
		g1 = gg.pages[gg.curpage - 1]->gbase;
		while (g1 != NULL)
		{
			if (g1->kind->simtype == simtype_common && g1->sig == gate->sig &&
					g1->pin[0] == oldnode)
				chggate(g1, 1, oldnode, n);
			g1 = g1->next;
		}
	}
	gg.curpage = savepg;
}


void chggatepin(log_grec *gate, short i, log_nrec **oldnode, log_nrec **savenode)
{
	log_nrec *n;
	log_hwrec *hw;
	log_vwrec *vw;
	log_grec *g1;
	short j, x, y, FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Change gate %p pin %d from node %p\n",
				gate, i, *oldnode);
	x = gate->pinpos[i - 1].x;
	y = gate->pinpos[i - 1].y;
	hw = gg.pages[gg.curpage - 1]->hwbase;
	while (hw != NULL)
	{
		if (hw->x1 <= x && x <= hw->x2 && hw->y == y && hw->node == *oldnode)
		{
			delgetnode(savenode, oldnode, &n);
			if (n != *oldnode)
				chghw(hw, *oldnode, n);
		}
		hw = hw->next;
	}
	vw = gg.pages[gg.curpage - 1]->vwbase;
	while (vw != NULL)
	{
		if (vw->y1 <= y && y <= vw->y2 && vw->x == x && vw->node == *oldnode)
		{
			delgetnode(savenode, oldnode, &n);
			if (n != *oldnode)
				chgvw(vw, *oldnode, n);
		}
		vw = vw->next;
	}
	g1 = gg.pages[gg.curpage - 1]->gbase;
	while (g1 != NULL)
	{
		if (g1 != gate &&
				P_imax2((long)abs(g1->x - x), (long)abs(g1->y - y)) <= g1->kind->bbmax)
		{
			FORLIM = g1->kind->numpins;
			for (j = 1; j <= FORLIM; j++)
			{
				if (x == g1->pinpos[j - 1].x && y == g1->pinpos[j - 1].y &&
						g1->pin[j - 1] == *oldnode) {
					delgetnode(savenode, oldnode, &n);
					if (n != *oldnode)
						chggate(g1, j, *oldnode, n);
				}
			}
		}
		g1 = g1->next;
	}
}


void disconnectgate(log_grec *gate)
{
	log_nrec *savenode;
	short i, j;
	log_nrec **oldnode;
	long done[log_maxpins / 32 + 2];
	short FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Disconnect gate %p\n", gate);
	stamp(&gate->kind->simtype->netstamp);
	calltoolgate(gate, act_disconnectgate);
	oldnode = (log_nrec **)Malloc(gate->kind->numpins * sizeof(log_nrec *));
	FORLIM = gate->kind->numpins;
	for (i = 0; i < FORLIM; i++)
	{
		oldnode[i] = gate->pin[i];
		stamp(&oldnode[i]->simtype->netstamp);
		switchnode(&gate->pin[i], NULL);
	}
	P_expset(done, 0L);
	FORLIM = gate->kind->numpins;
	for (i = 1; i <= FORLIM; i++)
	{
		if (!P_inset(i, done))
		{
			j = i;
			if (i == 1 && gate->kind->simtype == simtype_common && gate->sig != 0)
				savenode = NULL;
			else
				savenode = oldnode[i - 1];
			do {
				chggatepin(gate, j, &oldnode[i - 1], &savenode);
				P_addset(done, j);
				j = gate->kind->pin[j - 1].c;
			} while (j != i && j != 0);
		}
	}
	unlkgate(&gate);
}


void delgate(log_grec *gate)
{
	clipon();
	eragate(gate);
	disconnectgate(gate);
	disposegate(&gate);
	clipoff();
}

void frygate(log_grec *gate)
{
	long t0;

	remcursor();
	clipon();
	t0 = timers_sysclock();
	while (labs(timers_sysclock() - t0) < frytime)
	{
		drawgatec(gate->x, gate->y, gate->g, gg.color.conflict);
		eragate(gate);
	}
	clipoff();
	refreshsoon();
}

void initpinpos(log_grec *gate)
{
	short i, rxx, rxy, ryx, ryy;
	log_krec *WITH;
	short FORLIM;

	rxx = log_rotxx[gate->rot];
	rxy = log_rotxy[gate->rot];
	ryx = log_rotyx[gate->rot];
	ryy = log_rotyy[gate->rot];
	WITH = gate->kind;
	FORLIM = WITH->numpins;
	for (i = 0; i < FORLIM; i++)
	{
		gate->pinpos[i].x = gate->x + WITH->pin[i].x * rxx + WITH->pin[i].y * ryx;
		gate->pinpos[i].y = gate->y + WITH->pin[i].x * rxy + WITH->pin[i].y * ryy;
	}
}

/** @brief Check each pin of a gate and
			connect to wires, other pins as necessary.
*/
int connectgate(log_grec *gate)
{
	int Result;
	cnrec *cnbase;
	short i, j, xp, yp;
	log_hwrec *hw;
	log_vwrec *vw;
	log_nrec **n1;
	log_grec *g1;
	short ptrs[log_maxpins];
	int success;
	short FORLIM, FORLIM1;

	if (gg.traceflag)
		fprintf(tracefile, "Connect gate %p, type %s\n", gate, gate->kind->name);
	linkgate(&gate);
	initpinpos(gate);
	FORLIM = gate->kind->numpins;
	for (i = 0; i < FORLIM; i++)
	{
		ptrs[i] = 0;
		gate->pin[i] = NULL;
	}
	FORLIM = gate->kind->numpins;
	for (i = 1; i <= FORLIM; i++)
	{
		if (ptrs[i - 1] == 0)
		{
			newnoderef(&gate->pin[i - 1], gate->kind->pin[i - 1].s, 1L);
			if (gate->pin[i - 1] == NULL)
			{
				Result = false;
				goto _L1;   /*return*/
			}
			j = i;
			do
			{
				ptrs[j - 1] = i;
				j = gate->kind->pin[j - 1].c;
			} while (j != i && j != 0);
		}
	}
	cnbase = NULL;
	FORLIM = gate->kind->numpins;
	for (i = 0; i < FORLIM; i++)
	{
		xp = gate->pinpos[i].x;
		yp = gate->pinpos[i].y;
		n1 = NULL;
		hw = gg.pages[gg.curpage - 1]->hwbase;
		while (hw != NULL && n1 == NULL)
		{
			if (hw->x1 <= xp && xp <= hw->x2 && yp == hw->y)
				n1 = &hw->node;
			hw = hw->next;
		}
		vw = gg.pages[gg.curpage - 1]->vwbase;
		while (vw != NULL && n1 == NULL)
		{
			if (vw->y1 <= yp && yp <= vw->y2 && xp == vw->x)
				n1 = &vw->node;
			vw = vw->next;
		}
		g1 = gg.pages[gg.curpage - 1]->gbase;
		while (g1 != NULL && n1 == NULL)
		{
			if (g1 != gate && P_imax2((long)abs(g1->x - xp), (long)abs(g1->y - yp)) <=
					g1->kind->bbmax)
			{
				FORLIM1 = g1->kind->numpins;
				for (j = 0; j < FORLIM1; j++)
				{
					if (g1->pinpos[j].x == xp && g1->pinpos[j].y == yp)
						n1 = &g1->pin[j];
				}
			}
			g1 = g1->next;
		}
		if (n1 != NULL)
			queuecombine(&cnbase, &gate->pin[ptrs[i] - 1], n1);
	}
	if (gate->kind->simtype == simtype_common && gate->sig != 0)
		queuecombine(&cnbase, &gg.signaltab[gate->sig - 1].np, gate->pin);
	gg.actflag = true;
	calltoolgate(gate, act_connectgate);
	success = gg.actflag;
	if (success && checkcombine(&cnbase))
	{
		FORLIM = gate->kind->numpins;
		for (i = 0; i < FORLIM; i++)
		{
			if (ptrs[i] != i + 1)
				gate->pin[i] = refnode(gate->pin[ptrs[i] - 1]);
		}
		Result = true;
	}
	else
	{
		success = false;
	}

	if (!success)
	{
		Result = false;
		FORLIM = gate->kind->numpins;
		for (i = 1; i <= FORLIM; i++) {
			if (ptrs[i - 1] == i)
				unrefnode(&gate->pin[i - 1]);
		}
		unlkgate(&gate);
	}
_L1:
	return Result;
}


void pconnectgate(log_grec *gate, int *success)
{
	*success = connectgate(gate);
}

void addgate2(short x, short y, short gtype, short sig, log_gattrrec *attrs)
{
	log_grec *g, *g1;
	short i, gmax;
	int flag;

	newgate2(&g, gtype, sig, attrs);
	if (gatecount[gg.curpage - 1] == log_maxshint)
	{
		i = 0;
		gmax = 0;
		do
		{
			i++;
			g1 = gg.pages[gg.curpage - 1]->gbase;
			flag = false;
			while (g1 != NULL && !flag)
			{
				if (g1->gc > gmax)
					gmax = g1->gc;
				flag = (g1->gc == i);
			}
		} while (flag);
		g->gc = i;
		gatecount[gg.curpage - 1] = gmax;
	}
	else
	{
		gatecount[gg.curpage - 1]++;
		g->gc = gatecount[gg.curpage - 1];
	}
	g->x = x;
	g->y = y;
	clipon();
	drawgatex(g);
	clipoff();
	if (!connectgate(g))
	{
		frygate(g);
		disposegate(&g);
	}
	gg.refrflag = true;
	gg.neargate = g;
}

/// Add a gate to the circuit. Initialize all gate variables, etc.
void addgate1(short x, short y, short gtype)
{
	addgate2(x, y, gtype, 0, NULL);
}

void uaddgate(short x, short y, short gtype)
{
	short xx, yy;
	log_grec *g;
	log_krec *k;

	x = P_imin2(P_imax2((long)x, gridcen * 3L / 2), across - gridcen * 3L / 2);
	y = P_imin2(P_imax2((long)y, gridcen * 3L / 2), baseline - gridcen * 3L / 2);
	k = kind[(gtype & (log_kindoffset - 1)) - 1];
	do
	{
		xx = (x + gg.xoff) / gg.scale;
		yy = (y + gg.yoff) / gg.scale;
		g = gg.pages[gg.curpage - 1]->gbase;
		while (g != NULL && P_imax2((long)abs(g->x - xx), (long)abs(g->y - yy)) >
				g->kind->bbmax + k->bbmax)
			g = g->next;
		if (g != NULL)
		{
			x += gridcen * 2;
			if (x > across - gridcen * 3)
			{
				x = gridcen * 2;
				y += gridcen * 2;
			}
		}
	} while (g != NULL);
	addgate1(xx, yy, gtype);
}

void movexorgate(short x, short y, short g, short sig, short yy)
{
	clipoff();
	m_colormode((long)m_xor);
	if (yy < baseline)
	{
		drawgatec(x, y, g, gg.color.gate);
		signallabel((int)(x * gg.scale - gg.xoff), (int)(y * gg.scale - gg.yoff),
				g, sig, gg.color.signal);
	}
	else
	{
		udrawgatec((int)(x * gg.scale - gg.xoff), (int)(y * gg.scale - gg.yoff),
				g, gg.color.gate);
		if (zoom == 0 &&
				kind[(g & (log_kindoffset - 1)) - 1]->bbmax <= maxsafescale)
			drawgatedotsc(x, y, g, gg.color.gatepin);
	}
	m_colormode((long)m_normal);
}

/// Move a gate into position, then add to circuit or menu area.
int addgate(short gtype, short sig, log_gattrrec *attrs)
{
	int Result;
	short yy;
	int oo;
	log_krec *k;

	Result = false;
	if (gtype != 0 && (gtype & (log_kindoffset - 1)) <= maxkinds)
	{
		k = kind[(gtype & (log_kindoffset - 1)) - 1];
		if (k != NULL)
		{
			do
			{
				yy = gg.t.y;
				gg.posx = gg.gridx;
				gg.posy = gg.gridy;
				oo = gg.t.off;
				movexorgate(gg.posx, gg.posy, gtype, sig, yy);
				do
				{
					pass();
					trykbd();
					pen();
				} while (gg.gridx == gg.posx && gg.gridy == gg.posy &&
						gg.t.depressed && strcmp(gg.func, "REFR"));
				movexorgate(gg.posx, gg.posy, gtype, sig, yy);
				scroll();
			} while (gg.t.depressed);
			if (!oo && !gg.t.off)
			{
				if (gg.t.y < baseline)
				{
					addgate2(gg.posx, gg.posy, gtype, sig, attrs);
					Result = true;
				}
				else
				{
					changekind(gtype, sig, attrs);
				}
			}
		}
	}
	gg.startpoint = false;
	return Result;
}

void reportgateconflict(log_grec *g)
{
	if (g->conflict || !conflictenbl)
		return;
	if (!g->oconflict)
	{
		g->confnext = gateconflictbase;
		gateconflictbase = g;
		g->conflict1 = (3L - conflictdelay) & 1;
		g->conflict2 = ((3L - conflictdelay) / 2) & 1;
	}
	g->conflict = true;
}

