/* "LOG", the circuit editing and simulation system,
   "DigLOG", the digital simulator for LOG.
   Copyright (C) 1985, 1990 David Gillespie.
   Author's address: daveg@csvax.caltech.edu; 256-80 Caltech/Pasadena CA 91125.

   "AnaLOG", the analog simulator for LOG.
   Copyright (C) 1985, 1990 John Lazzaro.
   Author's address: lazzaro@csvax.caltech.edu; 256-80 Caltech.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation (any version).

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   */

/*
   Logic Simulator   Version 4.1
   For HP 9836C


   Copyright (c)  1985

   David Gillespie
   */
#define TOOL_KIND

#include "log.h"

#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>


#include <utils/p2c.h>

#include "logdef.h"
#include "log_action.h"
#include <utils/sysglobals.h>
#include <utils/sysdevs.h>

#include <utils/misc.h>
#include "help.h"
#include <graphics/newci.h>
#include <utils/filepack.h>
#include <graphics/mylib.h>
#include "logfont.h"
#include "logstuff.h"
#include "swap.h"
#include <graphics/newkbd.h>
#include <utils/strings.h>
#include <utils/newasm.h>
#include <graphics/newcrt.h>

#include "settings.h"
#include "logstructs.h"
#include "logglobals.h"

#include "attrs.h"
#include "window.h"
#include "display.h"
#include "label.h"
#include "gate.h"
#include "wire.h"
#include "node.h"
#include "page.h"
#include "utils.h"
#include "tool.h"
#include "keyboard.h"
#include "box.h"
#include "pen.h"

#include "pagewriter.h"
#include "pagereader.h"

#include "configreader.h"
#include "libraryreader.h"

enum cursors;
enum cursors cursortype;
enum cursors oldcursortype;

log_action_t gg;   ///< External global variables

short txacross;
short txdown;
short across;
short down;
short baseline;
short line1;
short line2;
short kindgroupbase;
short kindgroupleft;
short kindgroupright;
short kindgroupstart;
short kindgroupsize;
short histdown;

char (*index_)[9];   ///< Index into gate library
short *indexoffset;   ///< Offset of each gate
uchar *indexfile;   ///< File number of each gate
uchar *indexgroup;   ///< Group number of each gate
uchar *loadedgates;   ///< Tells whether a gate is in the catalog
log_krec *kind[maxmaxkinds];   ///< Kind of gate in each catalog space
kindgrouptype kindgroup;   ///< Kind of gate in each menu-area space
kindgrouptype kindsig;
kindattrtype kindattr;
strlist_t *gatefilenames;
short maxkinds;
short catwidth;
catboxrec *catboxes;
librstrrec *librstrs;

char *gatesname[maxgatesfiles];
char *loghelpname;   ///< Name of help file
char *lognewsname;   ///< Name of news file

int ospointflag;   ///< Starting point was touched

long xoff0;   ///< Good ol' XOFF
long yoff0;   ///< Good ol' YOFF

short curwcolor;   ///< Current VLSI wire color
char cureditmode;   ///< Current editing (gate-tapping) mode
short curlistgroup;   ///< Current group in library list

int vlsimode[log_maxpages];   ///< Page contains VLSI circuits
long xoffp[log_maxpages];  ///< XOFF for non-current pages
long yoffp[log_maxpages];  ///< YOFF for non-current pages
short gatecount[log_maxpages];   ///< Highest used gate-ID number
char *curfilename[log_maxpages];   ///< Name of last-loaded file
long realcurpage;

baseptrs copybuf;   ///< Del/Copy/Paste buffer

long htcount;   ///< Number of timesteps in list
short hncount;   ///< Number of names in list
short hnocount;   ///< Old number of names in list
short histtrig;   ///< Number of signal which triggers history
strlist_t *histlbl;
short histnumattrs;
log_kattrrec *histkattr;
log_gattrrec *histgattr;
short histgridmode;
short histgridwhich;
log_hnrec *histgridhn;
double histvalrange;
double histdivsacross;
int histonscreen;   ///< Leading edge of trace is visible
int histreset;   ///< Clear the history memory
double histtime;   ///< Current time for history mode

int probeflag;   ///< Has the probe been checked?

log_nrec *nodeconflictbase;   ///< Node conflict list
log_grec *gateconflictbase;   ///< Gate conflict list

cnfrec *cnfbase;   ///< List of tool configuration lines
strlist_t *colorbase;   ///< List of configured colors
macrorec *macrobase;   ///< List of macro keys
strlist_t *thingstodo; ///< List of delayed commands
strlist_t *nexttodo;   ///< List of delayed commands
strlist_t *messages;   ///< List of delayed messages
strlist_t *commandlist;   ///< List of tool-activation commands

char modename[9];   ///< Word in 'mode' area of menu
int modeflag;   ///< Mode area contains a word
long modetime;   ///< Time (in minutes) in mode display
log_krec *modeprobekind;

/*  menupicture, errorpicture : m_picturevar;              { Various saved images */
short bottomcount;   ///< How many nested BEGINBOTTOMs

uchar rcolormap[16];
uchar gcolormap[16];
uchar bcolormap[16];

short gategreen;
short gateyellow;
short gatered;
short gateorange;
short gateblack;
short gatewhite;

short defineboxcolor;
short definebackcolor;
short definetextcolor;
short catboxcolor;

short messagepos;   ///< Position of next message
short messageright;   ///< Maximum width of messages

long fastsavetime;

long watchdog;  ///< Miscellaneous timers

log_tool *simtype_ignore;   ///< Non-simulated type
log_tool *simtype_common;   ///< Common-pin type

double status_oldtime;
double status_oldtstep;
long status_oldmem;

long helpptr;   ///< Help descriptor

int popup_grid;   ///< Pop-up menus
logmenurec *popupmenus[4];

filerecfilerec *libf1[maxgatesfiles];   ///< Library files
short libfstart[maxgatesfiles]; ///< First descriptor record in LIBF1
short libptr;   ///< Most recently replaced catalog item
short idxsize;   ///< Number of gates in library
char *librgroupnames[9];

short refrtimer;    ///< Time until screen is refreshed
short refrtimer0;   ///< Time until screen is refreshed

char *dumpfname;
char *tracefname;
FILE *dumpfile;
FILE *tracefile;

char pushedbackkey;
char realkey;

long EXCP_LINE;   ///< Line number of last error
long excpline;

short conflictdelay;

int eightcolors;   ///< Has only 8 colors
int onescreen;   ///< Has only one screen
int doingcnffunction;   ///< This is an ACT_CNF, not an ACT_FUNC
int justonecommand;   ///< LOG -R mode, so exit after first cmd
int displaynews;   ///< LOGNEWS has not yet been displayed
int immedscroll;   ///< Don't need to defer 'REFR' function
int tempverbose;   ///< Verbose mode temporarily on
int commandfound;   ///< Function name was acknowledged
int cursorhide;   ///< Cursor is hidden
int reportnowait;   ///< Don't wait for key on bug reports
int firsttraining;   ///< Training used for first time
int training;   ///< Training mode for Help command
int snapflag;   ///< Snap-to-grid cursor
int showsolder;   ///< Display solder points
int briefprobe;   ///< Temporary probe mode
int conflictenbl;   ///< Conflict checking enabled
int conflictstop;   ///< Conflict stopping mode
int anyconflicts;   ///< Any active, reported conflicts?
int steppingoff;   ///< Single-step is turning simulation off
int suppressdots;   ///< Temporarily suppressing red dots
int glowsolder;   ///< Display solder dots in Glow mode
int vlsi;   ///< Current page is VLSI mode


static char *ioresult_message(char *Result, long iores)
{
	char s[81];

	misc_getioerrmsg(s, iores);   /* MISC */
	return strcpy(Result, s);
}

/// Report an internal error.
void report(int num, char *s)
{
	char ch;
	char STR2[256];

	if (reportnowait)
	{
		sprintf(STR2, "Internal error %d in %s", num, s);
		message(STR2);
		return;
	}
	clearshowalpha();
	nc_gotoXY(0, txdown - 1);
	nc_printf("\215--> Internal error %d in %s\210\n", num, s);
	if (gg.traceflag)
		fprintf(tracefile, "--> Internal error %d in %s\n", num, s);
	m_alpha_on();
	nc_printf("    Press any key to continue.");
	ch = m_inkey();
	if (ch == 'N' || ch == 'n')
		reportnowait = true;
	nc_putchar('\n');
}


/// Draw a word in the "mode" location of the menu area.
void log_setmode(const char *s)
{
	time_t h;
	char *cp;

	remcursor();
	clipoff();
	m_color((long)gg.color.backgr);
	if (gg.showpage > 0)
		drawstr2(across + menux3, line2, modename);
	refreditmode();   /*may have been trashed by previous mode name*/
	if (*s != '\0')
	{
		strcpy(modename, s);
		modeflag = true;
		m_color((long)gg.color.selword);
	}
	else
	{
		if (gg.probekind != NULL)
		{
			strcpy(modename, gg.probekind->name);
			m_color((long)gg.color.selword);
		}
		else
		{
			time(&h);
			cp = ctime(&h);
			sprintf(modename, "%.5s", cp + 11);
			modetime = timers_sysclock() / 6000;
			if (!strcmp(modename, "00:00"))
				strcpy(modename, "__@/ ");
			m_color((long)gg.color.menuword);
		}
		modeflag = false;
		modeprobekind = gg.probekind;
	}
	if (gg.showpage > 0)
		drawstr2(across + menux3, line2, modename);
}

void unprobe()
{
	gg.probenode = NULL;
	gg.probegate = NULL;
	gg.probepin = 0;
	gg.probesimtype = NULL;
	gg.probekind = NULL;
	gg.probehwire = NULL;
	gg.probevwire = NULL;
	probeflag = false;
}


static log_krec *peninkind(short xx, short yy)
{
	log_krec *Result;
	short i;

	Result = NULL;
	if (!(P_ibetween((long)kindgroupleft, (long)xx, kindgroupright - 1L) &&
				P_ibetween((long)baseline, (long)yy, (long)down)))
		return Result;
	i = kindgroup[(xx - kindgroupleft) / kindgroupspacing];
	if (i > 0)
		return (kind[(i & (log_kindoffset - 1)) - 1]);
	return Result;
}

/** @brief Find which wire, pin, or gate the Probe is touching
  and set PROBENODE or PROBEGATE to its address.
  */
void testprobe(short xx, short yy)
{
	log_grec *g;
	log_krec *k;
	short i, x, y;
	log_hwrec *hw;
	log_vwrec *vw;
	short FORLIM;

	x = (xx + gg.hscale + gg.xoff) / gg.scale;
	y = (yy + gg.hscale + gg.yoff) / gg.scale;
	unprobe();
	if (gg.showpage > 0 && !gg.invisible && !gg.showconflicts)
	{
		if (gg.incircuit)
		{
			g = gg.pages[gg.curpage - 1]->gbase;
			while (g != NULL && gg.probegate == NULL)
			{
				k = g->kind;
				if (g->x >= x - k->bbmax && g->x <= x + k->bbmax &&
						g->y >= y - k->bbmax && g->y <= y + k->bbmax)
				{
					if (insidegate(g, x, y))
					{
						gg.probegate = g;
					}
					else if (g->kind->simtype->simtype >= 8)
					{
						FORLIM = g->kind->numpins;
						for (i = 1; i <= FORLIM; i++) {
							if (x == g->pinpos[i - 1].x && y == g->pinpos[i - 1].y)
							{
								gg.probegate = g;
								gg.probenode = g->pin[i - 1];
								gg.probepin = i;
							}
						}
					}
				}
				g = g->next;
			}
			hw = gg.pages[gg.curpage - 1]->hwbase;
			while (hw != NULL && gg.probenode == NULL)
			{
				if (hw->x1 <= x && x <= hw->x2 && hw->y == y)
				{
					gg.probehwire = hw;
					gg.probenode = hw->node;
				}
				hw = hw->next;
			}
			vw = gg.pages[gg.curpage - 1]->vwbase;
			while (vw != NULL && gg.probenode == NULL)
			{
				if (vw->y1 <= y && y <= vw->y2 && vw->x == x)
				{
					gg.probevwire = vw;
					gg.probenode = vw->node;
				}
				vw = vw->next;
			}
		}
		else
		{
			gg.probekind = peninkind(xx, yy);
		}
	}
	if (gg.probegate != NULL)
		gg.probekind = gg.probegate->kind;
	if (gg.probenode != NULL)
		gg.probesimtype = gg.probenode->simtype;
	else if (gg.probekind != NULL)
		gg.probesimtype = gg.probekind->simtype;
}


void checkprobe(short xx, short yy)
{
	short x, y;

	x = (xx + gg.hscale + gg.xoff) / gg.scale;
	y = (yy + gg.hscale + gg.yoff) / gg.scale;
	if (gg.probegate != NULL &&
			(gg.probegate->x < x - gg.probekind->bbmax ||
			 gg.probegate->x > x + gg.probekind->bbmax ||
			 gg.probegate->y < y - gg.probekind->bbmax ||
			 gg.probegate->y > y + gg.probekind->bbmax))
	{
		unprobe();
		return;
	}

	if (gg.probekind != NULL && gg.probegate == NULL &&
			peninkind(xx, yy) != gg.probekind)
	{
		unprobe();
		return;
	}

	if (gg.probehwire != NULL &&
			(gg.probehwire->x1 > x || x > gg.probehwire->x2 ||
			 gg.probehwire->y != y))
		unprobe();
	else if (gg.probevwire != NULL &&
			(gg.probevwire->y1 > y || y > gg.probevwire->y2 ||
			 gg.probevwire->x != x))
		unprobe();
}

static void send_general(log_tool *sim, char *act)
{
	char savefunc[17];

	strcpy(savefunc, gg.genfunc);
	strcpy(gg.genfunc, act);
	calltool(sim, act_general);
	strcpy(gg.genfunc, savefunc);
}


static void send_gennode(log_nrec *n, char *act)
{
	char savefunc[17];

	strcpy(savefunc, gg.genfunc);
	strcpy(gg.genfunc, act);
	calltoolnode(n, act_gennode);
	strcpy(gg.genfunc, savefunc);
}


static void send_genkind(log_krec *k, char *act)
{
	char savefunc[17];

	strcpy(savefunc, gg.genfunc);
	strcpy(gg.genfunc, act);
	calltoolkind(k, act_genkind);
	strcpy(gg.genfunc, savefunc);
}


static void send_gengate(log_grec *g, char *act)
{
	char savefunc[17];

	strcpy(savefunc, gg.genfunc);
	strcpy(gg.genfunc, act);
	calltoolgate(g, act_gengate);
	strcpy(gg.genfunc, savefunc);
}


static void send_all(char *act)
{
	log_tool *tp;
	char savefunc[17];

	strcpy(savefunc, gg.genfunc);
	tp = gg.toolbase;
	while (tp != NULL)
	{
		if (tp->ready && tp->simulator)
		{
			strcpy(gg.genfunc, act);
			calltool(tp, act_general);
		}
		tp = tp->next;
	}
	strcpy(gg.genfunc, savefunc);
}


static void send_allnodes(char *act)
{
	log_tool *tp;
	char savefunc[17];

	strcpy(savefunc, gg.genfunc);
	tp = gg.toolbase;
	while (tp != NULL)
	{
		if (tp->ready && tp->simulator)
		{
			strcpy(gg.genfunc, act);
			gg.actnode = NULL;
			calltool(tp, act_gennode);
		}
		tp = tp->next;
	}
	strcpy(gg.genfunc, savefunc);
}


static void send_allkinds(char *act)
{
	log_tool *tp;
	char savefunc[17];

	strcpy(savefunc, gg.genfunc);
	tp = gg.toolbase;
	while (tp != NULL)
	{
		if (tp->ready && tp->simulator)
		{
			strcpy(gg.genfunc, act);
			gg.actkind = NULL;
			calltool(tp, act_genkind);
		}
		tp = tp->next;
	}
	strcpy(gg.genfunc, savefunc);
}


static void send_allgates(char *act)
{
	log_tool *tp;
	char savefunc[17];

	strcpy(savefunc, gg.genfunc);
	tp = gg.toolbase;
	while (tp != NULL)
	{
		if (tp->ready && tp->simulator)
		{
			strcpy(gg.genfunc, act);
			gg.actgate = NULL;
			calltool(tp, act_gengate);
		}
		tp = tp->next;
	}
	strcpy(gg.genfunc, savefunc);
}


long glowcol(log_nrec *n, long def)
{
	gg.actx = def;
	calltoolnode(n, act_glowcol);
	return (gg.actx);
}


void resetmessages()
{
	short opos;

	opos = messagepos;
	messagepos = topmessagepos;
	messageright = 0;
	if (messagepos != opos)
	{
		calltools(act_clearmsg);
		stamp(&gg.msgstamp);
	}
}

/** @brief Debug function to show current number of events received
  and current simulation time.
  */
void show_events()
{
	/*zEMBED
	  char buf[30];
	  int m_events_received;
	  static int old_events, old_serial, old_time;
	  int serial, time = timers_sysclock() / 200;
	  sprintf(buf, "%d", m_events_received);
	  m_color(log_red);
	  if (m_events_received > old_events)
	  drawstr2(5, 20, buf);
	  old_events = m_events_received;
	  sprintf(buf, "%d", serial = XNextRequest(m_display));
	  if (serial > old_serial && time != old_time)
	  drawstr2(5, 29, buf),
	  old_serial = serial + 2,
	  old_time = time;
	  */
}

#define spacing         10

void message(char *msg)
{
	strlist_t *l1;

	if (!gg.initdone)
	{
		l1 = strlist_append(&messages, msg);
		l1->value = (na_long)0;
		return;
	}

	if (messagepos + spacing >= baseline)
	{
		uerase(0, 0, messageright, messagepos);
		resetmessages();
		refreshsoon();
	}

	messageright = P_imax2((long)messageright,
			m_strwidth(logfont_lfont, msg) + 3);
	m_color((long)gg.color.message);
	m_drawstr(2L, (long)messagepos, logfont_lfont, msg);
	messagepos += spacing;
}

#undef spacing


static void vmessage(char *msg)
{
	strlist_t *l1;

	if (gg.initdone)
	{
		if (gg.verbose || tempverbose)
			message(msg);
	}
	else
	{
		l1 = strlist_append(&messages, msg);
		if (tempverbose)
			l1->value = (na_long)0;
		else
			l1->value = (na_long)1;
	}
	tempverbose = false;
}


static void vmessageflag(char *msg, int b)
{
	char STR1[256];

	if (b)
	{
		sprintf(STR1, "%sON", msg);
		vmessage(STR1);
	}
	else
	{
		sprintf(STR1, "%sOFF", msg);
		vmessage(STR1);
	}
}

void tracemessage(char *msg)
{
	if (!gg.traceflag)
		return;
	if (!gg.traceopen)
	{
		if (*tracefname == '\0')
		{
			do
			{
				sprintf(tracefname, "/tmp/%.10s%03ld",
						cuserid(NULL), rand() % 1000L);

			} while (access(tracefname, F_OK) == 0);
		}
		if (tracefile != NULL)
			tracefile = freopen(tracefname, "a", tracefile);
		else
			tracefile = fopen(tracefname, "a");
		if (tracefile == NULL)
			_EscIO(FileNotFound);
		gg.traceopen = true;
	}
	fprintf(tracefile, "%s\n", msg);
}


static void dumpmessage(char *msg)
{
	if (*dumpfname == '\0')
		return;
	if (!gg.dumpopen) {
		if (dumpfile != NULL)
			dumpfile = freopen(dumpfname, "a", dumpfile);
		else
			dumpfile = fopen(dumpfname, "a");
		if (dumpfile == NULL)
			_EscIO(FileNotFound);
		gg.dumpopen = true;
	}
	fprintf(dumpfile, "%s\n", msg);
}


static void closedump()
{
	if (gg.dumpopen)
	{
		TRY(try4);
		if (dumpfile != NULL)
			fclose(dumpfile);
		dumpfile = NULL;
		RECOVER(try4);
		if (P_escapecode == -20)
			_Escape(P_escapecode);
		ENDTRY(try4);
	}
	gg.dumpopen = false;
}


static void closetrace()
{
	if (gg.traceopen)
	{
		TRY(try5);
		if (tracefile != NULL)
			fclose(tracefile);
		tracefile = NULL;
		RECOVER(try5);
		if (P_escapecode == -20)
			_Escape(P_escapecode);
		ENDTRY(try5);
	}
	gg.traceopen = false;
}


static void setdumpname(char *name_)
{
	char name[256];

	strcpy(name, name_);
	newci_fixfname(name, "text", "");
	closedump();
	strcpy(dumpfname, name);
}


static void settracename(char *name_)
{
	char name[256];

	strcpy(name, name_);
	newci_fixfname(name, "text", "");
	closetrace();
	strcpy(tracefname, name);
	if (gg.traceflag)
		tracemessage("Trace mode ON");
}


static void closedumpfiles()
{
	closedump();
	closetrace();
}


/// Clear the menu area for an error message.
void beginerror()
{
	beginbottom();
	strlist_empty(&thingstodo);
}


/// Clean up after displaying an error message.
void enderror()
{
	nc_printf("  Press any key to continue.");
	m_alpha_on();
	waitforkey();
	endbottom();
}


static void drawhistdivisions(short x1, short x2)
{
	short x, y;

	m_color((long)gg.color.division);
	x = histleft;
	while (x <= x2)
	{
		if (x >= x1)
		{
			y = histdown;
			while (y >= 0)
			{
				m_drawpoint((long)x, (long)y);
				y -= histdivision;
			}
		}
		x += histdivision;
	}
}


static void historypointx(log_htrec *ht, short *x)
{
	gg.actval = (ht->time - histgattr[histfirsttime - 1].UU.r) /
		histgattr[histsecdiv - 1].UU.r;
	if (gg.actval > histdivsacross)
	{
		*x = across * 2;
		return;
	}
	if (gg.actval < 0)
		*x = -across;
	else
		*x = histleft + (long)floor(gg.actval * histdivision + 0.5);
}


static void historyvaluey(log_hnrec *hn, double value, double *outval, short *y)
{
	gg.actval = value;
	gg.actx = hn->sig;
	gg.actgattr = hn->attr;
	gg.acty = gg.color.chart;
	calltoolnode(gg.signaltab[gg.actx - 1].np, act_histval);
	*outval = gg.actval;
	if (gg.actval < -histvalrange)
		gg.actval = -histvalrange;
	else if (gg.actval > histvalrange)
		gg.actval = histvalrange;
	*y = hn->y - (long)floor(gg.actval * histdivision + 0.5);
}  /*also returns in gg.acty*/


static void historypointy(log_hnrec *hn, log_htrec *ht, short *y)
{
	double r;

	historyvaluey(hn, na_srtor(ht->val[hn->num - 1]), &r, y);
}


/** @brief Make one simulation pass through the circuit.
  For each page, call the simulator(s) and do other simulation-related chores.
  */
void pass()
{
	log_grec *g, *g1, *g2;
	log_nrec *n, *n1, *n2;
	log_hwrec *hw;
	log_vwrec *vw;
	log_srec *s;
	log_tool *st, *mintimetool;
	log_htrec *ht;
	log_hnrec *hn;
	long i1;
	short histsweep, hx1, hx2, hy1, hy2;
	int havetime, hadconflicts;
	double mintime, dtime;
	int flag, flag2;
	log_gattrrec *WITH;
	log_hwrec *WITH1;
	log_vwrec *WITH2;
	log_srec *WITH3;

	watchdog = timers_sysclock();
	gg.busyflag = false;
	gg.oldsimstate = gg.simstate;
	gg.oldsimstatetool = gg.simstatetool;
	gg.simstate = simst_null;
	gg.simstatetool = NULL;
	if (!gg.pwrflag)
		gg.fastspeed = gg.fastmin;
	if (gg.showpage > 0 && !modeflag &&
			(timers_sysclock() / 6000 != modetime || gg.probekind != modeprobekind))
		log_setmode("");

	if (gg.showpage > 0 || bottomcount > 0)
	{
		clipon();
	}
	else if (gg.showpage == log_page_history)
	{
		if (bottomcount == 0)
			m_clip((long)histleft, 0L, (long)across, (long)histdown);
		else
			m_clip((long)histleft, 0L, (long)across, baseline - 1L);
	}
	histsweep = histgattr[histsweeptype - 1].UU.nv;
	if (gg.resetflag)
	{
		stamp(&gg.resetstamp);
		gg.time = 0.0;
		gg.singlestepcount = 0;
		histreset = true;
		if (histsweep == histsweep_contin ||
				(histsweep == histsweep_onreset &&
				 (histgattr[histstarttime - 1].blnk ||
				  histgattr[histstarttime - 1].UU.r == 0)))
			gg.histactive = true;
		st = gg.toolbase;
		while (st != NULL)
		{
			if (st->simulator)
				st->deltatime = 0.0;
			st = st->next;
		}
	}

	if (!gg.histactive && (histsweep == histsweep_trig) && (histtrig != 0))
	{
		gg.actflag = false;
		gg.actx = histtrig;
		calltoolnode(gg.signaltab[gg.actx - 1].np, act_trigger);
		if (gg.actflag)
		{
			histreset = true;
			gg.histactive = true;
		}
	}

	if (histreset || hncount != hnocount)
	{
		if (gg.htbase != NULL)
		{
			if (gg.showpage == log_page_history && gg.htbase != gg.htlast)
			{
				historypointx(gg.htbase, &hx1);
				historypointx(gg.htlast, &hx2);
				if (hx1 < histleft)
					hx1 = histleft;
				if (hx2 > across)
					hx2 = across;
				if (hx1 <= hx2)
				{
					m_color((long)gg.color.backgr);
					m_fillrect((long)hx1, 0L, (long)hx2, (long)histdown);
					drawhistdivisions(hx1, hx2);
				}
			}
			while (gg.htbase != NULL)
			{
				ht = gg.htbase->next;
				Free(gg.htbase);
				gg.htbase = ht;
			}
			gg.htlast = NULL;
		}
		htcount = 0;
		WITH = &histgattr[histminstep - 1];
		if (WITH->blnk)
			histtime = gg.time;
		else
			histtime = gg.time - WITH->UU.r;
	}
	if (gg.histactive && !histgattr[histmaxstep - 1].blnk)
	{
		mintime = histgattr[histmaxstep - 1].UU.r;
		havetime = (mintime > 0);
		mintimetool = NULL;
	}
	else
	{
		havetime = false;
	}
	st = gg.toolbase;
	while (st != NULL)
	{
		if (st->simulator)
		{
			st->nexttstep = -1.0;
			calltool(st, act_pass);
			if (st->nexttstep >= 0.0)
			{
				dtime = st->nexttstep - st->deltatime;
				if (dtime < 0.0)
				{
					st->deltatime = st->nexttstep * 0.999999999;
					dtime = st->nexttstep - st->deltatime;
				}
				if (!havetime || dtime < mintime)
				{
					mintime = dtime;
					mintimetool = st;
				}
				havetime = true;
			}
		}
		st = st->next;
	}

	if (gg.histactive && gg.pwrflag &&
			htcount >= histgattr[histmaxcount - 1].UU.U73.i1 &&
			histgattr[histsweeptype - 1].UU.nv != histsweep_contin)
	{
		m_saveclip();
		m_noclip();
		vmessage("Warning: scope memory is 100% full");
		switch (histgattr[histfullaction - 1].UU.nv)
		{

			case 0:
				gg.histactive = false;
				break;

			case 1:
				gg.pwrflag = false;
				vmessage("Simulation is OFF (scope full)");
				break;
		}
		m_unclip();
	}

	if (gg.pwrflag)
	{
		gg.simstatetool = mintimetool;
		if (havetime)
		{
			if (mintime > 0)
			{
				gg.simstate = simst_running;
				if (histsweep == histsweep_onreset)
				{
					flag = (gg.time >= histgattr[histstarttime - 1].UU.r);
					flag2 = (gg.time >= histgattr[histstoptime - 1].UU.r);
				}
				gg.time += mintime;
				gg.prevtimestep = mintime;
				if (histsweep == histsweep_onreset)
				{
					if (!histgattr[histstarttime - 1].blnk && !flag &&
							gg.time >= histgattr[histstarttime - 1].UU.r)
						gg.histactive = true;
					if (!histgattr[histstoptime - 1].blnk && !flag2 &&
							gg.time >= histgattr[histstoptime - 1].UU.r)
						gg.histactive = false;
				}
				st = gg.toolbase;
				while (st != NULL)
				{
					if (st->simulator) {
						if (st->nexttstep > 0.0)
						{
							gg.actflag = (st->nexttstep - st->deltatime == mintime);
							if (gg.actflag)
								st->deltatime = 0.0;
							else
								st->deltatime += mintime;
							calltool(st, act_tstep);
						}
					}
					st = st->next;
				}
				if (gg.histactive && hncount > 0 &&
						(histgattr[histminstep - 1].blnk ||
						 gg.time - histtime >= histgattr[histminstep - 1].UU.r))
				{
					if (htcount >= histgattr[histmaxcount - 1].UU.U73.i1)
					{
						if (gg.showpage == log_page_history)
							historypointx(gg.htbase, &hx1);
						while (htcount >= histgattr[histmaxcount - 1].UU.U73.i1)
						{
							ht = gg.htbase->next;
							Free(gg.htbase);
							gg.htbase = ht;
							htcount--;
						}
						if (gg.showpage == log_page_history)
						{
							if (gg.htbase != NULL)
								historypointx(gg.htbase, &hx2);
							else
								hx2 = across;
							if (hx2 >= hx1)
							{
								m_color((long)gg.color.backgr);
								m_fillrect((long)hx1, 0L, (long)hx2, (long)histdown);
								drawhistdivisions(hx1, hx2);
							}
						}
					}
					if (gg.htbase == NULL)
					{
						histgattr[histfirsttime - 1].UU.r = gg.time;
						histgattr[histfirsttime - 1].changed = true;
						histonscreen = false;
					}

					ht = (log_htrec *)Malloc(sizeof(log_htrec) -
							log_million * sizeof(void *) +
							hncount * sizeof(void *));

					ht->next = NULL;
					ht->time = gg.time;
					hn = gg.hnbase;
					i1 = 0;
					while (hn != NULL)
					{
						gg.actval = 0.0;
						gg.actx = hn->sig;
						gg.actgattr = hn->attr;
						calltoolnode(gg.signaltab[gg.actx - 1].np, act_history);
						i1++;
						ht->val[i1 - 1] = na_rtosr(gg.actval);
						hn = hn->next;
					}

					if (gg.showpage == log_page_history && gg.htlast != NULL)
					{
						historypointx(gg.htlast, &hx1);
						historypointx(ht, &hx2);
						hn = gg.hnbase;
						while (hn != NULL)
						{
							historypointy(hn, gg.htlast, &hy1);
							historypointy(hn, ht, &hy2);
							if ((unsigned long)gg.acty <= 15)
							{
								m_color(gg.acty);
								m_drawline((long)hx1, (long)hy1, (long)hx2, (long)hy2);
							}
							hn = hn->next;
						}
					}

					if (gg.htlast == NULL)
						gg.htbase = ht;
					else
						gg.htlast->next = ht;

					gg.htlast = ht;
					htcount++;
					histtime = gg.time;
					if (histsweep != histsweep_contin)
					{
						i1 = histgattr[histmaxcount - 1].UU.U73.i1;
						if (htcount >= i1 * 9 / 10)
						{
							if (htcount == i1 * 9 / 10)
							{
								m_noclip();
								vmessage("Warning: scope memory is 90% full");
							}
							if (htcount == i1 * 19 / 20)
							{
								m_noclip();
								vmessage("Warning: scope memory is 95% full");
							}
						}
					}
				}
			}
			else
				gg.simstate = simst_notready;
		}
		else
			gg.simstate = simst_notactive;

		if (gg.glowmode && gg.showpage > 0 && !gg.invisible && !gg.showconflicts)
		{
			flag = false;
			hw = gg.pages[gg.curpage - 1]->hwbase;
			while (hw != NULL)
			{
				WITH1 = hw;
				if ((!WITH1->node->conflict1 || !WITH1->node->conflict2) &&
						WITH1->node->changed && hw != gg.movinghw)
				{
					hidecursor_hook();
					m_color(glowcol(hw->node,
								(long)gg.color.wire[hw->wcolr - log_wcol_normal]));
					hline(WITH1->x1, WITH1->x2, WITH1->y);
					flag = true;
				}
				hw = WITH1->next;
			}
			vw = gg.pages[gg.curpage - 1]->vwbase;
			while (vw != NULL)
			{
				WITH2 = vw;
				if ((!WITH2->node->conflict1 || !WITH2->node->conflict2) &&
						WITH2->node->changed && vw != gg.movingvw)
				{
					hidecursor_hook();
					m_color(glowcol(vw->node,
								(long)gg.color.wire[vw->wcolr - log_wcol_normal]));
					vline(WITH2->x, WITH2->y1, WITH2->y2);
					flag = true;
				}
				vw = WITH2->next;
			}
			if (glowsolder)
			{
				s = gg.pages[gg.curpage - 1]->sbase;
				while (s != NULL)
				{
					WITH3 = s;
					n = WITH3->hwire->node;
					if ((!n->conflict1 || !n->conflict2) && n->changed &&
							WITH3->hwire != gg.movinghw && WITH3->vwire != gg.movingvw)
					{
						hidecursor_hook();
						drawsolderc(WITH3->x, WITH3->y,
								(int)glowcol(n, (long)gg.color.solder));
						flag = true;
					}
					s = WITH3->next;
				}
			}
			if (flag)
			{
				n = gg.nbase;
				while (n != NULL)
				{
					n->changed = false;
					n = n->next;
				}
				unhidecursor_hook();
			}
		}  /*if gg.glowmode*/
	}
	else
		gg.simstate = simst_off;
	hadconflicts = anyconflicts;
	anyconflicts = false;
	n = nodeconflictbase;
	n1 = NULL;
	while (n != NULL)
	{
		n2 = n->confnext;
		if (n->conflict)
		{
			flag2 = (n->conflict1 && n->conflict2);
			if (gg.simstate == simst_running && !flag2 && n->oconflict)
			{
				n->conflict1 = !n->conflict1;
				if (!n->conflict1)
					n->conflict2 = true;
			}
			if (n->conflict1 && n->conflict2)
			{
				anyconflicts = true;
				if ((gg.refrflag || !flag2 || !n->oconflict) && gg.showpage > 0)
					drawnodec(n, gg.color.conflict);
			}
			n->oconflict = true;
		}
		else
		{
			if (gg.showpage > 0)
				drawnode(n);
			if (n1 == NULL)
				nodeconflictbase = n2;
			else
				n1->confnext = n2;
			n->oconflict = false;
			n->conflict1 = false;
			n->conflict2 = false;
			n = n1;
		}
		n1 = n;
		n = n2;
	}
	g = gateconflictbase;
	g1 = NULL;
	while (g != NULL)
	{
		g2 = g->confnext;
		if (g->conflict)
		{
			flag2 = (g->conflict1 && g->conflict2);
			if (gg.simstate == simst_running && !flag2 && g->oconflict)
			{
				g->conflict1 = !g->conflict1;
				if (!g->conflict1)
					g->conflict2 = true;
			}
			if (g->conflict2 && g->conflict2)
			{
				anyconflicts = true;
				if ((gg.refrflag || !flag2 || !g->oconflict) && gg.showpage > 0)
					drawgatexc(g, gg.color.conflict);
			}
			g->oconflict = true;
		}
		else
		{
			if (gg.showpage > 0)
				drawgatex(g);
			if (g1 == NULL)
				gateconflictbase = g2;
			else
				g1->confnext = g2;
			g->oconflict = false;
			g->conflict1 = false;
			g->conflict2 = false;
			g = g1;
		}
		g1 = g;
		g = g2;
	}
	/* clipoff; */
	if (gg.showpage > 0 && bottomcount == 0)
		refrbaseline(false);
	gg.baselinecolor = gg.color.baseline;
	gg.resetflag = false;
	histreset = false;
	hnocount = hncount;
	if (gg.singlestepcount > 0 && gg.simstate == simst_running)
	{
		gg.singlestepcount--;
		if (gg.singlestepcount == 0) {
			gg.pwrflag = false;
			if (steppingoff)
				vmessage("Simulation is OFF (single-step)");
			steppingoff = false;
		}
	}
	if (conflictstop && anyconflicts && !hadconflicts)
	{
		gg.pwrflag = false;
		vmessage("Simulation is OFF (conflict occurred)");
	}
	gg.refrflag = false;
}


static void clearconflicts(log_tool *tool)
{
	log_grec *g;
	log_nrec *n;

	n = nodeconflictbase;
	while (n != NULL)
	{
		if (n->simtype == tool)
			n->conflict = false;
		n = n->confnext;
	}
	g = gateconflictbase;
	while (g != NULL)
	{
		if (g->kind->simtype == tool)
			g->conflict = false;
		g = g->confnext;
	}
}


short getsignal (int d, char *n);


static void defsimulator(log_action_t *act)
{
	switch (act->action)
	{

		case act_init:
			act->acttool->simulator = true;
			act->acttool->ready = true;
			break;

		default:
			break;
	}
}


log_tool *findsimtype(uchar st)
{
	char simname[17];
	log_tool *stp;
	int rdy;

	sprintf(simname, "%d", st);
	stp = findtool(simname);
	if (gg.traceflag)
		fprintf(tracefile, "Findsimtype of %d gets %p\n", st, stp);
	if (!stp->ready || !stp->simulator)
	{
		rdy = stp->ready;
		if (rdy)
			calltool(stp, act_exit);
		stp->proc = defsimulator;
		gg.action = act_init;
		gg.acttool = stp;
		defsimulator(&gg);
		stp->nexttstep = 0.0;
		stp->nnumattrs = 0;
		stp->nattr = NULL;
		stp->hnumattrs = 0;
		stp->hattr = NULL;
		beginerror();
		if (rdy)
			nc_printf("Tool %d is not a simulator!\n", st);
		else
			nc_printf("No program loaded for gate-simulation type %d\n", st);
		enderror();
	}
	stp->keep = true;
	stp->simtype = st;
	if (!strcmp(stp->shortname, simname))
		sprintf(stp->shortname, "Type %s", simname);
	return stp;
}


static void dumpconflicts()
{
	/*no longer needed (I hope!)*/
	log_grec *g;
	log_nrec *n;

	n = nodeconflictbase;
	g = gateconflictbase;
	nc_printf("\f");
	if (n != NULL || g != NULL)
	{
		nc_printf("Conflicts:\n");
		while (n != NULL)
		{
			nc_printf("   Conflict node %ld  c=%c  delay=%d  oc=%c\n",
					(long)n, n->conflict ? 'T' : 'F',
					3 - n->conflict - n->conflict2 * 2, n->oconflict ? 'T' : 'F');
			n = n->confnext;
		}
		while (g != NULL)
		{
			nc_printf("   Conflict gate %ld  c=%c  delay=%d  oc=%c\n",
					(long)g, g->conflict ? 'T' : 'F',
					3 - g->conflict - g->conflict2 * 2, g->oconflict ? 'T' : 'F');
			g = g->confnext;
		}
	}
	else
	{
		nc_printf("No conflicts registered.\n");
	}
	m_alpha_on();
}


static void purgesignaltab()
{
	short i, pg;
	log_nrec *n1;
	log_hnrec *hn;
	log_grec *g;
	short FORLIM;
	log_sigrec *WITH;

	FORLIM = gg.maxsignal;
	for (i = 0; i < FORLIM; i++)
		gg.signaltab[i].f = gg.signaltab[i].keep;

	FORLIM = gg.numpages;
	for (pg = 0; pg < FORLIM; pg++)
	{
		g = gg.pages[pg]->gbase;
		while (g != NULL)
		{
			if (g->sig != 0)
				gg.signaltab[g->sig - 1].f = true;
			g = g->next;
		}
	}

	if (copybuf.valid)
	{
		g = copybuf.gcopy;
		while (g != NULL)
		{
			if (g->sig != 0)
				gg.signaltab[g->sig - 1].f = true;
			g = g->next;
		}
	}

	if (histtrig != 0)
		gg.signaltab[histtrig - 1].f = true;

	hn = gg.hnbase;
	while (hn != NULL)
	{
		gg.signaltab[hn->sig - 1].f = true;
		hn = hn->next;
	}

	FORLIM = kindgroupsize;
	for (i = 0; i < FORLIM; i++)
	{
		if (kindsig[i] != 0)
			gg.signaltab[kindsig[i] - 1].f = true;
	}

	g = gg.neargate;
	if (g != NULL && g->sig != 0)
		gg.signaltab[g->sig - 1].f = true;

	gg.lastsignal = 0;
	FORLIM = gg.maxsignal;
	for (i = 1; i <= FORLIM; i++)
	{
		if (!gg.signaltab[i - 1].f)
		{
			if (gg.signaltab[i -1].name && gg.signaltab[i - 1].np)
			{
				WITH = &gg.signaltab[i - 1];
				if (WITH->name != NULL)
					strdispose(&WITH->name);
				n1 = gg.signaltab[i - 1].np;
				if (n1 && nodeexists(n1))
					unrefnode(&gg.signaltab[i - 1].np);
				gg.signaltab[i - 1].np = NULL;
			}
		}
		else
		{
			gg.lastsignal = i;
		}
	}
}


/// Check each node and get rid of ones that are no longer used.
void garbagecoll()
{
	log_nrec *node;
	log_grec *g;
	log_hwrec *hw;
	log_vwrec *vw;
	short pg, FORLIM1;

	node = gg.nbase;
	while (node != NULL)
	{
		if (!node->keep)
			node->ref = 0;
		node = node->next;
	}

	for (pg = 0; pg < gg.numpages; pg++)
	{
		hw = gg.pages[pg]->hwbase;
		while (hw != NULL)
		{
			hw->node->ref++;
			hw = hw->next;
		}

		vw = gg.pages[pg]->vwbase;
		while (vw != NULL)
		{
			vw->node->ref++;
			vw = vw->next;
		}

		g = gg.pages[pg]->gbase;
		while (g != NULL)
		{
			FORLIM1 = g->kind->numpins;
			for (short i = 0; i < FORLIM1; i++)
				g->pin[i]->ref++;
			g = g->next;
		}
	}

	for (short i = 0; i < gg.lastsignal; i++)
	{
		if (gg.signaltab[i].np != NULL)
			gg.signaltab[i].np->ref++;
	}
	callsimtools(act_refnodes);
	purgesignaltab();

	node = gg.nbase;
	while (node != NULL)
	{
		log_nrec *nextnode = node->next;
		if (node->ref == 0 && !node->keep)
			disposenode(&node);
		node = nextnode;
	}
}


/** @brief Find the number associated with a signal name.

  If it doesn't exist, create it.
  If not enough room, issue an error and
  return zero.
  */
short getsignal(int d, char *n_)
{
	char n[256];
	short i;
	char STR1[256], STR2[256];
	short FORLIM;
	log_sigrec *WITH;

	strcpy(n, n_);
	strcpy(STR1, strltrim(strrtrim(strcpy(STR2, n))));
	strcpy(n, STR1);
	if (*n == '\0')
	{
		i = d;
	}
	else
	{
		i = 1;
		while (i <= gg.lastsignal &&
				(gg.signaltab[i - 1].name == NULL ||
				 strcmp(gg.signaltab[i - 1].name, n)))
			i++;
		if (i > gg.lastsignal)
		{
			stamp(&gg.sigstamp);
			i = 1;
			while (i <= gg.maxsignal && gg.signaltab[i - 1].name != NULL)
				i++;
			if (i > gg.maxsignal)
			{
				beginerror();
				nc_printf("Only %d distinct signal names are allowed!\n", gg.maxsignal);
				enderror();
				i = 0;
			}
			if (i != 0)
			{
				WITH = &gg.signaltab[i - 1];
				WITH->name = strdup(n);
				WITH->keep = false;
				newnoderef(&WITH->np, 0, 1L);
				WITH->temp = (na_long)0;
			}
		}
	}
	if (i > gg.lastsignal)
		gg.lastsignal = i;
	return i;
}


static void getsigname(char *name, log_sigrec **sig)
{
	short i;

	i = getsignal(0, name);
	if (i != 0)
	{
		*sig = &gg.signaltab[i - 1];
		(*sig)->keep = true;
	}
	else
	{
		*sig = NULL;
	}
}


static void setupregion(log_regrec **r, short pagenum)
{
	log_regrec *r2;

	if (pagenum < 1 || pagenum > gg.numpages)
	{
		if (*r != NULL)
		{
			if (*r == gg.pages[(*r)->pagenum - 1]->pageregions)
				gg.pages[(*r)->pagenum - 1]->pageregions = (*r)->next;
			else
			{
				r2 = gg.pages[(*r)->pagenum - 1]->pageregions;
				while (r2 != NULL && r2->next != *r)
					r2 = r2->next;
				if (r2 != NULL)
					r2->next = (*r)->next;
			}
			Free(*r);
		}
		while (gg.numpages > gg.curpage && isPageNull(gg.pages[gg.numpages - 1]))
			gg.numpages--;
		return;
	}
	*r = (log_regrec *)Malloc(sizeof(log_regrec));
	(*r)->pagenum = pagenum;
	(*r)->x1 = -32768L;
	(*r)->y1 = -32768L;
	(*r)->x2 = 32767;
	(*r)->y2 = 32767;
	stamp(&(*r)->regstamp);
	(*r)->info = (na_long)0;
	(*r)->next = gg.pages[pagenum - 1]->pageregions;
	gg.pages[pagenum - 1]->pageregions = *r;
}

/// Return a string containing the name of VLSI wire color C.
char *colorname(char *Result, short c)
{
	if (c == log_wcol_normal)
		return strcpy(Result, "U");
	else if (c == log_wcol_green)
		return strcpy(Result, "G");
	else if (c == log_wcol_red)
		return strcpy(Result, "R");
	else if (c == log_wcol_blue)
		return strcpy(Result, "B");
	else if (c == log_wcol_yellow)
		return strcpy(Result, "Y");
	else {
		sprintf(Result, "%02d", c);
		return Result;
	}
}

/// Display program Help information.
static void gethelp(char *s)
{
	char cmdline[256];
	char *pager;

	vmessage("Starting an xterm for help");

	pager = getenv("PAGER");
	if (!pager)
		pager = "more";

	if (*m_display_name == '\0')
		sprintf(cmdline, "xterm -e %s %s & \n", pager, loghelpname);
	else
		sprintf(cmdline, "xterm -display %s -e %s %s & \n",
				m_display_name, pager, loghelpname);

	system(cmdline);
}

/// Simulate READLN, but keep the circuit running by calling PASS.
static void readlnpass(char *s, short mode)
{
	uchar ch;
	short i, xleft, oxleft;
	long x, y;
	int redraw, starting;
	char STR1[256];

	m_alpha_on();
	if (mode != 3)   /*mode 0: normal*/
		*s = '\0';
	/*mode 2: exit on knob up/dn*/
	nc_getXY(&x, &y);   /*mode 3: 2 + keep default*/
	oxleft = 1;
	xleft = 1;
	i = 1;
	starting = true;
	redraw = true;
	do
	{
		xleft = P_imax2(1L, x + i - txacross + 1);
		if (redraw || xleft != oxleft)
		{
			hidecursor();
			nc_gotoXY((int)x, (int)y);
			nc_printf("%s\t", strsub(STR1, s, xleft, (int)(txacross - x)));
			hidecursor();
			oxleft = xleft;
			redraw = false;
		}
		nc_gotoXY((int)(x + i - xleft), (int)y);
		nc_cursor_on();
		if (!pollkbd2())
		{
			do
			{
				pass();
				pen();
			} while (!(pollkbd2() || gg.t.dn));
		}
		nc_cursor_off();
		if (gg.t.dn)
			ch = '\015';
		else
			ch = inkey2();
		if (ch == '\b' && i > 1)
			i--;
		else if (ch == '\034' && i <= strlen(s))
			i++;
		else if (ch == '\007' && i > 1)
		{
			i--;
			strcpy_overlap(s + i - 1, s + i);
			redraw = true;
		}
		else if (ch == '\177')
		{
			*s = '\0';
			i = 1;
			redraw = true;
		}
		else if (ch >= ' ' && ch != 250 && ch != 251 && strlen(s) < 255)
		{
			if (starting)
				*s = '\0';
			sprintf(STR1, "%c%s", ch, s + i - 1);
			strcpy(s + i - 1, STR1);
			i++;
			redraw = true;
		}
		if ((unsigned)mode < 32 && ((1L << mode) & 0xc) != 0 && ch < 32 &&
				((1L << ch) & 0x80000400L) != 0) {
			ungetkey2(ch);
			ch = '\015';
		}
		starting = false;
	} while (ch >= 32 || ((1L << ch) & 0x2008) == 0);
	remcursor();
	sprintf(STR1, "%c", ch);
	if (!strcmp(STR1, EXEC))
	{
		*s = '\0';
		nc_gotoXY((int)x, (int)y);
		nc_putchar('\t');
	}
	if (gg.t.dn)
	{
		gg.t.dn = false;
		gg.t.depressed = false;
	}
}


void clearfunc()
{
	if (doingcnffunction)
		*gg.func = '\0';
	else
		assertfunc("");
}

void refrfunc()
{
	if (gg.initdone && !doingcnffunction)
		assertfunc("REFR");
	else
		clearfunc();
}

void doimmedcnffunction()
{
	char STR1[256], STR2[256];

	if (*gg.func == '\0')
		return;
	if (!strcmp(gg.func, "SNAP"))
	{
		getbool(gg.funcarg, &snapflag);
		vmessageflag("Snap-to-grid is ", snapflag);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "GLOW"))
	{
		getbool(gg.funcarg, &gg.glowmode);
		refrfunc();
		vmessageflag("Glow mode is ", gg.glowmode);
		return;
	}
	if (!strcmp(gg.func, "AUTOWINDOW"))
	{
		getbool(gg.funcarg, &m_autoraise);
		refrfunc();
		vmessageflag("AUTOWINDOW mode is ", m_autoraise);
		return;
	}
	if (!strcmp(gg.func, "GLOWSOLDER"))
	{
		getbool(gg.funcarg, &glowsolder);
		refrfunc();
		vmessageflag("Glowing solder mode is ", glowsolder);
		return;
	}
	if (!strcmp(gg.func, "PROBE"))
	{
		if (gg.initdone)
			remcursor();
		getbool(gg.funcarg, &gg.probemode);
		briefprobe = false;
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "DOTS"))
	{
		getbool(gg.funcarg, &gg.dotsvisible);
		refrfunc();
		if (gg.dotsvisible)
			vmessage("Dots are now visible");
		else
			vmessage("Dots are now hidden");
		return;
	}
	if (!strcmp(gg.func, "ONOFF") || !strcmp(gg.func, "SIM") ||
			!strcmp(gg.func, "POWER"))
	{
		getbool(gg.funcarg, &gg.pwrflag);
		vmessageflag("Simulation is ", gg.pwrflag);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "ON"))
	{
		gg.pwrflag = true;
		vmessageflag("Simulation is ", gg.pwrflag);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "OFF"))
	{
		gg.pwrflag = false;
		vmessageflag("Simulation is ", gg.pwrflag);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "INVISIBLE"))
	{
		getbool(gg.funcarg, &gg.invisible);
		refrfunc();
		vmessageflag("Invisible mode is ", gg.invisible);
		return;
	}
	if (!strcmp(gg.func, "INVLABEL"))
	{
		getbool(gg.funcarg, &gg.textinvisible);
		refrfunc();
		vmessageflag("Invisible-labels mode is ", gg.textinvisible);
		return;
	}
	if (!strcmp(gg.func, "INVPINNUM"))
	{
		getbool(gg.funcarg, &gg.pnuminvisible);
		refrfunc();
		vmessageflag("Invisible-pin-number mode is ", gg.pnuminvisible);
		return;
	}
	if (!strcmp(gg.func, "SHOWCONFLICTS"))
	{
		getbool(gg.funcarg, &gg.showconflicts);
		refrfunc();
		if (gg.showconflicts)
			vmessage("Conflicts are highlighted");
		else
			vmessage("Conflicts are not highlighted");
		return;
	}
	if (!strcmp(gg.func, "QUIET"))
	{
		getbool(gg.funcarg, &gg.quiet);
		vmessageflag("Quiet mode is ", gg.quiet);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "CONFLICT"))
	{
		if (*gg.funcarg != '\0' && isdigit(gg.funcarg[0]))
		{
			conflictenbl = true;
			conflictdelay = getint(gg.funcarg, 1L);
			if (conflictdelay > 3)   /*from 0 to 3*/
				conflictdelay = 3;
		}
		else
			getbool(gg.funcarg, &conflictenbl);
		if (conflictenbl)
		{
			sprintf(STR2, "Conflicts are enabled with delay %d", conflictdelay);
			vmessage(STR2);
		}
		else
			vmessage("Conflicts are disabled");
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "CSTOP"))
	{
		getbool(gg.funcarg, &conflictstop);
		vmessageflag("Conflict-stopping is ", conflictstop);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "RESPONSE"))
	{
		gg.fastmin = getint(gg.funcarg, gg.fastmin);
		gg.fastmax = getint(gg.funcarg, gg.fastmax);
		gg.fastrate = getint(gg.funcarg, gg.fastrate);
		if (gg.fastmax < gg.fastmin)
			gg.fastmax = gg.fastmin;
		if (gg.fastspeed < gg.fastmin)
			gg.fastspeed = gg.fastmin;
		else if (gg.fastspeed > gg.fastmax)
			gg.fastspeed = gg.fastmax;
		if (gg.fastrate < 1)
			gg.fastrate = 1;
		sprintf(STR1, "Response: min=%ld, max=%ld, rate=%ld",
				gg.fastmin, gg.fastmax, gg.fastrate);
		vmessage(STR1);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "VERBOSE"))
	{
		getbool(gg.funcarg, &gg.verbose);
		if (gg.verbose)
			message("Verbose mode is ON");
		else
			message("Verbose mode is OFF");
		clearfunc();
		return;
	}
	if (strcmp(gg.func, "AVOID"))
		return;
	if (gg.initdone)
		remcursor();
	clearfunc();
}


void doimmedfunction()
{
	log_tool *tp;
	char cmd[256];
	strlist_t *l1;

	if (*gg.func != '\0')
	{
		l1 = strlist_find(commandlist, gg.func);
		if (l1 != NULL) {
			if (*(char *)l1->value != '\0')
			{
				tp = findtool((char *)l1->value);
				if (tp->ready)
					strlist_delete(&commandlist, &l1);
			}
		}
		doingcnffunction = false;
		strcpy(cmd, gg.func);
		tp = gg.toolbase;
		while (tp != NULL && !strcmp(gg.func, cmd))
		{
			if (tp->ready)
				calltool(tp, act_immed);
			tp = tp->next;
		}
		doimmedcnffunction();
	}
	if (*gg.func == '\0')
		return;
	if (!strcmp(gg.func, "HELP"))
	{
		gethelp(gg.funcarg);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "RESET"))
	{
		gg.resetflag = true;
		vmessage("Resetting simulation");
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "TRACE"))
	{
		getbool(gg.funcarg, &gg.traceflag);
		vmessageflag("Trace mode ", gg.traceflag);
		if (gg.traceflag)
			tracemessage("Trace mode ON");
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "DUMPFILE"))
	{
		setdumpname(gg.funcarg);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "TRACEFILE"))
	{
		settracename(gg.funcarg);
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "CLOSEFILES"))
	{
		closedumpfiles();
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "DUMP"))
	{
		dumpnodes();
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "HOME"))
	{
		zoomto(0);
		refrfunc();
		xoff0 = origin - gg.xoff;
		yoff0 = origin - gg.yoff;
		setscale(0);
		refrfunc();
		return;
	}
	if (!strcmp(gg.func, "ZOOMDN"))
	{
		zoomto(zoom - 1);
		return;
	}
	if (!strcmp(gg.func, "ZOOMUP"))
	{
		zoomto(zoom + 1);
		return;
	}
	if (!strcmp(gg.func, "GRID"))
	{
		remcursor();
		if (cursortype == grid)
			cursortype = normal;
		else
			cursortype = grid;
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "EXAMINE"))
	{
		clearfunc();
		remcursor();
		gg.probemode = true;
		briefprobe = true;
		return;
	}
	if (!strcmp(gg.func, "ARROW"))
	{
		remcursor();
		cursortype = normal;
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "ALTPOSN"))
	{
		xoff0 = xoffp[gg.curpage - 1] - gg.xoff;
		yoff0 = yoffp[gg.curpage - 1] - gg.yoff;
		xoffp[gg.curpage - 1] = gg.xoff;
		yoffp[gg.curpage - 1] = gg.yoff;
		refrfunc();
		return;
	}
	if (!strcmp(gg.func, "PAGE"))
	{
		if (!strcmp(gg.funcarg, "+"))
		{
			if (gg.curpage < log_maxpages)
				newpage((int)(gg.curpage + 1));
		}
		else if (!strcmp(gg.funcarg, "-"))
		{
			if (gg.curpage > 1)
				newpage((int)(gg.curpage - 1));
		}
		else if (strlen(gg.funcarg) == 1 && gg.funcarg[0] >= '1' &&
				gg.funcarg[0] <= '9')
			newpage(gg.funcarg[0] - 48);
		refrfunc();
		return;
	}
	if (strcmp(gg.func, "DO"))
	{
		if (immedscroll)
			scroll();
		return;
	}
	if (*gg.funcarg != '\0')
	{
		assertfunc(gg.funcarg);
		return;
	}
	beginbottom();
	m_alpha_on();
	nc_printf("Do: ");
	readlnpass(cmd, 0);
	endbottom();
	assertfunc(cmd);
}


void assertfunc(char *name_)
{
	char name[256];
	char cmd[17];

	strcpy(name, name_);
	remcursor();
	commandfound = true;
	while (*name == ':')
		strcpy_overlap(name, name + 1);
	getword(name, cmd);
	doingcnffunction = false;
	if (!strcmp(gg.func, cmd))
	{
		*gg.func = '\0';
		return;
	}
	strcpy(gg.func, cmd);
	strcpy(gg.funcarg, name);
	if (briefprobe)
	{
		remcursor();
		gg.probemode = false;
	}
	doimmedfunction();
}



void confirmsimtype(log_nrec *n)
{
	log_grec *g;
	short i, pg;
	int found;
	log_hnrec *hn;
	short FORLIM, FORLIM1;

	if (gg.traceflag)
		fprintf(tracefile, "Confirm simtype of %ld\n", (long)n);
	if (n == NULL || n->simtype->simtype < 8)
		return;
	found = false;
	FORLIM = gg.numpages;
	for (pg = 0; pg < FORLIM; pg++)
	{
		g = gg.pages[pg]->gbase;
		while (g != NULL && !found)
		{
			FORLIM1 = g->kind->numpins;
			for (i = 0; i < FORLIM1; i++)
			{
				if (g->pin[i] == n)
				{
					if (g->kind->pin[i].s >= 8)
						found = true;
				}
			}
			g = g->next;
		}
	}

	if (found)
		return;

	if (gg.traceflag)
		fprintf(tracefile, "Deleting simtype of %ld\n", (long)n);
	calltoolnode(n, act_disposenode);
	disposeattrs(&n->attr, n->simtype->nnumattrs, n->simtype->nattr);
	hn = gg.hnbase;
	while (hn != NULL)
	{
		if (gg.signaltab[hn->sig - 1].np == n)
			disposeattrs(&hn->attr, n->simtype->hnumattrs, n->simtype->hattr);
		hn = hn->next;
	}
	n->simtype = simtype_ignore;
}


void scancn (cnrec *cn, struct LOC_checkcombine *LINK);

static void scan(log_nrec **n, struct LOC_checkcombine *LINK)
{
	cnrec *cn;

	if (!LINK->good)
		return;
	if (*n != NULL)
		checknode(*n, LINK);
	cn = *LINK->cnbase;
	while (cn != NULL && LINK->good)
	{
		if ((cn->n == *n && *n != NULL) || cn->np == n)
			scancn(cn, LINK);
		cn = cn->next;
	}
}

void scancn(cnrec *cn, struct LOC_checkcombine *LINK)
{
	if (cn->done)
		return;
	cn->done = true;
	scan(&cn->n, LINK);
	scan(cn->np, LINK);
}

void frysolder(short x, short y)
{
	long t0;

	remcursor();
	clipon();
	t0 = timers_sysclock();
	while (labs(timers_sysclock() - t0) < frytime)
	{
		drawsolderc(x, y, gg.color.conflict);
		drawsolderc(x, y, gg.color.backgr);
	}
	clipoff();
	refreshsoon();
}


void addblobs(blobrec **blbase, short x1, short y1, short x2, short y2)
{
	blobrec *blp;
	log_srec *s;

	s = gg.pages[gg.curpage - 1]->sbase;
	while (s != NULL)
	{
		if (P_ibetween((long)x1, (long)s->x, (long)x2) &&
				P_ibetween((long)y1, (long)s->y, (long)y2))
		{
			blp = (blobrec *)Malloc(sizeof(blobrec));
			blp->x = s->x;
			blp->y = s->y;
			blp->next = *blbase;
			*blbase = blp;
		}
		s = s->next;
	}
}


void doblobs(blobrec *blp)
{
	log_hwrec *hw;
	log_vwrec *vw;
	short x, y;

	while (blp != NULL)
	{
		if (blp->x != -32768L)
		{
			x = blp->x;
			y = blp->y;
			hw = gg.pages[gg.curpage - 1]->hwbase;
			while (hw != NULL && (hw->y != y || hw->x1 > x || hw->x2 < x))
				hw = hw->next;
			vw = gg.pages[gg.curpage - 1]->vwbase;
			while (vw != NULL && (vw->x != x || vw->y1 > y || vw->y2 < y))
				vw = vw->next;
			if (hw != NULL && vw != NULL && hw->node != vw->node)
			{
				if (trycombinenodes(&hw->node, &vw->node)) {
					clipon();
					addsolder(x, y, hw, NULL, vw, NULL);
					clipoff();
				}
			}
		}
		blp = blp->next;
	}
}


void dispblobs(blobrec **blbase)
{
	blobrec *blp;

	while (*blbase != NULL)
	{
		blp = *blbase;
		*blbase = blp->next;
		Free(blp);
	}
}


static void initbuf(baseptrs *bases)
{
	bases->gcopy = NULL;
	bases->scopy = NULL;
	bases->hwcopy = NULL;
	bases->vwcopy = NULL;
	bases->lcopy = NULL;
	bases->bcopy = NULL;
	bases->pgnum = 0;
	bases->valid = false;
}

/// Clear a copy buffer.
static void clearbuf(baseptrs *bases)
{
	log_grec *g;
	log_srec *s;
	log_hwrec *hw;
	log_vwrec *vw;
	log_lrec *l;
	log_brec *b;

	while (bases->gcopy != NULL)
	{
		g = bases->gcopy->next;
		disposegate(&bases->gcopy);
		bases->gcopy = g;
	}

	while (bases->scopy != NULL)
	{
		s = bases->scopy->next;
		Free(bases->scopy);
		bases->scopy = s;
	}

	while (bases->hwcopy != NULL)
	{
		hw = bases->hwcopy->next;
		Free(bases->hwcopy);
		bases->hwcopy = hw;
	}

	while (bases->vwcopy != NULL)
	{
		vw = bases->vwcopy->next;
		Free(bases->vwcopy);
		bases->vwcopy = vw;
	}

	while (bases->lcopy != NULL)
	{
		l = bases->lcopy->next;
		Free(bases->lcopy);
		bases->lcopy = l;
	}

	while (bases->bcopy != NULL)
	{
		b = bases->bcopy->next;
		Free(bases->bcopy);
		bases->bcopy = b;
	}

	bases->pgnum = 0;
	bases->valid = false;
}


static int bufempty(baseptrs *bases)
{
	return (bases->gcopy == NULL && bases->scopy == NULL &&
			bases->hwcopy == NULL && bases->vwcopy == NULL &&
			bases->lcopy == NULL && bases->bcopy == NULL);
}


static int bufissmall(baseptrs *bases)
{
	return ((bases->gcopy == NULL || bases->gcopy->next == NULL ||
				bases->gcopy->next->next == NULL ||
				bases->gcopy->next->next->next == NULL) &&
			(bases->hwcopy == NULL || bases->hwcopy->next == NULL ||
			 bases->hwcopy->next->next == NULL ||
			 bases->hwcopy->next->next->next == NULL) &&
			(bases->vwcopy == NULL || bases->vwcopy->next == NULL ||
			 bases->vwcopy->next->next == NULL ||
			 bases->vwcopy->next->next->next == NULL) &&
			(bases->lcopy == NULL || bases->lcopy->next == NULL ||
			 bases->lcopy->next->next == NULL) &&
			(bases->bcopy == NULL || bases->bcopy->next == NULL ||
			 bases->bcopy->next->next == NULL));
}


/** @brief Cut or Copy all gates/wires/solder points in a rectangular area.
  Previous contents of Copy buffer are lost.
  */
static void cutcopy(baseptrs *bases, short x1, short y1, short x2, short y2, int cut, int tap)
{
	log_grec *g, *g1, *g2;
	log_hwrec *hw, *hw1, *hw2;
	log_vwrec *vw, *vw1, *vw2;
	log_srec *s, *s2;
	log_lrec *l, *l1, *l2;
	log_brec *b, *b1, *b2;
	short xx1, yy1, xx2, yy2, anchorx, anchory;
	int found;
	blobrec *blbase;

	clearbuf(bases);
	anchorx = x2;
	anchory = y2;
	if (tap)
	{
		x2 = x1;
		y2 = y1;
		closergate(x1, y1);
		if (gg.nearlabel == NULL && gg.nearbox == NULL && gg.neargate == NULL)
		{
			closerwire(x1, y1);
			if (gg.nearhw != NULL)
			{
				gg.nearvw = NULL;
				x1 = gg.nearhw->x1;
				x2 = gg.nearhw->x2;
				y1 = gg.nearhw->y;
				y2 = gg.nearhw->y;
			}
			else if (gg.nearvw != NULL)
			{
				x1 = gg.nearvw->x;
				x2 = gg.nearvw->x;
				y1 = gg.nearvw->y1;
				y2 = gg.nearvw->y2;
			}
		}
		else
		{
			gg.nearhw = NULL;
			gg.nearvw = NULL;
		}
	}
	bases->x1copy = x1;
	bases->y1copy = y1;
	bases->x2copy = x2;
	bases->y2copy = y2;
	bases->dxcopy = x1 - x2;
	bases->dycopy = y1 - y2;
	sortshints(&x1, &x2);
	sortshints(&y1, &y2);
	bases->pgnum = gg.curpage;
	g = gg.pages[gg.curpage - 1]->gbase;
	while (g != NULL)
	{
		g1 = g->next;
		if (tap)
			found = (g == gg.neargate);
		else
			found = (x1 <= g->x && g->x <= x2 && y1 <= g->y && g->y <= y2);

		if (found)
		{
			copygate(g, &g2);
			g2->next = bases->gcopy;
			bases->gcopy = g2;
			g2->x = g->x - anchorx;
			g2->y = g->y - anchory;
			if (cut)
				delgate(g);
		}
		g = g1;
	}
	s = gg.pages[gg.curpage - 1]->sbase;
	while (s != NULL)
	{
		if (tap)
			found = ((s->hwire == gg.nearhw && s->hwire != NULL) ||
					(s->vwire == gg.nearvw && s->vwire != NULL));
		else
			found = (x1 <= s->x && s->x <= x2 && y1 <= s->y && s->y <= y2);

		if (found &&
				(s->hwire == NULL || (s->hwire->x1 != s->x && s->hwire->x2 != s->x)) &&
				(s->vwire == NULL || (s->vwire->y1 != s->y && s->vwire->y2 != s->y))) {
			s2 = (log_srec *)Malloc(sizeof(log_srec));
			s2->next = bases->scopy;
			bases->scopy = s2;
			s2->x = s->x - anchorx;
			s2->y = s->y - anchory;
		}
		s = s->next;
	}
	hw = gg.pages[gg.curpage - 1]->hwbase;
	while (hw != NULL)
	{
		hw1 = hw->next;
		xx1 = hw->x1;
		xx2 = hw->x2;
		yy1 = hw->y;
		if (tap)
			found = (hw == gg.nearhw);
		else
			found = (x1 <= xx2 && xx1 <= x2 && y1 <= yy1 && yy1 <= y2);
		if (found)
		{
			hw2 = (log_hwrec *)Malloc(sizeof(log_hwrec));
			hw2->x1 = P_imax2((long)xx1, (long)x1) - anchorx;
			hw2->x2 = P_imin2((long)xx2, (long)x2) - anchorx;
			hw2->y = yy1 - anchory;
			hw2->wcolr = hw->wcolr;

			if (hw2->x1 == hw2->x2)
			{
				Free(hw2);
			}
			else
			{
				hw2->next = bases->hwcopy;
				bases->hwcopy = hw2;
				if (cut)
				{
					blbase = NULL;
					if (xx1 < x1)
						addblobs(&blbase, xx1, yy1, x1, yy1);
					if (xx2 > x2)
						addblobs(&blbase, x2, yy1, xx2, yy1);
					delhwire(hw);
					if (xx1 < x1)
						addhwire(xx1, x1, yy1, hw2->wcolr);
					if (xx2 > x2)
						addhwire(x2, xx2, yy1, hw2->wcolr);
					doblobs(blbase);
					dispblobs(&blbase);
				}
			}
		}
		hw = hw1;
	}
	vw = gg.pages[gg.curpage - 1]->vwbase;
	while (vw != NULL)
	{
		vw1 = vw->next;
		xx1 = vw->x;
		yy1 = vw->y1;
		yy2 = vw->y2;
		if (tap)
			found = (vw == gg.nearvw);
		else
			found = (y1 <= yy2 && yy1 <= y2 && x1 <= xx1 && xx1 <= x2);

		if (found)
		{
			vw2 = (log_vwrec *)Malloc(sizeof(log_vwrec));
			vw2->x = xx1 - anchorx;
			vw2->y1 = P_imax2((long)yy1, (long)y1) - anchory;
			vw2->y2 = P_imin2((long)yy2, (long)y2) - anchory;
			vw2->wcolr = vw->wcolr;
			if (vw2->y1 == vw2->y2)
			{
				Free(vw2);
			}
			else
			{
				vw2->next = bases->vwcopy;
				bases->vwcopy = vw2;
				if (cut)
				{
					blbase = NULL;
					if (yy1 < y1)
						addblobs(&blbase, xx1, y1, xx1, yy1);
					if (yy2 > y2)
						addblobs(&blbase, xx1, yy2, xx1, y2);
					delvwire(vw);
					if (yy1 < y1)
						addvwire(xx1, yy1, y1, vw2->wcolr);
					if (yy2 > y2)
						addvwire(xx1, y2, yy2, vw2->wcolr);
					doblobs(blbase);
					dispblobs(&blbase);
				}
			}
		}
		vw = vw1;
	}
	l = gg.pages[gg.curpage - 1]->lbase;
	while (l != NULL)
	{
		l1 = l->next;
		if (tap)
		{
			found = (l == gg.nearlabel);
		}
		else
		{
			found = (x1 <= l->x + m_strwidth(logfont_lfont, l->name) / gg.scale &&
					l->x <= x2 && y1 <= l->y + 2 && l->y <= y2);
		}

		if (found)
		{
			l2 = (log_lrec *)Malloc(sizeof(log_lrec));
			l2->next = bases->lcopy;
			bases->lcopy = l2;
			l2->x = l->x - anchorx;
			l2->y = l->y - anchory;
			strcpy(l2->name, l->name);
			if (cut)
			{
				eralabel(l);
				displabel(&l);
			}
		}
		l = l1;
	}
	b = gg.pages[gg.curpage - 1]->bbase;
	while (b != NULL)
	{
		b1 = b->next;
		if (tap)
			found = (b == gg.nearbox);
		else
			found = (x1 <= b->x1 && b->x2 <= x2 && y1 <= b->y1 && b->y2 <= y2);

		if (found)
		{
			b2 = (log_brec *)Malloc(sizeof(log_brec));
			b2->next = bases->bcopy;
			bases->bcopy = b2;
			b2->x1 = b->x1 - anchorx;
			b2->y1 = b->y1 - anchory;
			b2->x2 = b->x2 - anchorx;
			b2->y2 = b->y2 - anchory;
			if (cut)
			{
				drawboxc(b, gg.color.backgr);
				dispbox(&b);
			}
		}
		b = b1;
	}
	bases->valid = true;
	if (isPageEmpty(gg.pages[gg.curpage - 1]) && curfilename[gg.curpage - 1] != NULL)
		Free(curfilename[gg.curpage - 1]);
}


static void checkabort()
{
	if (!gg.stillnear || (gg.t.dn && !gg.incircuit) || *gg.func != '\0')
		_Escape(5);
}


void waitnear()
{
	do
	{
		gg.stillnear = true;
		pass();
		trykbdscroll();
		pen();
	} while (!(gg.stillnear || *gg.func != '\0'));
	checkabort();
}


static void getfirstpoint()
{
	checkabort();
	do
	{
		pass();
		trykbdscroll();
		pen();
		checkabort();
	} while (!gg.t.dn);
	gg.posx = gg.gridx;
	gg.posy = gg.gridy;
}


static int getrectangle(short *x1, short *y1, short *x2, short *y2, int first)
{
	int flag;
	long time;

	if (first)
		getfirstpoint();
	*x1 = gg.posx;
	*y1 = gg.posy;
	time = timers_sysclock();
	do
	{
		pen();
		*x2 = gg.gridx;
		*y2 = gg.gridy;
		m_colormode((long)m_xor);
		m_color((long)gg.color.select);
		flag = gg.incircuit;
		if (flag)
			rect(*x1, *y1, *x2, *y2);
		m_colormode((long)m_normal);
		do
		{
			pass();
			trykbd();
			pen();
		} while (gg.gridx == *x2 && gg.gridy == *y2 && gg.t.depressed &&
				*gg.func == '\0');
		m_colormode((long)m_xor);
		m_color((long)gg.color.select);
		if (flag)
			rect(*x1, *y1, *x2, *y2);
		m_colormode((long)m_normal);
		scroll();
		gg.stillnear = true;
		checkabort();
	} while (gg.t.depressed);
	remcursor();
	if (!gg.incircuit)
		_Escape(5);
	return (abs(*x1 - *x2) < 2 && abs(*y1 - *y2) < 2 &&
			timers_sysclock() - time < tapdelay);
}

/// Delete the object touched, or all objects in a specific area.
static void delcommand()
{
	short x1, y1, x2, y2;
	int flag;

	log_setmode("DEL");
	clearfunc();
	cursortype = delete__;
	waitnear();
	do
	{
		flag = getrectangle(&x1, &y1, &x2, &y2, true);
		cutcopy(&copybuf, x1, y1, x2, y2, true, flag);
	} while (true);
}


static void deleverything()
{
	short x1, y1, x2, y2;

	if (pagembb((int)gg.curpage, &x1, &y1, &x2, &y2))
		cutcopy(&copybuf, x1, y1, x2, y2, true, false);
}


static void changekind2(short i, short gtype, short sig, log_gattrrec *attrs)
{
	log_krec *k;

	clipoff();
	kdrawgatec(i, gg.color.backgr);
	if (kindattr[i - 1] != NULL)
	{
		k = kind[(kindgroup[i - 1] & (log_kindoffset - 1)) - 1];
		disposeattrs(&kindattr[i - 1], k->numattrs, k->attr);
		kindattr[i - 1] = NULL;
	}
	k = kind[(gtype & (log_kindoffset - 1)) - 1];
	if (!k->flag.U3.named)
		sig = 0;
	kindgroup[i - 1] = gtype;
	kindsig[i - 1] = sig;
	if (attrs != NULL)
		copyattrs(&kindattr[i - 1], attrs, k->numattrs, k->attr);
	kdrawgatec(i, gg.color.kindgate);
}

/// Drop a new gate type into one of the slots in the menu area.
void changekind(short gtype, short sig, log_gattrrec *attrs)
{
	short i;

	if (kindgroupleft <= gg.t.x && gg.t.x < kindgroupright)
	{
		i = (gg.t.x - kindgroupleft) / kindgroupspacing + 1;
		changekind2(i, gtype, sig, attrs);
	}
}


static void showstick(long x1, long y1, long x2, long y2, long mode)
{
	m_colormode(mode);
	m_color((long)gg.color.select);
	x1 = x1 * gg.scale - gg.xoff;
	y1 = y1 * gg.scale - gg.yoff;
	x2 = x2 * gg.scale - gg.xoff;
	y2 = y2 * gg.scale - gg.yoff;
	m_drawline(x1, y1, x2, y2);
	m_drawline(x1, y1, x2, y2);
	m_drawline(x2, y2, x1, y1);
	m_colormode((long)m_normal);
}


static void yardstick()
{
	long x1, y1, x2, y2;

	x1 = gg.posx;
	y1 = gg.posy;
	do
	{
		x2 = gg.gridx;
		y2 = gg.gridy;
		showstick(x1, y1, x2, y2, (long)m_xor);
		do
		{
			pass();
			trykbd();
			pen();
		} while (gg.gridx == x2 && gg.gridy == y2 && gg.t.depressed &&
				*gg.func == '\0');
		showstick(x1, y1, x2, y2, (long)m_xor);
		scroll();
	} while (gg.t.depressed && *gg.func == '\0');
	x1 -= x2;
	y1 -= y2;
	while (gg.t.near_ && *gg.func == '\0')
	{
		remcursor();
		showstick(x2 + x1, y2 + y1, x2, y2, (long)m_normal);
		do
		{
			x2 = gg.gridx;
			y2 = gg.gridy;
			showstick(x2 + x1, y2 + y1, x2, y2, (long)m_xor);
			do
			{
				pass();
				trykbd();
				pen();
			} while (gg.gridx == x2 && gg.gridy == y2 && gg.t.near_ && !gg.t.dn &&
					*gg.func == '\0');
			showstick(x2 + x1, y2 + y1, x2, y2, (long)m_xor);
			scroll();
		} while (!gg.t.dn && gg.t.near_ && *gg.func == '\0');
	}
}


static void yardstickcommand()
{
	log_setmode("YARD");
	clearfunc();
	cursortype = copy_;
	getfirstpoint();
	waitnear();
	yardstick();
	cursortype = normal;
	log_setmode("");
}


static void nextkindsig(short i)
{
	char buf[256];
	short j;
	char STR1[256];

	if (kindsig[i - 1] == 0)
		return;
	strcpy(buf, gg.signaltab[kindsig[i - 1] - 1].name);
	j = strlen(buf);
	if (!(j > 0 && isdigit(buf[j - 1])))
		return;
	while (j > 0 && buf[j - 1] == '9')
	{
		buf[j - 1] = '0';
		j--;
	}

	if (j > 0 && isdigit(buf[j - 1]))
	{
		buf[j - 1]++;
	}
	else
	{
		sprintf(STR1, "1%s", buf + j);
		strcpy(buf + j, STR1);
	}
	kdrawgatec(i, gg.color.backgr);
	kindsig[i - 1] = getsignal(0, buf);
	kdrawgatec(i, gg.color.kindgate);
}


/* Local variables for xorcopybuf: */
struct LOC_xorcopybuf {
	long *count, max, time;
	int aborting;
} ;

static int abortit(struct LOC_xorcopybuf *LINK)
{
	if (LINK->max == LONG_MAX)
	{
		if (LINK->aborting || timers_sysclock() - LINK->time <= 10)
			return LINK->aborting;
		clipoff();
		pen();
		clipon();
		m_color((long)gg.color.select);
		if (gg.posx != gg.gridx || gg.posy != gg.gridy || gg.t.dn ||
				!gg.stillnear || pollkbd2())
			LINK->aborting = true;
		return LINK->aborting;
	}
	else
	{
		return (*LINK->count >= LINK->max);
	}
}

/// XOR a picture of gates to be pasted.
static void xorcopybuf(baseptrs *bases, short x, short y, long *count_)
{
	struct LOC_xorcopybuf V;
	log_grec *g;
	log_hwrec *hw;
	log_vwrec *vw;
	log_brec *b;
	log_lrec *l;

	V.count = count_;
	V.max = *V.count;
	*V.count = 0;
	V.time = timers_sysclock();
	V.aborting = false;
	clipon();
	m_colormode((long)m_xor);
	b = bases->bcopy;

	while (b != NULL && !abortit(&V))
	{
		drawboxc0(b->x1 + x, b->y1 + y, b->x2 + x, b->y2 + y, gg.color.select);
		(*V.count)++;
		b = b->next;
	}
	m_color((long)gg.color.select);
	hw = bases->hwcopy;

	while (hw != NULL && !abortit(&V))
	{
		hline(hw->x1 + x, hw->x2 + x, hw->y + y);
		(*V.count)++;
		hw = hw->next;
	}
	vw = bases->vwcopy;

	while (vw != NULL && !abortit(&V))
	{
		vline(vw->x + x, vw->y1 + y, vw->y2 + y);
		(*V.count)++;
		vw = vw->next;
	}
	g = bases->gcopy;

	while (g != NULL && !abortit(&V))
	{
		drawgatec(g->x + x, g->y + y, g->g, gg.color.select);
		(*V.count)++;
		g = g->next;
	}
	m_color((long)gg.color.select);
	l = bases->lcopy;

	while (l != NULL && !abortit(&V))
	{
		m_drawstr((l->x + x) * gg.scale - gg.xoff,
				(l->y + y) * gg.scale - gg.yoff + 2, logfont_lfont, l->name);
		(*V.count)++;
		l = l->next;
	}
	m_colormode((long)m_normal);
	clipoff();
}


/// Paste a group of objects into the page.
static void pastebuf(baseptrs *bases, short x, short y)
{
	log_grec *g, *g1;
	log_hwrec *hw;
	log_vwrec *vw;
	log_srec *s;
	log_lrec *l, *l1;
	log_brec *b, *b1;

	clipon();
	g = bases->gcopy;
	while (g != NULL)
	{
		copygate(g, &g1);
		g1->x = g->x + x;
		g1->y = g->y + y;
		if (connectgate(g1))
		{
			drawgatex(g1);
		}
		else
		{
			frygate(g1);
			disposegate(&g1);
			clipon();
		}
		g = g->next;
	}
	clipoff();
	hw = bases->hwcopy;

	while (hw != NULL)
	{
		addhwire(hw->x1 + x, hw->x2 + x, hw->y + y, hw->wcolr);
		hw = hw->next;
	}
	vw = bases->vwcopy;

	while (vw != NULL)
	{
		addvwire(vw->x + x, vw->y1 + y, vw->y2 + y, vw->wcolr);
		vw = vw->next;
	}
	s = bases->scopy;

	while (s != NULL)
	{
		closerwire(s->x + x, s->y + y);
		if (gg.nearhw != NULL && gg.nearvw != NULL)
		{
			if (trycombinenodes(&gg.nearvw->node, &gg.nearhw->node))
			{
				clipon();
				addsolder(s->x + x, s->y + y, gg.nearhw, NULL, gg.nearvw, NULL);
				clipoff();
			}
			else
			{
				frysolder(s->x + x, s->y + y);
			}
		}
		s = s->next;
	}
	l = bases->lcopy;

	while (l != NULL)
	{
		l1 = (log_lrec *)Malloc(sizeof(log_lrec));
		l1->next = gg.pages[gg.curpage - 1]->lbase;
		gg.pages[gg.curpage - 1]->lbase = l1;
		l1->x = l->x + x;
		l1->y = l->y + y;
		strcpy(l1->name, l->name);
		l1->w = m_strwidth(logfont_lfont, l1->name) / log_scale0;
		l1->temp = (na_long)0;
		clipon();
		m_color((long)gg.color.labeltext);
		if (!gg.textinvisible)
		{
			m_drawstr(l1->x * gg.scale - gg.xoff, l1->y * gg.scale - gg.yoff + 2,
					logfont_lfont, l1->name);
		}
		clipoff();
		chpageplace(gg.pages[gg.curpage - 1], l1->x, l1->y, l1->x + l1->w, l1->y + 2);
		stamp(&gg.labelstamp);
		l = l->next;
	}
	b = bases->bcopy;

	while (b != NULL)
	{
		b1 = (log_brec *)Malloc(sizeof(log_brec));
		b1->next = gg.pages[gg.curpage - 1]->bbase;
		gg.pages[gg.curpage - 1]->bbase = b1;
		b1->x1 = b->x1 + x;
		b1->y1 = b->y1 + y;
		b1->x2 = b->x2 + x;
		b1->y2 = b->y2 + y;
		b1->temp = (na_long)0;
		clipon();
		if (!gg.textinvisible)
			drawboxc(b1, gg.color.dashbox);
		clipoff();
		chpageplace(gg.pages[gg.curpage - 1], b1->x1, b1->y1, b1->x2, b1->y2);
		stamp(&gg.boxstamp);
		b = b->next;
	}
}


static void xororiginal()
{
	remcursor();
	clipon();
	m_colormode((long)m_xor);
	m_color((long)gg.color.dashbox);
	m_linestyle(1L);
	rect(copybuf.x1copy, copybuf.y1copy, copybuf.x2copy, copybuf.y2copy);
	m_linestyle(0L);
	m_colormode((long)m_normal);
	clipoff();
}

/** @brief Enter Paste mode.

  When cursor is pressed,a dd a copy of all gates/wires in Copy buffer at the pen position.

  If the pen is held still, XOR a picture of what would be added if pen is pressed.
  */
static void pastecommand(long movemode)
{
	short thepage;
	long max;
	baseptrs deleted;
	long time;
	int smallbuf;

	clearfunc();
	if (!copybuf.valid)
		return;
	log_setmode("PASTE");
	smallbuf = (bufissmall(&copybuf) && !bufempty(&copybuf));
	if (smallbuf)
		cursortype = normal;
	else
		cursortype = paste;
	xororiginal();
	waitnear();
	xororiginal();
	do
	{
		do
		{
			if (smallbuf)
			{
				pen();
				checkabort();
				if (!gg.t.dn)
				{
					xororiginal();
					gg.posx = gg.gridx;
					gg.posy = gg.gridy;
					max = LONG_MAX;
					xorcopybuf(&copybuf, gg.posx, gg.posy, &max);
					while (gg.posx == gg.gridx && gg.posy == gg.gridy && !gg.t.dn &&
							gg.stillnear && *gg.func == '\0')
					{
						pass();
						trykbd();
						pen();
					}
					xorcopybuf(&copybuf, gg.posx, gg.posy, &max);
					xororiginal();
				}
			}
			else
			{
				xororiginal();
				time = timers_sysclock();
				do
				{
					gg.posx = gg.gridx;
					gg.posy = gg.gridy;
					pass();
					trykbd();
					pen();
					if (gg.posx != gg.gridx || gg.posy != gg.gridy)
						time = timers_sysclock();
				} while (!(timers_sysclock() - time > 25 || smallbuf || gg.t.dn) &&
						gg.stillnear && *gg.func == '\0');
				xororiginal();
				scroll();
				checkabort();
				if (gg.posx == gg.gridx && gg.posy == gg.gridy && !gg.t.dn)
				{
					xororiginal();
					max = LONG_MAX;
					xorcopybuf(&copybuf, gg.posx, gg.posy, &max);
					while (gg.posx == gg.gridx && gg.posy == gg.gridy && !gg.t.dn &&
							gg.stillnear && *gg.func == '\0')
					{
						pass();
						trykbd();
						pen();
					}
					xorcopybuf(&copybuf, gg.posx, gg.posy, &max);
					xororiginal();
				}
			}
			scroll();
			checkabort();
		} while (!gg.t.dn);
		remcursor();

		if (movemode >= 1)
		{
			if (gg.posx != copybuf.x1copy || gg.posy != copybuf.y1copy ||
					gg.curpage != copybuf.pgnum)
			{
				initbuf(&deleted);
				thepage = gg.curpage;
				if (thepage != copybuf.pgnum)
				{
					newpage(copybuf.pgnum);
					refrwindow1();
				}
				cutcopy(&deleted, copybuf.x1copy, copybuf.y1copy, copybuf.x2copy,
						copybuf.y2copy, true, false);
				if (thepage != copybuf.pgnum)
				{
					newpage(thepage);
					refrwindow1();
				}
				clearbuf(&deleted);
				pastebuf(&copybuf, gg.posx, gg.posy);
			}

			if (movemode == 2)
				movemode = 0;
		}
		else
		{
			pastebuf(&copybuf, gg.posx, gg.posy);
		}
		gg.refrflag = true;
	} while (movemode != 1);
	cursortype = normal;
	log_setmode("");
}


static void extract()
{
	int okay;
	short x1, y1, x2, y2;

	okay = pagembb((int)gg.curpage, &x1, &y1, &x2, &y2);
	if (okay)
		cutcopy(&copybuf, x1, y1, x2, y2, true, false);
	if (!strcmp(gg.funcarg, "*"))
		dumpnodes();
	if (okay)
		pastebuf(&copybuf, x2, y2);
	clearfunc();
}

/// Enter Copy mode.
static void copycommand()
{
	short x1, y1, x2, y2;
	int flag;

	log_setmode("COPY");
	clearfunc();
	cursortype = copy_;
	waitnear();
	do
	{
		flag = getrectangle(&x1, &y1, &x2, &y2, true);
		cutcopy(&copybuf, x1, y1, x2, y2, false, flag);
		if (!bufempty(&copybuf))
			pastecommand(0L);
	} while (true);
}

/** @brief Enter Move mode.

  Sweep out an area, then move all objects in that area to a new position.
  */
static void movecommand(int waitflag)
{
	short x1, y1, x2, y2;
	int flag;

	log_setmode("MOVE");
	clearfunc();
	cursortype = copy_;
	if (waitflag)
		waitnear();
	do
	{
		flag = getrectangle(&x1, &y1, &x2, &y2, waitflag);
		cutcopy(&copybuf, x1, y1, x2, y2, false, flag);
		if (!bufempty(&copybuf))
		{
			if (waitflag)
				pastecommand(1L);
			else
				pastecommand(2L);
		}
	} while (waitflag);
	_Escape(5);
}

static int near_(short x, short y)
{
	return (abs(x - y) < 2);
}

/// Move a gate, wire, label, or dashed box.
static void moveobject()
{
	short gtype, x1, y1, yy, hx1, hx2, hy, vx, vy1, vy2, hc, vc, oldpg, newpg,
		  oldx1, oldy1, oldx2, oldy2;
	int fh1, fv1, fh2, fv2, oo;
	log_srec *s;
	log_hwrec *hw1;
	log_vwrec *vw1;
	blobrec *blbase, *blp;

	closergate(gg.posx, gg.posy);
	oldpg = gg.curpage;
	gg.gridx = gg.posx;
	gg.gridy = gg.posy;
	if (gg.markers && ((gg.gridx == gg.markerx1 && gg.gridy == gg.markery1) ||
				(gg.gridx == gg.markerx2 && gg.gridy == gg.markery2)))
	{
		remcursor();
		drawmarkersc(gg.color.backgr);
		gg.markers = false;
		stamp(&gg.markerstamp);
		do
		{
			gg.posx = gg.gridx;
			gg.posy = gg.gridy;
			m_colormode((long)m_xor);
			drawmarkersc(gg.color.marker);
			m_colormode((long)m_normal);
			do
			{
				pass();
				trykbd();
				pen();
			} while (gg.gridx == gg.posx && gg.gridy == gg.posy && gg.t.depressed &&
					strcmp(gg.func, "REFR"));

			m_colormode((long)m_xor);
			drawmarkersc(gg.color.marker);
			m_colormode((long)m_normal);
			if (gg.posx == gg.markerx1 && gg.posy == gg.markery1 &&
					gg.gridx < gg.markerx2 && gg.gridy < gg.markery2 && gg.incircuit)
			{
				gg.markerx1 = gg.gridx;
				gg.markery1 = gg.gridy;
			}
			else if (gg.posx == gg.markerx2 && gg.posy == gg.markery2 &&
					gg.gridx > gg.markerx1 && gg.gridy > gg.markery1 &&
					gg.incircuit)
			{
				gg.markerx2 = gg.gridx;
				gg.markery2 = gg.gridy;
			}
			scroll();
		} while (gg.t.depressed);
		stamp(&gg.markerstamp);
		gg.markers = true;
		remcursor();
		drawmarkersc(gg.color.marker);
		refreshsoon();
		gg.startpoint = false;
		return;
	}

	if (gg.nearlabel != NULL)
	{
		remcursor();
		oldx1 = gg.nearlabel->x;
		oldy1 = gg.nearlabel->y;
		oldx2 = oldx1 + gg.nearlabel->w;
		oldy2 = oldy1 + 2;
		chpageplace(gg.pages[gg.curpage - 1], oldx1, oldy1, oldx2, oldy2);
		stamp(&gg.labelstamp);
		clipon();
		eralabel(gg.nearlabel);
		clipoff();
		x1 = gg.nearlabel->x - gg.posx;
		y1 = gg.nearlabel->y - gg.posy;
		do
		{
			gg.posx = gg.gridx;
			gg.posy = gg.gridy;
			oo = gg.t.off;
			clipon();
			xorlabel(gg.posx + x1, gg.posy + y1, gg.nearlabel);
			clipoff();
			do
			{
				pass();
				trykbd();
				pen();
			} while (gg.gridx == gg.posx && gg.gridy == gg.posy && gg.t.depressed &&
					strcmp(gg.func, "REFR"));
			clipon();
			xorlabel(gg.posx + x1, gg.posy + y1, gg.nearlabel);
			clipoff();
			if (!strcmp(gg.func, "REFR"))
			{
				scroll();
				clipon();
				eralabel(gg.nearlabel);
				clipoff();
			}
		} while (gg.t.depressed);
		remcursor();
		chpageplace(gg.pages[oldpg - 1], oldx1, oldy1, oldx2, oldy2);
		if (gg.incircuit && !oo && !gg.t.off)
		{
			gg.posx += x1;
			gg.posy += y1;
			gg.nearlabel->x = gg.posx;
			gg.nearlabel->y = gg.posy;
			if (gg.curpage != oldpg)
			{
				newpg = gg.curpage;
				gg.curpage = oldpg;
				unlinklabel(gg.nearlabel);
				gg.curpage = newpg;
				linklabel(gg.nearlabel);
			}
			m_color((long)gg.color.labeltext);
			m_drawstr(gg.posx * gg.scale - gg.xoff,
					gg.posy * gg.scale - gg.yoff + 2, logfont_lfont,
					gg.nearlabel->name);
			gg.refrflag = true;
			chpageplace(gg.pages[gg.curpage - 1], gg.nearlabel->x, gg.nearlabel->y,
					gg.nearlabel->x + gg.nearlabel->w, gg.nearlabel->y + 2);
		}
		else
		{
			newpg = gg.curpage;
			gg.curpage = oldpg;
			displabel(&gg.nearlabel);
			gg.curpage = newpg;
		}
		stamp(&gg.labelstamp);
		refreshsoon();
		gg.startpoint = false;
		return;
	}

	if (gg.nearbox != NULL)
	{
		remcursor();
		oldx1 = gg.nearbox->x1;
		oldy1 = gg.nearbox->y1;
		oldx2 = gg.nearbox->x2;
		oldy2 = gg.nearbox->y2;
		chpageplace(gg.pages[gg.curpage - 1], oldx1, oldy1, oldx2, oldy2);
		stamp(&gg.boxstamp);
		clipon();
		drawboxc(gg.nearbox, gg.color.backgr);
		clipoff();
		hx1 = gg.nearbox->x1;
		hx2 = gg.nearbox->x2;
		vy1 = gg.nearbox->y1;
		vy2 = gg.nearbox->y2;
		do
		{
			gg.posx = gg.gridx;
			gg.posy = gg.gridy;
			oo = gg.t.off;
			clipon();
			m_colormode((long)m_xor);
			drawboxc0(hx1, vy1, hx2, vy2, gg.color.dashbox);
			m_colormode((long)m_normal);
			clipoff();
			do
			{
				pass();
				trykbd();
				pen();
			} while (gg.gridx == gg.posx && gg.gridy == gg.posy && gg.t.depressed &&
					strcmp(gg.func, "REFR"));

			clipon();
			m_colormode((long)m_xor);
			drawboxc0(hx1, vy1, hx2, vy2, gg.color.dashbox);
			m_colormode((long)m_normal);
			clipoff();
			if (near_(gg.posx, hx1) && near_(gg.posy, vy1))
			{
				hx1 = gg.gridx;
				vy1 = gg.gridy;
			}
			else if (near_(gg.posx, hx1) && near_(gg.posy, vy2))
			{
				hx1 = gg.gridx;
				vy2 = gg.gridy;
			}
			else if (near_(gg.posx, hx2) && near_(gg.posy, vy1))
			{
				hx2 = gg.gridx;
				vy1 = gg.gridy;
			}
			else if (near_(gg.posx, hx2) && near_(gg.posy, vy2))
			{
				hx2 = gg.gridx;
				vy2 = gg.gridy;
			}
			else
			{
				hx1 += gg.gridx - gg.posx;
				hx2 += gg.gridx - gg.posx;
				vy1 += gg.gridy - gg.posy;
				vy2 += gg.gridy - gg.posy;
			}

			if (!strcmp(gg.func, "REFR"))
			{
				scroll();
				clipon();
				drawboxc(gg.nearbox, gg.color.backgr);
				clipoff();
			}
		} while (gg.t.depressed);
		remcursor();
		sortshints(&hx1, &hx2);
		sortshints(&vy1, &vy2);
		gg.nearbox->x1 = hx1;
		gg.nearbox->x2 = hx2;
		gg.nearbox->y1 = vy1;
		gg.nearbox->y2 = vy2;
		chpageplace(gg.pages[oldpg - 1], oldx1, oldy1, oldx2, oldy2);
		if (gg.incircuit && !oo && !gg.t.off)
		{
			if (gg.curpage != oldpg)
			{
				newpg = gg.curpage;
				gg.curpage = oldpg;
				unlinkbox(gg.nearbox);
				gg.curpage = newpg;
				linkbox(gg.nearbox);
			}
			clipon();
			drawboxc(gg.nearbox, gg.color.dashbox);
			clipoff();
			chpageplace(gg.pages[gg.curpage - 1], gg.nearbox->x1, gg.nearbox->y1,
					gg.nearbox->x2, gg.nearbox->y2);
			gg.refrflag = true;
		}
		else
		{
			newpg = gg.curpage;
			gg.curpage = oldpg;
			dispbox(&gg.nearbox);
			gg.curpage = newpg;
		}
		stamp(&gg.boxstamp);
		refreshsoon();
		gg.startpoint = false;
		return;
	}

	if (gg.neargate != NULL)
	{
		gtype = gg.neargate->g;
		remcursor();
		clipon();
		eragate(gg.neargate);
		clipoff();
		disconnectgate(gg.neargate);
		x1 = gg.neargate->x - gg.posx;
		y1 = gg.neargate->y - gg.posy;
		do
		{
			yy = gg.t.y;
			gg.posx = gg.gridx;
			gg.posy = gg.gridy;
			oo = gg.t.off;
			movexorgate(gg.posx + x1, gg.posy + y1, gtype, gg.neargate->sig, yy);
			do
			{
				pass();
				trykbd();
				pen();
			} while (gg.gridx == gg.posx && gg.gridy == gg.posy && gg.t.depressed &&
					strcmp(gg.func, "REFR"));
			movexorgate(gg.posx + x1, gg.posy + y1, gtype, gg.neargate->sig, yy);
			scroll();
		} while (gg.t.depressed);
		remcursor();

		if (gg.incircuit && !oo && !gg.t.off)
		{
			gg.posx += x1;
			gg.posy += y1;
			gg.neargate->x = gg.posx;
			gg.neargate->y = gg.posy;
			clipon();
			drawgatex(gg.neargate);
			clipoff();
			if (!connectgate(gg.neargate))
			{
				frygate(gg.neargate);
				disposegate(&gg.neargate);
			}
			gg.refrflag = true;
		}
		else if (gg.t.y >= baseline)
		{
			changekind(gg.neargate->g, gg.neargate->sig, gg.neargate->attr);
			disposegate(&gg.neargate);
		}
		else
		{
			disposegate(&gg.neargate);
		}
		refreshsoon();
		gg.startpoint = false;
		return;
	}
	closerwire(gg.posx, gg.posy);
	if (gg.nearhw == NULL && gg.nearvw == NULL)
	{
		gg.startpoint = false;
		assertfunc("MOVE *");
		return;
	}
	remcursor();
	clipon();
	m_color((long)gg.color.backgr);
	if (gg.nearhw != NULL)
	{
		hx1 = gg.nearhw->x1;
		hx2 = gg.nearhw->x2;
		hy = gg.nearhw->y;
		fh1 = (gg.posx == hx1 && gg.posy == hy);
		fh2 = (gg.posx == hx2 && gg.posy == hy);
		hline(hx1, hx2, hy);
		gg.movinghw = gg.nearhw;
		hc = gg.nearhw->wcolr;
	}

	if (gg.nearvw != NULL)
	{
		vx = gg.nearvw->x;
		vy1 = gg.nearvw->y1;
		vy2 = gg.nearvw->y2;
		fv1 = (gg.posy == vy1 && gg.posx == vx);
		fv2 = (gg.posy == vy2 && gg.posx == vx);
		vline(vx, vy1, vy2);
		gg.movingvw = gg.nearvw;
		vc = gg.nearvw->wcolr;
	}
	blbase = NULL;
	s = gg.pages[gg.curpage - 1]->sbase;
	while (s != NULL)
	{
		if ((gg.nearhw != NULL && (s->hwire == gg.nearhw || s->hwire2 == gg.nearhw)) ||
				(gg.nearvw != NULL && (s->vwire == gg.nearvw || s->vwire2 == gg.nearvw)))
		{
			drawsolderc(s->x, s->y, gg.color.backgr);
			blp = (blobrec *)Malloc(sizeof(blobrec));
			if (s->hwire == gg.nearhw)
				blp->hw = NULL;
			else
				blp->hw = s->hwire;

			if (s->vwire == gg.nearvw)
				blp->vw = NULL;
			else
				blp->vw = s->vwire;

			blp->next = blbase;
			blbase = blp;
		}
		s = s->next;
	}
	clipoff();
	do
	{
		gg.posx = gg.gridx;
		gg.posy = gg.gridy;
		oo = gg.t.off;
		clipon();
		m_colormode((long)m_xor);
		if (gg.nearhw != NULL)
		{
			if (fh1)
			{
				hx1 = gg.posx;
				hy = gg.posy;
			} else if (fh2)
			{
				hx2 = gg.posx;
				hy = gg.posy;
			}
			else
			{
				if (gg.posx < hx1)
					hx1 = gg.posx;
				if (gg.posx > hx2)
					hx2 = gg.posx;
				hy = gg.posy;
			}
			m_color((long)gg.color.wire[gg.nearhw->wcolr - log_wcol_normal]);
			hline(hx1, hx2, hy);
		}
		if (gg.nearvw != NULL)
		{
			if (fv1)
			{
				vy1 = gg.posy;
				vx = gg.posx;
			}
			else if (fv2)
			{
				vy2 = gg.posy;
				vx = gg.posx;
			}
			else
			{
				if (gg.posy < vy1)
					vy1 = gg.posy;
				if (gg.posy > vy2)
					vy2 = gg.posy;
				vx = gg.posx;
			}
			m_color((long)gg.color.wire[gg.nearvw->wcolr - log_wcol_normal]);
			vline(vx, vy1, vy2);
		}
		m_colormode((long)m_normal);
		clipoff();
		do
		{
			pass();
			trykbd();
			pen();
		} while (gg.gridx == gg.posx && gg.gridy == gg.posy && gg.t.depressed &&
				strcmp(gg.func, "REFR"));
		clipon();
		m_colormode((long)m_xor);
		if (gg.nearhw != NULL)
		{
			m_color((long)gg.color.wire[gg.nearhw->wcolr - log_wcol_normal]);
			hline(hx1, hx2, hy);
		}

		if (gg.nearvw != NULL)
		{
			m_color((long)gg.color.wire[gg.nearvw->wcolr - log_wcol_normal]);
			vline(vx, vy1, vy2);
		}

		m_colormode((long)m_normal);
		clipoff();
		scroll();
	} while (gg.t.depressed);
	gg.movinghw = NULL;
	gg.movingvw = NULL;
	remcursor();
	gg.curpage = oldpg;
	if (gg.nearhw != NULL)
	{
		delhwire(gg.nearhw);
	}
	else
	{
		hx1 = 0;
		hx2 = 0;
	}

	if (gg.nearvw != NULL)
	{
		delvwire(gg.nearvw);
	}
	else
	{
		vy1 = 0;
		vy2 = 0;
	}

	gg.curpage = realcurpage;
	if (gg.incircuit && !oo && !gg.t.off)
	{
		blp = blbase;
		while (blp != NULL)
		{
			if (blp->hw != NULL)
			{
				hw1 = gg.pages[gg.curpage - 1]->hwbase;
				while (hw1 != NULL &&
						(hw1->node != blp->hw->node ||
						 !P_ibetween((long)hw1->x1, (long)vx, (long)hw1->x2) ||
						 !P_ibetween((long)vy1, (long)hw1->y, (long)vy2)))
					hw1 = hw1->next;

				if (hw1 != NULL)
				{
					blp->x = vx;
					blp->y = hw1->y;
				}
				else
				{
					blp->x = -32768L;
				}
			}
			else if (blp->vw != NULL)
			{
				vw1 = gg.pages[gg.curpage - 1]->vwbase;
				while (vw1 != NULL &&
						(vw1->node != blp->vw->node ||
						 !P_ibetween((long)vw1->y1, (long)hy, (long)vw1->y2) ||
						 !P_ibetween((long)hx1, (long)vw1->x, (long)hx2)))
					vw1 = vw1->next;

				if (vw1 != NULL)
				{
					blp->x = vw1->x;
					blp->y = hy;
				}
				else
				{
					blp->x = -32768L;
				}
			}
			else
			{
				blp->x = vx;
				blp->y = hy;
			}
			blp = blp->next;
		}

		if (hx1 != hx2)
			addhwire(hx1, hx2, hy, hc);
		if (vy1 != vy2)
			addvwire(vx, vy1, vy2, vc);
		doblobs(blbase);
	}
	dispblobs(&blbase);
	refreshsoon();
	gg.startpoint = false;

	/*thrown away*/
	/* move a wire */
}


static void shiftgate(log_grec *g, short dx, short dy)
{
	short i, FORLIM;
	log_pinposrec *WITH;

	g->x += dx;
	g->y += dy;
	FORLIM = g->kind->numpins;
	for (i = 0; i < FORLIM; i++)
	{
		WITH = &g->pinpos[i];
		WITH->x += dx;
		WITH->y += dy;
	}
}


static void xorrect(short x1, short y1, short x2, short y2)
{
	short y3;

	sortshints(&x1, &x2);
	sortshints(&y1, &y2);
	y3 = (y1 + y2) / 2;
	m_colormode((long)m_xor);
	m_color((long)gg.color.select);
	rect(x1, y1, x2, y2);
	hline(x1, x2, y3);
	line(x2 - 1, y3 - 1, x2, y3);
	line(x2 - 1, y3 + 1, x2, y3);
	m_colormode((long)m_normal);
}

/// Open space horizontally.
/**
  This means "put everything aside horizontally."
  */
static void openhoriz()
{
	log_grec *g;
	log_hwrec *hw;
	log_vwrec *vw;
	log_srec *s;
	log_lrec *l;
	log_brec *b;
	short x1, y1;
	int flag;

	log_setmode("OPNH");
	clearfunc();
	cursortype = copy_;
	waitnear();
	do
	{
		do
		{
			gg.posx = gg.gridx;
			gg.posy = gg.gridy;
			pass();
			trykbdscroll();
			pen();
		} while (!gg.t.dn && gg.stillnear && *gg.func == '\0');

		if (gg.incircuit && gg.stillnear && *gg.func == '\0')
		{
			x1 = gg.posx;
			y1 = gg.posy;
			do
			{
				pen();
				x1 = gg.gridx;
				y1 = gg.gridy;
				xorrect(gg.posx, gg.posy, x1, y1);
				do
				{
					pass();
					trykbd();
					pen();
				} while (gg.gridx == x1 && gg.gridy == y1 && gg.t.depressed &&
						*gg.func == '\0');
				xorrect(gg.posx, gg.posy, x1, y1);
				scroll();
			} while ((gg.t.depressed || x1 == gg.posx || y1 == gg.posy) &&
					*gg.func == '\0');
			sortshints(&gg.posx, &x1);
			sortshints(&gg.posy, &y1);
			remcursor();

			if (gg.incircuit && *gg.func == '\0')
			{
				chpage(gg.pages[gg.curpage - 1]);
				vw = gg.pages[gg.curpage - 1]->vwbase;
				flag = false;
				while (vw != NULL)
				{
					if (vw->x > gg.posx &&
							((vw->y1 < gg.posy && gg.posy < vw->y2) ||
							 (vw->y1 < y1 && y1 < vw->y2)))
						flag = true;
					vw = vw->next;
				}

				if (!flag)
				{
					g = gg.pages[gg.curpage - 1]->gbase;
					while (g != NULL)
					{
						if (gg.posy <= g->y && g->y <= y1 && g->x >= gg.posx)
							shiftgate(g, x1 - gg.posx, 0);
						g = g->next;
					}
					s = gg.pages[gg.curpage - 1]->sbase;

					while (s != NULL)
					{
						if (!vlsi && gg.posy <= s->vwire->y2 && s->vwire->y1 <= y1 &&
								s->vwire->x > gg.posx)
							s->x += x1 - gg.posx;
						s = s->next;
					}
					hw = gg.pages[gg.curpage - 1]->hwbase;

					while (hw != NULL)
					{
						if (gg.posy <= hw->y && hw->y <= y1)
						{
							if (hw->x1 > gg.posx)
								hw->x1 += x1 - gg.posx;
							if (hw->x2 >= gg.posx)
								hw->x2 += x1 - gg.posx;
						}
						hw = hw->next;
					}
					vw = gg.pages[gg.curpage - 1]->vwbase;

					while (vw != NULL)
					{
						if (gg.posy <= vw->y2 && vw->y1 <= y1 && vw->x > gg.posx)
							vw->x += x1 - gg.posx;
						vw = vw->next;
					}
					b = gg.pages[gg.curpage - 1]->bbase;

					while (b != NULL)
					{
						if (gg.posy <= b->y1 && b->y2 <= y1)
						{
							if (b->x1 >= gg.posx)
								b->x1 += x1 - gg.posx;
							if (b->x2 >= gg.posx)
								b->x2 += x1 - gg.posx;
						}
						b = b->next;
					}
					l = gg.pages[gg.curpage - 1]->lbase;

					while (l != NULL)
					{
						if (gg.posy <= l->y && l->y <= y1 && l->x >= gg.posx)
							l->x += x1 - gg.posx;
						l = l->next;
					}
					refrfunc();
					scroll();
				}
				else
				{
					beginerror();
					nc_printf("Can't Open: Blocked by vertical wires\n");
					enderror();
				}
			}
		}
	} while (gg.incircuit && gg.stillnear && *gg.func == '\0');
	gg.startpoint = false;
	log_setmode("");
	if (gg.t.dn && !gg.incircuit)
	{
		gg.t.dn = false;
		gg.t.depressed = false;
	}
	cursortype = normal;
}


static void xorrect_(short x1, short y1, short x2, short y2)
{
	short x3;

	sortshints(&x1, &x2);
	sortshints(&y1, &y2);
	x3 = (x1 + x2) / 2;
	m_colormode((long)m_xor);
	m_color((long)gg.color.select);
	rect(x1, y1, x2, y2);
	vline(x3, y1, y2);
	line(x3 - 1, y2 - 1, x3, y2);
	line(x3 + 1, y2 - 1, x3, y2);
	m_colormode((long)m_normal);
}

/// Open space vertically.
/**
  This means "put everything aside vertically."
  */
static void openvert()
{
	log_grec *g;
	log_hwrec *hw;
	log_vwrec *vw;
	log_srec *s;
	log_lrec *l;
	log_brec *b;
	short x1, y1;
	int flag;

	log_setmode("OPNV");
	clearfunc();
	cursortype = copy_;
	waitnear();
	do
	{
		do
		{
			gg.posx = gg.gridx;
			gg.posy = gg.gridy;
			pass();
			trykbdscroll();
			pen();
		} while (!gg.t.dn && gg.stillnear && *gg.func == '\0');

		if (gg.incircuit && gg.stillnear && *gg.func == '\0')
		{
			x1 = gg.posx;
			y1 = gg.posy;
			do
			{
				x1 = gg.gridx;
				y1 = gg.gridy;
				xorrect_(gg.posx, gg.posy, x1, y1);
				do
				{
					pass();
					trykbd();
					pen();
				} while (gg.gridx == x1 && gg.gridy == y1 && gg.t.depressed &&
						*gg.func == '\0');
				xorrect_(gg.posx, gg.posy, x1, y1);
				scroll();
			} while ((gg.t.depressed || x1 == gg.posx || y1 == gg.posy) &&
					*gg.func == '\0');
			sortshints(&gg.posx, &x1);
			sortshints(&gg.posy, &y1);
			remcursor();
			if (gg.incircuit && *gg.func == '\0')
			{
				chpage(gg.pages[gg.curpage - 1]);
				hw = gg.pages[gg.curpage - 1]->hwbase;
				flag = false;

				while (hw != NULL)
				{
					if (hw->y > gg.posy &&
							((hw->x1 < gg.posx && gg.posx < hw->x2) ||
							 (hw->x1 < x1 && x1 < hw->x2)))
						flag = true;
					hw = hw->next;
				}

				if (!flag)
				{
					g = gg.pages[gg.curpage - 1]->gbase;
					while (g != NULL)
					{
						if (gg.posx <= g->x && g->x <= x1 && g->y >= gg.posy)
							shiftgate(g, 0, y1 - gg.posy);
						g = g->next;
					}
					s = gg.pages[gg.curpage - 1]->sbase;

					while (s != NULL)
					{
						if (!vlsi && gg.posx <= s->hwire->x2 && s->hwire->x1 <= x1 &&
								s->hwire->y > gg.posy)
							s->y += y1 - gg.posy;
						s = s->next;
					}
					vw = gg.pages[gg.curpage - 1]->vwbase;

					while (vw != NULL)
					{
						if (gg.posx <= vw->x && vw->x <= x1)
						{
							if (vw->y1 > gg.posy)
								vw->y1 += y1 - gg.posy;
							if (vw->y2 >= gg.posy)
								vw->y2 += y1 - gg.posy;
						}
						vw = vw->next;
					}
					hw = gg.pages[gg.curpage - 1]->hwbase;

					while (hw != NULL)
					{
						if (gg.posx <= hw->x2 && hw->x1 <= x1 && hw->y > gg.posy)
							hw->y += y1 - gg.posy;
						hw = hw->next;
					}

					b = gg.pages[gg.curpage - 1]->bbase;
					while (b != NULL)
					{
						if (gg.posx <= b->x1 && b->x2 <= x1)
						{
							if (b->y1 >= gg.posy)
								b->y1 += y1 - gg.posy;
							if (b->y2 >= gg.posy)
								b->y2 += y1 - gg.posy;
						}
						b = b->next;
					}
					l = gg.pages[gg.curpage - 1]->lbase;
					while (l != NULL)
					{
						if (gg.posx <= l->x && l->x <= x1 && l->y >= gg.posy)
							l->y += y1 - gg.posy;
						l = l->next;
					}
					refrfunc();
					scroll();
				}
				else
				{
					beginerror();
					nc_printf("Can't Open: Blocked by horizontal wires\n");
					enderror();
				}
			}
		}
	} while (gg.incircuit && gg.stillnear && *gg.func == '\0');
	gg.startpoint = false;
	log_setmode("");
	if (gg.t.dn && !gg.incircuit)
	{
		gg.t.dn = false;
		gg.t.depressed = false;
	}
	cursortype = normal;
}

static void xorrect__(short x1, short y1, short x2, short y2)
{
	short y3;

	sortshints(&x1, &x2);
	sortshints(&y1, &y2);
	y3 = (y1 + y2) / 2;
	m_colormode((long)m_xor);
	m_color((long)gg.color.select);
	rect(x1, y1, x2, y2);
	hline(x1, x2, y3);
	line(x1 + 1, y3 - 1, x1, y3);
	line(x1 + 1, y3 + 1, x1, y3);
	m_colormode((long)m_normal);
}

/// Close space horizontally.
static void closehoriz()
{
	log_grec *g, *g1;
	log_hwrec *hw, *hw1;
	log_vwrec *vw, *vw1;
	log_srec *s, *s1;
	log_lrec *l, *l1;
	log_brec *b, *b1;
	short x1, y1;
	int flag;

	log_setmode("CLSH");
	clearfunc();
	cursortype = copy_;
	waitnear();
	do
	{
		do
		{
			gg.posx = gg.gridx;
			gg.posy = gg.gridy;
			pass();
			trykbdscroll();
			pen();
		} while (!gg.t.dn && gg.stillnear && *gg.func == '\0');

		if (gg.incircuit && gg.stillnear && *gg.func == '\0')
		{
			x1 = gg.posx;
			y1 = gg.posy;
			do
			{
				x1 = gg.gridx;
				y1 = gg.gridy;
				xorrect__(gg.posx, gg.posy, x1, y1);
				do
				{
					pass();
					trykbd();
					pen();
				} while (gg.gridx == x1 && gg.gridy == y1 && gg.t.depressed &&
						*gg.func == '\0');
				xorrect__(gg.posx, gg.posy, x1, y1);
				scroll();
			} while ((gg.t.depressed || x1 == gg.posx || y1 == gg.posy) &&
					*gg.func == '\0');
			sortshints(&gg.posx, &x1);
			sortshints(&gg.posy, &y1);
			remcursor();
			if (gg.incircuit && *gg.func == '\0')
			{
				chpage(gg.pages[gg.curpage - 1]);
				vw = gg.pages[gg.curpage - 1]->vwbase;
				flag = false;
				while (vw != NULL)
				{
					if (vw->x >= gg.posx &&
							((vw->y1 < gg.posy && gg.posy < vw->y2) ||
							 (vw->y1 < y1 && y1 < vw->y2)))
						flag = true;
					vw = vw->next;
				}

				if (!flag)
				{
					g = gg.pages[gg.curpage - 1]->gbase;
					while (g != NULL)
					{
						g1 = g->next;
						if (gg.posy <= g->y && g->y <= y1 && g->x >= gg.posx)
						{
							if (g->x <= x1)
								delgate(g);
							else
								shiftgate(g, gg.posx - x1, 0);
						}
						g = g1;
					}
					s = gg.pages[gg.curpage - 1]->sbase;

					while (s != NULL)
					{
						s1 = s->next;
						if (!vlsi && gg.posy <= s->vwire->y2 && s->vwire->y1 <= y1 &&
								s->vwire->x >= gg.posx)
						{
							if (s->vwire->x > x1)
								s->x += gg.posx - x1;
						}
						s = s1;
					}
					hw = gg.pages[gg.curpage - 1]->hwbase;

					while (hw != NULL)
					{
						hw1 = hw->next;
						if (gg.posy <= hw->y && hw->y <= y1 && hw->x2 >= gg.posx) {
							if (hw->x1 >= gg.posx && hw->x2 < x1)
								delhwire(hw);
						}
						hw = hw1;
					}
					vw = gg.pages[gg.curpage - 1]->vwbase;

					while (vw != NULL)
					{
						vw1 = vw->next;
						if (gg.posy <= vw->y2 && vw->y1 <= y1 && vw->x >= gg.posx &&
								vw->x <= x1)
							delvwire(vw);
						vw = vw1;
					}
					hw = gg.pages[gg.curpage - 1]->hwbase;

					while (hw != NULL)
					{
						if (gg.posy <= hw->y && hw->y <= y1 && hw->x2 >= gg.posx)
						{
							if (hw->x1 >= gg.posx)
							{
								if (hw->x1 < x1)
								{
									hw->x1 = gg.posx;
								}
								else
								{
									hw->x1 += gg.posx - x1;
									hw->x2 += gg.posx - x1;
								}
							}
							else if (hw->x2 >= x1)
							{
								hw->x2 += gg.posx - x1;
							}
							else if (hw->x2 >= gg.posx)
							{
								hw->x2 = gg.posx;
							}
						}
						hw = hw->next;
					}
					vw = gg.pages[gg.curpage - 1]->vwbase;

					while (vw != NULL)
					{
						if (gg.posy <= vw->y2 && vw->y1 <= y1 && vw->x >= gg.posx)
							vw->x += gg.posx - x1;
						vw = vw->next;
					}
					b = gg.pages[gg.curpage - 1]->bbase;

					while (b != NULL)
					{
						b1 = b->next;
						if (gg.posy <= b->y1 && b->y2 <= y1)
						{
							if (b->x1 >= x1)
							{
								b->x1 += gg.posx - x1;
								b->x2 += gg.posx - x1;
							}
							else if (b->x1 >= gg.posx)
							{
								if (b->x2 < x1)
								{
									dispbox(&b);
								}
								else
								{
									b->x1 = gg.posx;
									b->x2 += gg.posx - x1;
								}
							}
							else if (b->x2 >= x1)
							{
								b->x2 += gg.posx - x1;
							}
							else if (b->x2 >= gg.posx)
							{
								b->x2 = gg.posx;
							}
						}
						b = b1;
					}
					l = gg.pages[gg.curpage - 1]->lbase;

					while (l != NULL)
					{
						l1 = l->next;
						if (gg.posy <= l->y && l->y <= y1 && l->x >= gg.posx)
						{
							if (l->x + strlen(gg.nearlabel->name) * 7 / gg.scale <= x1)
								displabel(&l);
							else
								l->x += gg.posx - x1;
						}
						l = l1;
					}
					refrfunc();
					scroll();
				}
				else
				{
					beginerror();
					nc_printf("Can't Close: Blocked by vertical wires\n");
					enderror();
				}
			}
		}
	} while (gg.incircuit && gg.stillnear && *gg.func == '\0');
	gg.startpoint = false;
	log_setmode("");
	if (gg.t.dn && !gg.incircuit)
	{
		gg.t.dn = false;
		gg.t.depressed = false;
	}
	cursortype = normal;
}


static void xorrect___(short x1, short y1, short x2, short y2)
{
	short x3;

	sortshints(&x1, &x2);
	sortshints(&y1, &y2);
	x3 = (x1 + x2) / 2;
	m_colormode((long)m_xor);
	m_color((long)gg.color.select);
	rect(x1, y1, x2, y2);
	vline(x3, y1, y2);
	line(x3 - 1, y1 + 1, x3, y1);
	line(x3 + 1, y1 + 1, x3, y1);
	m_colormode((long)m_normal);
}

/// Close space vertically.
static void closevert()
{
	log_grec *g, *g1;
	log_hwrec *hw, *hw1;
	log_vwrec *vw, *vw1;
	log_srec *s, *s1;
	log_lrec *l, *l1;
	log_brec *b, *b1;
	short x1, y1;
	int flag;

	log_setmode("CLSV");
	clearfunc();
	cursortype = copy_;
	waitnear();
	do
	{
		do
		{
			gg.posx = gg.gridx;
			gg.posy = gg.gridy;
			pass();
			trykbdscroll();
			pen();
		} while (!gg.t.dn && gg.stillnear && *gg.func == '\0');

		if (gg.incircuit && gg.stillnear && *gg.func == '\0')
		{
			x1 = gg.posx;
			y1 = gg.posy;
			do
			{
				x1 = gg.gridx;
				y1 = gg.gridy;
				xorrect___(gg.posx, gg.posy, x1, y1);
				do
				{
					pass();
					trykbd();
					pen();
				} while (gg.gridx == x1 && gg.gridy == y1 && gg.t.depressed &&
						*gg.func == '\0');
				xorrect___(gg.posx, gg.posy, x1, y1);
				scroll();
			} while ((gg.t.depressed || x1 == gg.posx || y1 == gg.posy) &&
					*gg.func == '\0');
			sortshints(&gg.posx, &x1);
			sortshints(&gg.posy, &y1);
			remcursor();
			if (gg.incircuit && *gg.func == '\0')
			{
				chpage(gg.pages[gg.curpage - 1]);
				hw = gg.pages[gg.curpage - 1]->hwbase;
				flag = false;
				while (hw != NULL)
				{
					if (hw->y >= gg.posy &&
							((hw->x1 < gg.posx && gg.posx < hw->x2) ||
							 (hw->x1 < x1 && x1 < hw->x2)))
						flag = true;
					hw = hw->next;
				}

				if (!flag)
				{
					g = gg.pages[gg.curpage - 1]->gbase;
					while (g != NULL)
					{
						g1 = g->next;
						if (gg.posx <= g->x && g->x <= x1 && g->y >= gg.posy)
						{
							if (g->y <= y1)
								delgate(g);
							else
								shiftgate(g, 0, gg.posy - y1);
						}
						g = g1;
					}
					s = gg.pages[gg.curpage - 1]->sbase;

					while (s != NULL)
					{
						s1 = s->next;
						if (!vlsi && gg.posx <= s->hwire->x2 && s->hwire->x1 <= x1 &&
								s->hwire->y >= gg.posy)
						{
							if (s->hwire->y > y1)
								s->y += gg.posy - y1;
						}
						s = s1;
					}
					vw = gg.pages[gg.curpage - 1]->vwbase;

					while (vw != NULL)
					{
						vw1 = vw->next;
						if (gg.posx <= vw->x && vw->x <= x1 && vw->y2 >= gg.posy)
						{
							if (vw->y1 >= gg.posy && vw->y2 < y1)
								delvwire(vw);
						}
						vw = vw1;
					}
					hw = gg.pages[gg.curpage - 1]->hwbase;

					while (hw != NULL)
					{
						hw1 = hw->next;
						if (gg.posx <= hw->x2 && hw->x1 <= x1 && hw->y >= gg.posy &&
								hw->y <= y1)
							delhwire(hw);
						hw = hw1;
					}
					vw = gg.pages[gg.curpage - 1]->vwbase;

					while (vw != NULL)
					{
						if (gg.posx <= vw->x && vw->x <= x1 && vw->y2 >= gg.posy)
						{
							if (vw->y1 >= gg.posy)
							{
								if (vw->y1 < y1)
								{
									vw->y1 = gg.posy;
								}
								else
								{
									vw->y1 += gg.posy - y1;
									vw->y2 += gg.posy - y1;
								}
							}
							else if (vw->y2 >= y1)
							{
								vw->y2 += gg.posy - y1;
							}
							else if (vw->y2 >= gg.posy)
							{
								vw->y2 = gg.posy;
							}
						}
						vw = vw->next;
					}
					hw = gg.pages[gg.curpage - 1]->hwbase;

					while (hw != NULL)
					{
						if (gg.posx <= hw->x2 && hw->x1 <= x1 && hw->y >= gg.posy)
							hw->y += gg.posy - y1;
						hw = hw->next;
					}
					b = gg.pages[gg.curpage - 1]->bbase;

					while (b != NULL)
					{
						b1 = b->next;
						if (gg.posx <= b->x1 && b->x2 <= x1)
						{
							if (b->y1 >= y1)
							{
								b->y1 += gg.posy - y1;
								b->y2 += gg.posy - y1;
							}
							else if (b->y1 >= gg.posy)
							{
								if (b->y2 < y1)
								{
									dispbox(&b);
								}
								else
								{
									b->y1 = gg.posy;
									b->y2 += gg.posy - y1;
								}
							}
							else if (b->y2 >= y1)
							{
								b->y2 += gg.posy - y1;
							}
							else if (b->y2 >= gg.posy)
							{
								b->y2 = gg.posy;
							}
						}
						b = b1;
					}
					l = gg.pages[gg.curpage - 1]->lbase;

					while (l != NULL)
					{
						l1 = l->next;
						if (gg.posx <= l->x && l->y >= gg.posy &&
								l->x + strlen(gg.nearlabel->name) * 7 / gg.scale <= x1)
						{
							if (l->y <= y1)
								displabel(&l);
							else
								l->y += gg.posy - y1;
						}
						l = l1;
					}
					refrfunc();
					scroll();
				}
				else
				{
					beginerror();
					nc_printf("Can't Close: Blocked by horizontal wires\n");
					enderror();
				}
			}
		}
	} while (gg.incircuit && gg.stillnear && *gg.func == '\0');
	gg.startpoint = false;
	log_setmode("");
	if (gg.t.dn && !gg.incircuit)
	{
		gg.t.dn = false;
		gg.t.depressed = false;
	}
	cursortype = normal;
}


static void centercommand()
{
	log_grec *g;
	log_hwrec *hw;
	log_vwrec *vw;
	log_srec *s;
	log_lrec *l;
	log_brec *b;
	short dx, dy, x1, y1, x2, y2;

	if (pagembb((int)gg.curpage, &x1, &y1, &x2, &y2))
	{
		dx = (origin + across / 2) / log_scale0 - (x2 + x1) / 2;
		dy = (origin + baseline / 2) / log_scale0 - (y2 + y1) / 2;
		g = gg.pages[gg.curpage - 1]->gbase;
		while (g != NULL)
		{
			shiftgate(g, dx, dy);
			g = g->next;
		}
		hw = gg.pages[gg.curpage - 1]->hwbase;

		while (hw != NULL)
		{
			hw->x1 += dx;
			hw->x2 += dx;
			hw->y += dy;
			hw = hw->next;
		}
		vw = gg.pages[gg.curpage - 1]->vwbase;

		while (vw != NULL)
		{
			vw->x += dx;
			vw->y1 += dy;
			vw->y2 += dy;
			vw = vw->next;
		}
		s = gg.pages[gg.curpage - 1]->sbase;

		while (s != NULL)
		{
			s->x += dx;
			s->y += dy;
			s = s->next;
		}
		l = gg.pages[gg.curpage - 1]->lbase;

		while (l != NULL)
		{
			l->x += dx;
			l->y += dy;
			l = l->next;
		}
		b = gg.pages[gg.curpage - 1]->bbase;

		while (b != NULL)
		{
			b->x1 += dx;
			b->x2 += dx;
			b->y1 += dy;
			b->y2 += dy;
			b = b->next;
		}
	}
	xoff0 = origin - gg.xoff;
	yoff0 = origin - gg.yoff;
	setscale(0);
	refrfunc();
}


static void getgate(char *name, short *gtype)
{
	*gtype = readlibrary(name);
}


static void getgategroup(strlist_t *grp)
{
	strlist_t *l1;
	short pos, i, count;
	int done;
	catboxrec *cb;

	l1 = grp;
	count = 0;
	while (l1 != NULL)
	{
		count += readlibrary_at(l1->s, 1, false);
		l1 = l1->next;
	}

	if (count != 0)
	{
		if (count > maxkinds)
			count = maxkinds;
		pos = 0;
		do
		{
			pos++;
			if (count > catwidth)
			{
				done = ((pos - 1) % catwidth == 0);
			}
			else
			{
				done = ((pos - 1) / catwidth == (pos + count - 2) / catwidth);
			}

			for (i = pos - 1; i <= pos + count - 2; i++)
			{
				if (kind[i] != NULL)
					done = false;
			}
		} while (!(done || pos + count > maxkinds));

		if (done)
		{
			if (count <= catwidth)
			{
				cb = (catboxrec *)Malloc(sizeof(catboxrec));
				cb->pos = pos;
				cb->count = count;
				cb->next = catboxes;
				catboxes = cb;
			}
		}
		else
		{
			pos = 1;
		}
	}
	else
	{
		pos = 1;
	}
	l1 = grp;
	while (l1 != NULL)
	{
		i = readlibrary_at(l1->s, pos, true);
		l1 = l1->next;
	}
	strlist_empty(&grp);
}


#define nummap          "0.,+123-456*789/.()^"


static void showpinname(log_grec *g, short i, short c, char *name)
{
	short x, y, w, y0, y1, y2;
	int small;
	char buf[256];

	if (zoom < 2)
	{
		*buf = '\0';
		small = true;
		size_t namelenght = strlen(name);
		for (size_t j = 0; j < namelenght; j++)
		{
			size_t j2 = strposc(nummap, name[j], 1L);
			if (j2 > 0)
				sprintf(buf + strlen(buf), "%c", j2 + 127);
			else if (isupper(name[j]))
				sprintf(buf + strlen(buf), "%c", name[j] + 103);
			else
				small = false;
		}
	}
	else
	{
		small = false;
	}

	if (small)
	{
		y0 = -5;
		y1 = -3;
		y2 = 3;
	}
	else
	{
		y0 = -3;
		y1 = -4;
		y2 = 4;
		strcpy(buf, name);
	}
	x = g->pinpos[i - 1].x * gg.scale - gg.xoff;
	y = g->pinpos[i - 1].y * gg.scale - gg.yoff;
	w = m_strwidth(logfont_lfont, buf);
	m_color((long)c);
	if (c == gg.color.backgr)
		m_fillrect(x - w / 2L - 1, (long)(y + y1), x + w / 2L + 1, (long)(y + y2));
	else
		m_centerstr((long)x, (long)(y + y0), logfont_lfont, buf);
}

#undef nummap


#define spc             12
#define xdiff           30
#define ydiff           20
#define border          3



static void showgateinfo(strlist_t *info, log_grec *g)
{
	strlist_t *l1;
	long width, height, i;
	short gx, gy, gx1, gy1, gx2, gy2, mx, my, y, z;
	log_krec *WITH;

	width = 10;
	height = 0;
	l1 = info;
	while (l1 != NULL)
	{
		i = m_strwidth(logfont_lfont, l1->s);
		if (i > width)
			width = i;
		height += spc;
		l1 = l1->next;
	}
	width += border * 2;
	height += border * 2 - 1;
	mx = -1;
	if (g != NULL)
	{
		gx = g->x * gg.scale - gg.xoff;
		gy = g->y * gg.scale - gg.yoff;
		z = g->rot;
		WITH = g->kind;
		gx1 = gx + (WITH->x1 * log_rotxx[z] + WITH->y1 * log_rotyx[z]) *
			gg.scale / log_scale0;
		gx2 = gx + (WITH->x2 * log_rotxx[z] + WITH->y2 * log_rotyx[z]) *
			gg.scale / log_scale0;
		sortshints(&gx1, &gx2);
		gy1 = gy + (WITH->x1 * log_rotxy[z] + WITH->y1 * log_rotyy[z]) *
			gg.scale / log_scale0;
		gy2 = gy + (WITH->x2 * log_rotxy[z] + WITH->y2 * log_rotyy[z]) *
			gg.scale / log_scale0;
		sortshints(&gy1, &gy2);
		gx1 -= 2;
		gx2 += 2;
		gy1 -= 2;
		gy2 += 2;
		if (gx2 + xdiff + width < across)
			mx = gx2 + xdiff;
		else if (gx1 - xdiff - width > 0)
			mx = gx1 - xdiff - width;
		else if (gx < across / 2)
			mx = across - width;
		else
			mx = 0;

		if (gy2 + ydiff + height < baseline)
			my = gy2 + ydiff;
		else if (gy1 - ydiff - height > 0)
			my = gy1 - ydiff - height;
		else if (gy < baseline / 2)
			my = baseline - height;
		else
			my = 0;

		if (false)
		{
			if (gx < across / 3)
				mx = (gx2 + across - width) / 2;
			else if (gx > across * 2 / 3 || (gy >= baseline / 3 && gy <= baseline))
				mx = (gx1 - width) / 2;
			else
				mx = (across - width) / 2;
			mx = P_imax2(P_imin2((long)mx, across - width), 0L);
			if (gy < baseline / 3)
				my = (gy2 + baseline - height) / 2;
			else if (gy > baseline * 2 / 3 || (gx >= across / 3 && gx <= across))
				my = (gy1 - height) / 2;
			else
				my = (baseline - height) / 2;
			my = P_imax2(P_imin2((long)my, baseline - height), 0L);
		}

		m_color((long)defineboxcolor);
		m_linestyle(2L);
		if (mx < gx1 || my < gy1)
			m_drawline((long)gx1, (long)gy1, (long)mx, (long)my);
		if (mx + width > gx2 || my < gy1)
			m_drawline((long)gx2, (long)gy1, mx + width, (long)my);
		if (mx < gx1 || my + height > gy2)
			m_drawline((long)gx1, (long)gy2, (long)mx, my + height);
		if (mx + width > gx2 || my + height > gy2)
			m_drawline((long)gx2, (long)gy2, mx + width, my + height);
		m_linestyle(0L);
	}

	if (mx < 0)
	{
		mx = P_imax2(across * 3L / 5 - width / 2, 0L);
		my = P_imax2(baseline * 3L / 5 - height / 2, 0L);
	}
	m_color((long)definebackcolor);
	m_fillrect((long)mx, (long)my, mx + width, my + height);
	m_color((long)defineboxcolor);
	m_drawrect((long)mx, (long)my, mx + width, my + height);
	m_color((long)gg.color.selword);
	y = my + border;
	l1 = info;
	while (l1 != NULL)
	{
		m_drawstr((long)(mx + border), (long)y, logfont_lfont, l1->s);
		y += spc;
		m_color((long)definetextcolor);
		l1 = l1->next;
	}
	strlist_empty(&info);
}

#undef spc
#undef xdiff
#undef ydiff
#undef border


static void showgatedef(log_krec *k, log_grec *g)
{
	strlist_t *l1;
	char STR1[42];

	gg.actstrlist = NULL;
	if (g != NULL)
	{
		send_gengate(g, "DUMPKIND");
		gg.actflag = false;
		send_gengate(g, "SHOWPINS");
	}

	if (gg.actstrlist == NULL)
		send_genkind(k, "DUMPKIND");

	if (gg.actstrlist == NULL)
	{
		sprintf(STR1, "No definition available for gate %s", k->name);
		l1 = strlist_append(&gg.actstrlist, STR1);
	}
	showgateinfo(gg.actstrlist, g);
}


static void gatedefinitioncommand()
{
	long i;
	char STR1[256];

	if (*gg.funcarg == '\0')
	{
		clearfunc();
		do
		{
			do
			{
				beginbottom();
				m_alpha_on();
				nc_gotoXY(0, txdown - 1);
				nc_printf("Select a gate to view its simulator definition.");
				do
				{
					pass();
					trykbd();
					pen();
					m_alpha_on();
				} while (!(gg.t.dn || *gg.func != '\0'));
				endbottom();
				scroll();
			} while (!(gg.t.dn || *gg.func != '\0'));
			if (*gg.func == '\0')
			{
				closergate(gg.gridx, gg.gridy);
				if (gg.neargate != NULL)
					showgatedef(gg.neargate->kind, gg.neargate);
			}
			else
			{
				gg.neargate = NULL;
			}
		} while (gg.neargate != NULL);
		return;
	}
	i = readlibrary(strupper(STR1, gg.funcarg));
	clearfunc();
	if (i != 0)
		showgatedef(kind[i - 1], NULL);
}


static librstrrec *findlibrstr(char *name_)
{
	size_t name_length = strlen(name_);
	char *name = (char*) malloc(sizeof(char) * name_length);

	librstrrec *lsp = librstrs;
	strreverse(name, name_);

	while (lsp != NULL && strcmp(lsp->name, name))
	{
		if (strcmp(name, lsp->name) < 0)
			lsp = lsp->left;
		else
			lsp = lsp->right;
	}

	free(name);
	return lsp;
}


#define maxmaxi         12
#define maxmaxj         45
#define huge_           30000


/* Local variables for listlibrary: */
struct LOC_listlibrary {
	short maxi, maxj;
	short karr[maxmaxi + 1][maxmaxj];
} ;

static short kfunc(short *i, short *j, struct LOC_listlibrary *LINK)
{
	/* K for Kludge */
	*i = (gg.t.ax - 2) / 10;
	*j = gg.t.ay - 1;
	if (!gg.t.near_)
	{
		return -1;
	}
	else if (*j <= 0 || (unsigned)(*i) > LINK->maxi)
	{
		return -2;
	}
	else if (*j > LINK->maxj)
	{
		if (*i < 1)
			return -3;
		else if (*i < 3)
			return -4;
		else
			return -2;
	}
	else if (LINK->karr[*i][*j - 1] == 0)
	{
		return (huge_ + *j * 8 + *i);
	}
	else
	{
		return (LINK->karr[*i][*j - 1]);
	}
}

/** @brief List the gates available in the library.

  - + or space bar displays next page;
  - - displays previous page;
  - anything else returns to Instructions.
  */
static void listlibrary()
{
	struct LOC_listlibrary V;
	short i, j, ii, jj, k, nn0, nn;
	librstrrec *lsp;
	char ch;
	int exitflag;
	char STR1[256];
	short FORLIM, FORLIM1;

	V.maxi = P_imin2((txacross - 9L) / 10, (long)maxmaxi);
	V.maxj = P_imin2(txdown - 3L, (long)maxmaxj);
	nn0 = 1;
	exitflag = false;
	do
	{
		clearshowalpha();
		nc_printf("Listing of gates ");
		if (librgroupnames[curlistgroup] != NULL &&
				strlen(librgroupnames[curlistgroup]) <= txacross - 26)
			nc_puts(strcjust(STR1, librgroupnames[curlistgroup], txacross - 26L));
		else
			nc_printf("%*c", txacross - 26, ' ');

		nc_printf("Group%2d\n", curlistgroup);
		FORLIM = V.maxi;
		for (i = 0; i <= FORLIM; i++)
		{
			FORLIM1 = V.maxj;
			for (j = 0; j < FORLIM1; j++)
				V.karr[i][j] = 0;
		}
		i = 0;
		j = 1;
		nn = nn0;
		do
		{
			if (indexgroup[nn - 1] == curlistgroup)
			{
				V.karr[i][j - 1] = nn;
				nc_gotoXY(i * 10 + 2, j + 1);
				if (P_getbits_UB(loadedgates, nn - 1, 0, 3))
					nc_putchar(chryellow);
				nc_printf("%s%c", index_[nn - 1], chrgreen);
				j++;
				if (j > V.maxj)
				{
					j = 1;
					i++;
				}
			}
			nn++;
		} while (i <= V.maxi && nn <= idxsize);
		nc_gotoXY(0, txdown);
		nc_printf("+ for next group, - for last group, space bar to quit.");
		do
		{
			k = kfunc(&i, &j, &V);
			lsp = NULL;
			if (k > 0)
			{
				if (k < huge_)
				{
					lsp = findlibrstr(index_[k - 1]);
					if (lsp != NULL)
					{
						nc_gotoXY(0, txdown - 1);
						nc_printf("%.*s", txacross, lsp->str);
					}
				}
				nc_gotoXY(i * 10 + 2, j + 1);
				if (k < huge_)
				{
					if (P_getbits_UB(loadedgates, k - 1, 0, 3))
						nc_putchar(chryellow);
					nc_printf("%c%s%c%c", chrinverse, index_[k - 1], chrgreen, chrplain);
				}
			}
			else
			{
				switch (k)
				{

					case -1:
						/* blank case */
						break;

					case -2:
						nc_gotoXY(49, txdown);
						nc_printf("%cquit%c.", chrinverse, chrplain);
						break;

					case -3:
						nc_gotoXY(0, txdown);
						nc_printf("%c+%c", chrinverse, chrplain);
						break;

					case -4:
						nc_gotoXY(18, txdown);
						nc_printf("%c-%c", chrinverse, chrplain);
						break;
				}
			}

			if (gg.t.near_ && gg.t.inalpha)
				nc_gotoXY(gg.t.ax, gg.t.ay);
			else
				noblink();

			do
			{
				pass();
				pen();
			} while (!(pollkbd2() || gg.t.dn || k != kfunc(&ii, &jj, &V)));

			remcursor();
			if (k > 0)
			{
				if (k < huge_)
				{
					if (lsp != NULL)
					{
						nc_gotoXY(0, txdown - 1);
						nc_putchar('\t');
					}
					nc_gotoXY(i * 10 + 2, j + 1);
					if (P_getbits_UB(loadedgates, k - 1, 0, 3))
						nc_putchar(chryellow);
					nc_printf("%s%c", index_[k - 1], chrgreen);
				}
			}
			else
			{
				switch (k)
				{

					case -1:
						/* blank case */
						break;

					case -2:
						nc_gotoXY(49, txdown);
						nc_printf("quit.");
						break;

					case -3:
						nc_gotoXY(0, txdown);
						nc_putchar('+');
						break;

					case -4:
						nc_gotoXY(18, txdown);
						nc_putchar('-');
						break;
				}
			}
		} while (!(pollkbd2() || gg.t.dn));
		ch = '\0';
		if (pollkbd2())
			ch = inkey2();

		if (gg.t.dn)
		{
			if (k >= huge_)
			{
				ch = '\003';
			}
			else if (k > 0)
			{
				if (readlibrary(index_[k - 1]) != 0)
				{
					remcursor();
					nc_gotoXY(i * 10 + 2, j + 1);
					nc_printf("%c%s%c", chryellow, index_[k - 1], chrgreen);
				}
			}
			else
			{
				switch (k)
				{

					case -1:
					case -2:
						ch = '\003';
						break;

					case -3:
						ch = '+';
						break;

					case -4:
						ch = '-';
						break;
				}
			}
		}
		if (ch >= '0' && ch <= '8')
		{
			curlistgroup = ch - 48;
		}
		else if (ch == '\b' || ch == '-')
		{
			curlistgroup = (curlistgroup + 8) % 9;
		}
		else if (ch == '\034' || ch == '+')
		{
			curlistgroup = (curlistgroup + 1) % 9;
		}
		else if (ch == '\003' || ch == '\015' || ch == 'Q' || ch == 'q' ||
				(ch & 255) == 171 || ch == ' ')
		{
			exitflag = true;
		}
	} while (!exitflag);
	clearwindow();
	m_graphics_on();
	clearalpha();
}

#undef maxmaxi
#undef maxmaxj
#undef huge_


/// Process CAT menu item. Display catalog, call ADDGATE if a gate is touched.
static void gatecatalog(int librmode)
{
	short x, y, x0, y0, x00, y00, w, x1, y1;
	char nm[9];
	/*  pic : m_picturevar;  */
	uchar ch;
	log_krec *k;
	char bot1[256], bot2[256], bot3[256];
	strlist_t *lp;
	int flag, oo, done, botflag, refrflag;
	long bottime;
	librstrrec *lsp;

	clearfunc();
	done = false;
	do
	{
		refrflag = false;
		if (librmode)
		{
			if (ch >= '0' && ch <= '8')
				curlistgroup = ch - 48;
			listlibrary();
			librmode = false;
		}
		else
		{
			showcatalog();
			m_color((long)gg.color.selword);
			drawstr2(10, down - 27, "LIBR");
			drawstr2(across - 32, down - 27, "LIBR");
			clearalpha();
			pen();
			do
			{
				ch = '\0';
				x0 = gg.t.x / (gridcen * 2);
				y0 = gg.t.y / (gridcen * 2);
				*bot1 = '\0';
				*bot2 = '\0';
				*bot3 = '\0';
				botflag = false;
				if ((unsigned)x0 < catwidth && y0 >= 0 &&
						y0 * catwidth + x0 < maxkinds &&
						gg.t.near_ && kind[y0 * catwidth + x0] != NULL)
				{
					flag = true;
					nm[8] = '\0';
					k = kind[y0 * catwidth + x0];
					strrtrim(strcpy(nm, k->name));
					w = m_strwidth(logfont_lfont, nm);
					x00 = (x0 * 2 + 1) * gridcen - w / 2;
					y00 = (y0 + 1) * gridcen * 2 + 2;
					remcursor();
					/*  m_getcpicture(imax(x00-2,0), y00-2, x00+w+4, y00+10, pic);
						m_color(gg.color.backgr);
						m_fillrect(x00-2, y00-2, x00+w+4, y00+10);
						m_color(gg.color.selword);
						m_centerstr((x0*2+1)*gridcen, y00, logfont_lfont, nm);*/
					m_colormode((long)m_xor);
					m_color((long)gg.color.selword);
					m_centerstr((x0 * 2L + 1) * gridcen, (long)y00, logfont_lfont, nm);
					m_colormode((long)m_normal);
					lp = k->lbl;

					while (lp != NULL && lp->kind != '\0')
						lp = lp->next;

					while (lp != NULL && *bot3 == '\0')
					{
						if (*bot1 == '\0')
							strcpy(bot1, lp->s);
						else if (*bot2 == '\0')
							strcpy(bot2, lp->s);
						else if (*bot3 == '\0')
							strcpy(bot3, lp->s);
						lp = lp->next;
					}

					if (*bot1 == '\0')
					{
						lsp = findlibrstr(k->name);
						if (lsp != NULL)
							strcpy(bot1, lsp->str);
					}
					bottime = timers_sysclock();
				}
				else
				{
					flag = false;
				}

				do
				{
					x = gg.t.x;
					y = gg.t.y;
					pass();
					pen();
					if (*bot1 != '\0' && !botflag)
					{
						if (labs(x - gg.t.x) > 10 || labs(y - gg.t.y) > 10)
						{
							bottime = timers_sysclock();
						}
						else if (timers_sysclock() > bottime + 10)
						{
							m_color((long)gg.color.selword);
							m_centerstr(across / 2L, down - 27L, NULL, bot1);
							if (*bot2 != '\0')
								m_centerstr(across / 2L, down - 17L, NULL, bot2);
							if (*bot3 != '\0')
								m_centerstr(across / 2L, down - 7L, NULL, bot3);
							botflag = true;
						}
					}
					if (pollkbd2())
						ch = inkey2();
					x1 = gg.t.x / (gridcen * 2);
					y1 = gg.t.y / (gridcen * 2);
				} while (!gg.t.dn && gg.t.near_ && x0 == x1 && y0 == y1 && ch == '\0');
				if (flag)
				{
					remcursor();
					/*  m_putcpicture(imax(x00-2,0), y00-2, pic);
						m_disposepicture(pic);  */
					m_colormode((long)m_xor);
					m_color((long)gg.color.selword);
					m_centerstr((x0 * 2L + 1) * gridcen, (long)y00, logfont_lfont, nm);
					m_colormode((long)m_normal);
					m_color((long)gg.color.backgr);

					if (botflag)
					{
						m_centerstr(across / 2L, down - 27L, NULL, bot1);
						if (*bot2 != '\0')
							m_centerstr(across / 2L, down - 17L, NULL, bot2);
						if (*bot3 != '\0')
							m_centerstr(across / 2L, down - 7L, NULL, bot3);
					}
				}

				if (gg.t.dn)
				{
					if ((x < 40 || x >= across - 37) && y >= down - 32)
						librmode = true;
					else
						done = true;
				}

				if (ch == 251 || ch == 250 || ch == ' ')
				{
					refrflag = true;
				}
				else if (ch == '\003' || ch == 171 || ch == 'Q' || ch == 'q')
				{
					done = true;
				}
				else if ((ch >= '0' && ch <= '8') || ch == 'L' || ch == 'l')
				{
					librmode = true;
				}
				else if (ch == 'D' || ch == 'd')
				{
					if ((unsigned)x0 < catwidth && y0 >= 0 &&
							y0 * catwidth + x0 < maxkinds &&
							gg.t.near_ && kind[y0 * catwidth + x0] != NULL)
						showgatedef(kind[y0 * catwidth + x0], NULL);
				}
			} while (!(done || librmode || refrflag));
			oo = gg.t.off;
			pen();
			pen();
			if (done && gg.t.depressed && !oo && !gg.t.off && flag)
			{
				refrwindow();
				addgate(y0 * catwidth + x0 + 1, 0, NULL);
			}
		}
	} while (!done);
}


/// Turn VLSI mode on or off.
void setvlsimode(int flag)
{
	if (flag)
	{
		beginerror();
		nc_printf("VLSI mode not supported in version %s\n", log_version);
		enderror();
		flag = false;
	}

	if (flag)
		curwcolor = log_wcol_blue;
	else
		curwcolor = log_wcol_normal;
	vlsi = flag;
	vlsimode[gg.curpage - 1] = vlsi;
}


/* was static void, cause gcc problems in some releases */
void prealunit(double r, short p, char *u, char *s)
{
	realunit(s, r, p, u, false);
}

/* was static void, cause gcc problems in some releases */
void prealunit2(double r, short p, char *u, char *s)
{
	realunit(s, r, p, u, true);
}


/* Local variables for editattrs: */
struct LOC_editattrs {
	log_gattrrec *gattr;
	short numattrs;
	log_kattrrec *kattr;
	strlist_t *lbl;
	char name[256];
	void (*chproc) ();
	log_tool *tool;
	log_grec *gate;
	log_nrec *node;
	short acty, p, ybase;
} ;

static void drawlabelline(long i, struct LOC_editattrs *LINK)
{
	long j;
	strlist_t *l1;
	long FORLIM;

	l1 = LINK->lbl;
	while (l1 != NULL && l1->kind != '\0')
		l1 = l1->next;
	FORLIM = i + LINK->ybase;
	for (j = 1; j <= FORLIM; j++)
	{
		if (l1 != NULL)
			l1 = l1->next;
	}

	if (l1 != NULL)
		nc_putStr(0, (int)i, l1->s);
}

static void eraselabelline(long i, struct LOC_editattrs *LINK)
{
	remcursor();
	nc_gotoXY(0, (int)i);
	nc_putchar('\t');
}

static void drawlabel(struct LOC_editattrs *LINK)
{
	short i, FORLIM;

	remcursor();
	nc_putStr(txacross - strlen(LINK->name), 0, LINK->name);
	FORLIM = txdown;
	for (i = 0; i <= FORLIM; i++)
		drawlabelline((long)i, LINK);
}

static void drawvalue(short i, int highlight, struct LOC_editattrs *LINK)
{
	strlist_t *l1;
	log_kattrrec *WITH;
	char STR1[81];
	char STR4[256];

	WITH = &LINK->kattr[i - 1];
	if (WITH->y < LINK->ybase || WITH->y > LINK->ybase + txdown)
		return;
	remcursor();
	nc_gotoXY(WITH->x, WITH->y - LINK->ybase);
	if (highlight)
		nc_putchar(129);
	/* Note: WRITE statement contains color/attribute characters [203] */
	if (!LINK->gattr[i - 1].blnk)
	{
		switch (WITH->dtype)
		{

			case 'R':
				nc_puts(realstr(STR1, LINK->gattr[i - 1].UU.r, WITH->prec));
				break;

			case 'U':
				nc_puts(realunit(STR1, LINK->gattr[i - 1].UU.r, WITH->prec,
							WITH->UU.U82.u, false));
				break;

			case 'F':
				nc_printf("%s%s",
						realstr(STR1, LINK->gattr[i - 1].UU.r, WITH->prec),
						WITH->UU.U82.u);
				break;

			case 'I':
				nc_printf("%*ld", WITH->prec, LINK->gattr[i - 1].UU.U73.i1);
				break;

			case 'H':
				nc_puts(strhex(STR4, LINK->gattr[i - 1].UU.U73.i1, (long)WITH->prec));
				break;

			case 'C':
				nc_printf("%.*s", txacross - WITH->x + 1, LINK->gattr[i - 1].UU.c);
				break;

			case 'A':
				nc_printf("%.*s", txacross - WITH->x + 1, LINK->gattr[i - 1].UU.sp);
				break;

			case 'B':
				if (WITH->prec == 1)
				{
					if (LINK->gattr[i - 1].UU.b)
						nc_printf("Yes");
					else
						nc_printf("No");
				}
				else
				{
					if (LINK->gattr[i - 1].UU.b)
						nc_printf("True");
					else
						nc_printf("False");
				}
				break;

			case 'V':
				l1 = WITH->UU.U86.v;
				while (l1 != NULL && (long)l1->value != LINK->gattr[i - 1].UU.nv)
					l1 = l1->next;
				if (l1 != NULL)
					nc_puts(l1->s);
				else
					nc_printf("(value not found)");
				break;
		}
	}
	if (WITH->x == XPOS)
		nc_putchar(' ');
	LINK->gattr[i - 1].x2 = XPOS - 1;   /*from CRT*/
	/*prob. no longer used*/
	nc_printf("\t\200");
	LINK->gattr[i - 1].changed = false;
}

static void maskvalues(struct LOC_editattrs *LINK)
{
	short i, FORLIM;

	FORLIM = LINK->numattrs;
	for (i = 0; i < FORLIM; i++)
	{
		if (LINK->kattr[i].vra != 0 &&
				LINK->kattr[i].vr != LINK->gattr[LINK->kattr[i].vra - 1].UU.nv)
		{
			if (!LINK->gattr[i].supr)
				eraselabelline((long)LINK->kattr[i].y, LINK);
			LINK->gattr[i].supr = true;
		}
		else
		{
			if (LINK->gattr[i].supr)
			{
				drawlabelline((long)LINK->kattr[i].y, LINK);
				drawvalue(i + 1, false, LINK);
			}
			LINK->gattr[i].supr = false;
		}
	}
}

static double cleanup(double r, double r1, struct LOC_editattrs *LINK)
{
	long i;
	char s[81];
	char *STR1;

	if (fabs(r) * 2 < r1)
	{
		return 0.0;
	}
	else if (fabs(r) < 1e9 && r >= 1e-5)
	{
		sprintf(s, "%20.0f", r / r1);
		i = strlen(s) + 1;
		s[i - 1] = '\0';
		r = strtod(s, &STR1);
		i = STR1 - s + 1;
		return (r * r1);
	}
	else
	{
		return r;
	}
}

static double scrnincr(short p, struct LOC_editattrs *LINK)
{
	long i;
	double r1;
	char s[81];
	log_kattrrec *WITH;

	WITH = &LINK->kattr[p - 1];
	switch (WITH->dtype)
	{

		case 'R':
			realstr(s, LINK->gattr[p - 1].UU.r, WITH->prec);
			break;

		case 'U':
			realunit(s, LINK->gattr[p - 1].UU.r, WITH->prec, "", false);
			break;

		case 'F':
			realstr(s, LINK->gattr[p - 1].UU.r, WITH->prec);
			break;
	}
	strcat(s, " ");
	i = 0;
	do
	{
		i++;
		if (s[i - 1] == '-')
			s[i - 1] = '+';
		else if (isdigit(s[i - 1]))
			s[i - 1] = '0';
	} while (s[i - 1] == '.' || s[i - 1] == '+' || s[i - 1] == '0');

	if (i > 1)
		s[i - 2] = '1';
	r1 = 0.0;
	readreal(s, &r1);
	return r1;
}

static void callconfig(void (*proc) (), struct LOC_editattrs *LINK)
{
	gg.actx = LINK->p;
	gg.acty = LINK->acty;
	gg.acttool = LINK->tool;
	gg.actgate = LINK->gate;
	gg.actnode = LINK->node;
	gg.actgattr = LINK->gattr;
	gg.actkattr = LINK->kattr;
	(*proc)();
}

static int tryconfig(struct LOC_editattrs *LINK)
{
	gg.actflag = true;
	callconfig(LINK->chproc, LINK);
	return (gg.actflag);
}


static void editattrs(log_gattrrec *gattr_, short numattrs_, log_kattrrec *kattr_, strlist_t *lbl_, char *name_,
		void (*proc) (), void (*chproc_) (), void (*relproc) (), long *attrstamp)
{
	struct LOC_editattrs V;
	short i, i1, p0, olday, ytotal;
	strlist_t *l1;
	long j, j1, savei;
	double r1, saver;
	char ch;
	char buf[256], savebuf[256];
	int exitflag, saveb, saveb2, touched, understood;
	short FORLIM;
	log_gattrrec *WITH;
	char STR1[256];
	char *STR2;
	char STR3[256];

	V.gattr = gattr_;
	V.numattrs = numattrs_;
	V.kattr = kattr_;
	V.lbl = lbl_;
	strcpy(V.name, name_);
	V.chproc = chproc_;
	if (V.lbl == NULL)
		return;
	V.tool = gg.acttool;
	V.gate = gg.actgate;
	V.node = gg.actnode;
	V.acty = gg.acty;
	clearshowalpha();
	V.ybase = 0;
	ytotal = 0;
	l1 = V.lbl;
	while (l1 != NULL)
	{
		if (l1->kind == '\0')
			ytotal++;
		l1 = l1->next;
	}
	drawlabel(&V);
	gg.actflag = true;
	callconfig(proc, &V);
	FORLIM = V.numattrs;
	for (i = 0; i < FORLIM; i++)
	{
		WITH = &V.gattr[i];
		touched = false;
		WITH->supr = false;
	}
	maskvalues(&V);
	FORLIM = V.numattrs;
	for (i = 1; i <= FORLIM; i++)
	{
		if (!V.gattr[i - 1].supr)
			drawvalue(i, false, &V);
	}
	showalpha();
	if (V.numattrs == 0)
	{
		do
		{
			noblink();
			do
			{
				pass();
				pen();
			} while (!(pollkbd2() || gg.t.dn));

			if (pollkbd2())
				ch = inkey2();
			else
				ch = '\015';
		} while (ch != '\003' && ch != '\015' && ch != ' ');
	}
	else
	{
		V.p = 1;
		olday = -1;
		exitflag = false;
		do
		{
			noblink();
			if (V.kattr[V.p - 1].y <= V.ybase ||
					V.kattr[V.p - 1].y >= V.ybase + txdown)
			{
				i1 = V.ybase;
				if (V.kattr[V.p - 1].y <= V.ybase)
				{
					V.ybase = V.kattr[V.p - 1].y - 1;
					if (V.p == 1)
						V.ybase = P_imax2(0L, (long)(V.kattr[V.p - 1].y - txdown));
				}
				else
				{
					V.ybase = V.kattr[V.p - 1].y - txdown + 1;
					if (V.p == V.numattrs)
						V.ybase = P_imin2(ytotal - txdown - 1L, (long)V.kattr[V.p - 1].y);
				}

				if (abs(V.ybase - i1) < txdown)
					nc_scrollXY(0, V.ybase - i1);
				else
					clearshowalpha();

				drawlabel(&V);
				FORLIM = V.numattrs;
				for (i = 0; i < FORLIM; i++)
					V.gattr[i].supr = false;

				maskvalues(&V);
				FORLIM = V.numattrs;
				for (i = 1; i <= FORLIM; i++)
				{
					if (!V.gattr[i - 1].supr)
						drawvalue(i, i == V.p, &V);
				}
			}
			else
			{
				drawvalue(V.p, true, &V);
			}

			do
			{
				ch = '\0';
				pass();
				gg.actflag = false;
				callconfig(proc, &V);
				FORLIM = V.numattrs;
				for (i = 1; i <= FORLIM; i++)
				{
					if (V.gattr[i - 1].changed)
					{
						stamp(attrstamp);
						if (!V.gattr[i - 1].supr)
						{
							drawvalue(i, i == V.p, &V);

							if (V.kattr[i - 1].dtype == 'V')
								maskvalues(&V);

							if (V.gattr[V.p - 1].supr)
								ch = '\037';
							noblink();
						}
					}
				}
				pen();
				/*  if gg.t.near and (gg.t.ay <> olday) then
					begin
i := 1;
while (i <= numattrs) and (kattr^[i].y <> gg.t.ay) do
i := i + 1;
if (i <= numattrs) and (not gattr^[i].supr) then
if i < p then
ch := #31
else if i > p then
ch := #10;
end;  */
				if (pollkbd2())
					ch = inkey2();
			} while (!(ch != '\0' || gg.t.dn));
			drawvalue(V.p, false, &V);

			if (gg.t.dn)
				ch = '\003';
			p0 = V.p;
			understood = true;
			WITH = &V.gattr[V.p - 1];

			if (ch == '\003' || ch == ' ')
			{
				exitflag = true;
			}
			else if ((ch & 255) != 251 && (ch & 255) != 250)
			{
				if (ch == '\n')
				{
					do
					{
						V.p++;
					} while (V.p <= V.numattrs && V.gattr[V.p - 1].supr);

					if (V.p > V.numattrs)
						V.p = p0;
				}
				else if (ch == '\037')
				{
					do
					{
						V.p--;
					} while (V.p >= 1 && V.gattr[V.p - 1].supr);

					if (V.p < 1)   /*should never happen!*/
						V.p = p0;
				}
				else if (ch == '\034')
				{
					gg.actflag = false;
					gg.actflag2 = true;
					callconfig(relproc, &V);

					if (!gg.actflag)
					{
						switch (V.kattr[V.p - 1].dtype)
						{

							case 'R':
							case 'U':
							case 'F':
								if (!WITH->blnk)
								{
									r1 = scrnincr(V.p, &V);
									saver = WITH->UU.r;
									WITH->UU.r = cleanup(WITH->UU.r + r1, r1, &V);
									if (tryconfig(&V))
									{
										drawvalue(V.p, false, &V);
										touched = true;
									}
									else
									{
										WITH->UU.r = saver;
									}
								}
								break;

							case 'I':
							case 'H':
								if (!WITH->blnk)
								{
									WITH->UU.U73.i1++;
									if (tryconfig(&V))
									{
										drawvalue(V.p, false, &V);
										touched = true;
									}
									else
									{
										WITH->UU.U73.i1--;
									}
								}
								break;

							case 'C':
							case 'A':
								understood = false;
								break;

							case 'B':
								saveb = WITH->UU.b;
								saveb2 = WITH->blnk;
								WITH->UU.b = true;
								WITH->blnk = false;

								if (!tryconfig(&V))
								{
									WITH->UU.b = saveb;
									WITH->blnk = saveb2;
								}

								if (WITH->UU.b != saveb || WITH->blnk != saveb2)
								{
									drawvalue(V.p, false, &V);
									touched = true;
								}
								break;

							case 'V':
								savei = WITH->UU.nv;
								if (WITH->UU.nv < V.kattr[V.p - 1].UU.U86.nv - 1)
									WITH->UU.nv++;

								if (!tryconfig(&V))
									WITH->UU.nv = savei;

								if (WITH->UU.nv != savei)
								{
									maskvalues(&V);
									drawvalue(V.p, false, &V);
									touched = true;
								}
								break;
						}
					}
					stamp(attrstamp);
					chpage(gg.pages[gg.curpage - 1]);
				}
				else if (ch == '\b')
				{
					gg.actflag = false;
					gg.actflag2 = false;
					callconfig(relproc, &V);
					if (!gg.actflag)
					{
						switch (V.kattr[V.p - 1].dtype)
						{

							case 'R':
							case 'U':
							case 'F':
								if (!WITH->blnk)
								{
									r1 = scrnincr(V.p, &V);
									saver = WITH->UU.r;
									WITH->UU.r = cleanup(WITH->UU.r - r1, r1, &V);
									if (tryconfig(&V))
									{
										drawvalue(V.p, false, &V);
										touched = true;
									}
									else
									{
										WITH->UU.r = saver;
									}
								}
								break;

							case 'I':
							case 'H':
								if (!WITH->blnk)
								{
									WITH->UU.U73.i1--;
									if (tryconfig(&V))
									{
										drawvalue(V.p, false, &V);
										touched = true;
									}
									else
									{
										WITH->UU.U73.i1++;
									}
								}
								break;

							case 'C':
							case 'A':
								understood = false;
								break;

							case 'B':
								saveb = WITH->UU.b;
								saveb2 = WITH->blnk;
								WITH->UU.b = false;
								WITH->blnk = false;
								if (!tryconfig(&V))
								{
									WITH->UU.b = saveb;
									WITH->blnk = saveb2;
								}

								if (WITH->UU.b != saveb || WITH->blnk != saveb2)
								{
									drawvalue(V.p, false, &V);
									touched = true;
								}
								break;

							case 'V':
								savei = WITH->UU.nv;
								if (WITH->UU.nv > 0)
									WITH->UU.nv--;

								if (!tryconfig(&V))
									WITH->UU.nv = savei;

								if (WITH->UU.nv != savei)
								{
									maskvalues(&V);
									drawvalue(V.p, false, &V);
									touched = true;
								}
								break;
						}
					}
					stamp(attrstamp);
					chpage(gg.pages[gg.curpage - 1]);
				}
				else
				{
					understood = false;
				}
			}

			if (!understood)
			{
				nc_gotoXY(V.kattr[V.p - 1].x, V.kattr[V.p - 1].y - V.ybase);
				nc_putchar('\t');

				if (ch == '\015')
					ungetkey2('\003');
				else
					ungetkey2(ch);

				switch (V.kattr[V.p - 1].dtype)
				{

					case 'R':
					case 'U':
					case 'F':
						readlnpass(buf, 2);
						strcpy(STR1, strltrim(buf));
						strcpy(buf, STR1);
						saveb2 = touched;
						saveb = WITH->blnk;
						saver = WITH->UU.r;
						if (*buf == '\0' ||
								!(buf[0] == '-' || buf[0] == '+' || buf[0] == '.' ||
									isdigit(buf[0])))
						{
							if (V.kattr[V.p - 1].opt)
							{
								WITH->blnk = true;
							}
							else
							{
								WITH->blnk = false;
								WITH->UU.r = V.kattr[V.p - 1].UU.U82.r;
							}
							touched = true;
						}
						else
						{
							TRY(try16);
							readreal(buf, &r1);
							WITH->UU.r = r1;
							WITH->blnk = false;
							touched = true;
							RECOVER(try16);
							if (P_escapecode == -20)
								_Escape(P_escapecode);
							ENDTRY(try16);
						}

						if (!tryconfig(&V))
						{
							WITH->UU.r = saver;
							WITH->blnk = saveb;
							touched = saveb2;
						}
						break;

					case 'I':
						readlnpass(buf, 2);
						strcpy(STR1, strltrim(buf));
						strcpy(buf, STR1);
						saveb2 = touched;
						saveb = WITH->blnk;
						savei = WITH->UU.U73.i1;
						if (*buf == '\0' ||
								!(buf[0] == '+' || buf[0] == '-' || isdigit(buf[0])))
						{
							if (V.kattr[V.p - 1].opt)
							{
								WITH->blnk = true;
							}
							else
							{
								WITH->blnk = false;
								WITH->UU.U73.i1 = V.kattr[V.p - 1].UU.U73.i1;
							}
							touched = true;
						}
						else
						{
							TRY(try17);
							j1 = strtol(buf, &STR2, 10);
							j = STR2 - buf + 1;
							WITH->UU.U73.i1 = j1;
							WITH->blnk = false;
							touched = true;
							RECOVER(try17);
							if (P_escapecode == -20)
								_Escape(P_escapecode);
							ENDTRY(try17);
						}

						if (!tryconfig(&V))
						{
							WITH->UU.U73.i1 = savei;
							WITH->blnk = saveb;
							touched = saveb2;
						}
						break;

					case 'H':
						readlnpass(buf, 2);
						strcpy(STR1, strltrim(buf));
						strcpy(buf, STR1);
						saveb2 = touched;
						saveb = WITH->blnk;
						savei = WITH->UU.U73.i1;
						if (*buf == '\0' ||
								!((buf[0] >= 'a' && buf[0] <= 'f') ||
									(buf[0] >= 'A' && buf[0] <= 'F') || isdigit(buf[0])))
						{
							if (V.kattr[V.p - 1].opt)
							{
								WITH->blnk = true;
							}
							else
							{
								WITH->blnk = false;
								WITH->UU.U73.i1 = V.kattr[V.p - 1].UU.U73.i1;
							}
							touched = true;
						}
						else
						{
							TRY(try18);
							WITH->UU.U73.i1 = strtol(buf, NULL, 16);
							WITH->blnk = false;
							touched = true;
							RECOVER(try18);
							if (P_escapecode == -20)
								_Escape(P_escapecode);
							ENDTRY(try18);
						}

						if (!tryconfig(&V))
						{
							WITH->UU.U73.i1 = savei;
							WITH->blnk = saveb;
							touched = saveb2;
						}
						break;

					case 'C':
						strcpy(buf, WITH->UU.c);
						readlnpass(buf, 3);
						strcpy(savebuf, WITH->UU.c);
						strcpy(STR1, strltrim(strrtrim(strcpy(STR3, buf))));
						strcpy(buf, STR1);
						if (*buf == '\0' && !V.kattr[V.p - 1].opt)
						{
							strcpy(buf, V.kattr[V.p - 1].UU.c);
						}
						else if (strlen(buf) > V.kattr[V.p - 1].prec)
						{
							buf[V.kattr[V.p - 1].prec] = '\0';
						}
						strcpy(WITH->UU.c, buf);
						if (tryconfig(&V))
							touched = true;
						else
							strcpy(WITH->UU.c, savebuf);
						WITH->blnk = (*WITH->UU.c == '\0');
						break;

					case 'A':
						strcpy(buf, WITH->UU.c);
						readlnpass(buf, 3);
						strcpy(savebuf, WITH->UU.sp);
						strcpy(STR1, strltrim(strrtrim(strcpy(STR3, buf))));
						strcpy(buf, STR1);
						if (*buf == '\0' && !V.kattr[V.p - 1].opt)
							strcpy(buf, V.kattr[V.p - 1].UU.sp);
						strchange(&WITH->UU.sp, buf);
						if (tryconfig(&V))
							touched = true;
						else
							strchange(&WITH->UU.sp, savebuf);
						WITH->blnk = (*WITH->UU.sp == '\0');
						break;

					case 'B':
						saveb = WITH->blnk;
						saveb2 = WITH->UU.b;
						ch = inkey2();
						if (ch == '1' || ch == 'y' || ch == 'Y' || ch == 't' || ch == 'T')
						{
							WITH->UU.b = true;
							WITH->blnk = false;
						}
						else if (ch == '0' || ch == 'n' || ch == 'N' || ch == 'f' || ch == 'F')
						{
							WITH->UU.b = false;
							WITH->blnk = false;
						}
						else if ((ch == '\003' || ch == '\015' || ch == 'x' || ch == 'X') && V.kattr[V.p - 1].opt)
						{
							WITH->blnk = true;
						}
						else if (ch == '\003' || ch == '\015' || ch == 'z' || ch == 'Z')
						{
							WITH->UU.b = !WITH->UU.b;
							WITH->blnk = false;
						}

						if (tryconfig(&V))
						{
							touched = true;
						}
						else
						{
							WITH->blnk = saveb;
							WITH->UU.b = saveb2;
						}
						break;

					case 'V':
						readlnpass(buf, 2);
						savei = WITH->UU.nv;
						strcpy(STR1, strltrim(strrtrim(strcpy(STR3, buf))));
						strcpy(buf, STR1);
						l1 = V.kattr[V.p - 1].UU.U86.v;
						while (l1 != NULL && strcicmp(l1->s, buf) != 0)
							l1 = l1->next;

						if (l1 != NULL)
							WITH->UU.nv = (long)l1->value;

						if (tryconfig(&V))
						{
							maskvalues(&V);
							touched = true;
						}
						else
						{
							WITH->UU.nv = savei;
						}
						break;
				}
				stamp(attrstamp);
				chpage(gg.pages[gg.curpage - 1]);
				drawvalue(V.p, false, &V);
			}
		} while (!exitflag);
	}
	m_graphics_on();
	clearalpha();

	/*ignore*/
}


static void editattrsx(log_gattrrec *gattr, short numattrs, log_kattrrec *kattr, strlist_t *lbl, char *name, void (*proc) (), void (*chproc) (), void (*relproc) ())
{
	long stamp;

	editattrs(gattr, numattrs, kattr, lbl, name, proc, chproc, relproc, &stamp);
}


static void gproc1()
{
	calltool(gg.acttool, act_configgate);
}


static void gproc2()
{
	calltool(gg.acttool, act_configchgate);
}


static void gproc3()
{
	calltool(gg.acttool, act_configrelgate);
}


static void configgate(log_grec *g)
{
	void (*TEMP) ();
	void (*TEMP5) ();
	void (*TEMP6) ();

	gg.actgate = g;
	gg.acttool = g->kind->simtype;
	TEMP = gproc1;
	TEMP5 = gproc2;
	TEMP6 = gproc3;
	editattrs(g->attr, g->kind->numattrs, g->kind->attr, g->kind->lbl,
			g->kind->name, TEMP, TEMP5, TEMP6, &gg.gattrstamp);
}


static void nproc1()
{
	calltool(gg.acttool, act_confignode);
}


static void nproc2()
{
	calltool(gg.acttool, act_configchnode);
}


static void nproc3()
{
	calltool(gg.acttool, act_configrelnode);
}


static void confignode(log_nrec *n, char *name)
{
	void (*TEMP) ();
	void (*TEMP5) ();
	void (*TEMP6) ();

	gg.actnode = n;
	gg.acttool = n->simtype;
	TEMP = nproc1;
	TEMP5 = nproc2;
	TEMP6 = nproc3;
	editattrs(n->attr, n->simtype->nnumattrs, n->simtype->nattr,
			n->simtype->nlbl, name, TEMP, TEMP5, TEMP6, &gg.nattrstamp);
}


static void configkind(short i)
{
	log_krec *k;
	log_grec *g;

	if (kindgroup[i - 1] == 0)
		return;
	k = kind[(kindgroup[i - 1] & (log_kindoffset - 1)) - 1];
	newgate(&g, kindgroup[i - 1]);
	if (kindattr[i - 1] != NULL)
	{
		disposeattrs(&g->attr, k->numattrs, k->attr);
		copyattrs(&g->attr, kindattr[i - 1], k->numattrs, k->attr);
	}
	configgate(g);
	if (kindattr[i - 1] != NULL)
		disposeattrs(&kindattr[i - 1], k->numattrs, k->attr);
	copyattrs(&kindattr[i - 1], g->attr, k->numattrs, k->attr);
	disposegate(&g);
}


/* Local variables for setattr: */
struct LOC_setattr {
	log_gattrrec *gattr;
	log_kattrrec *kattr;
	short p;
} ;

static int tryconfig_(struct LOC_setattr *LINK)
{
	gg.actflag = true;
	gg.actx = LINK->p;
	gg.actgattr = LINK->gattr;
	gg.actkattr = LINK->kattr;
	(*gg.acttool->proc)(&gg);
	return (gg.actflag);
}

static int setattr(log_gattrrec *gattr_, log_kattrrec *kattr_, short p_, char *buf)
{
	struct LOC_setattr V;
	int touched;
	long savei;
	double r1, saver;
	int saveb, saveb2;
	char savebuf[256];
	long j, j1;
	strlist_t *l1;
	log_gattrrec *WITH;
	char STR1[256];
	char *STR2;
	char STR3[256];

	V.gattr = gattr_;
	V.kattr = kattr_;
	V.p = p_;
	touched = false;
	WITH = &V.gattr[V.p - 1];
	switch (V.kattr[V.p - 1].dtype)
	{

		case 'R':
		case 'U':
		case 'F':
			strcpy(STR1, strltrim(buf));
			strcpy(buf, STR1);
			saveb2 = touched;
			saveb = WITH->blnk;
			saver = WITH->UU.r;
			if (*buf == '\0' || !(buf[0] == '-' || buf[0] == '+' || buf[0] == '.' || isdigit(buf[0])))
			{
				if (V.kattr[V.p - 1].opt)
				{
					WITH->blnk = true;
				}
				else
				{
					WITH->blnk = false;
					WITH->UU.r = V.kattr[V.p - 1].UU.U82.r;
				}
				touched = true;
			}
			else
			{
				TRY(try19);
				readreal(buf, &r1);
				WITH->UU.r = r1;
				WITH->blnk = false;
				touched = true;
				RECOVER(try19);
				if (P_escapecode == -20)
					_Escape(P_escapecode);
				ENDTRY(try19);
			}

			if (!tryconfig_(&V))
			{
				WITH->UU.r = saver;
				WITH->blnk = saveb;
				touched = saveb2;
			}
			break;

		case 'I':
			strcpy(STR1, strltrim(buf));
			strcpy(buf, STR1);
			saveb2 = touched;
			saveb = WITH->blnk;
			savei = WITH->UU.U73.i1;
			if (*buf == '\0' || !(buf[0] == '+' || buf[0] == '-' || isdigit(buf[0])))
			{
				if (V.kattr[V.p - 1].opt)
				{
					WITH->blnk = true;
				}
				else
				{
					WITH->blnk = false;
					WITH->UU.U73.i1 = V.kattr[V.p - 1].UU.U73.i1;
				}
				touched = true;
			}
			else
			{
				TRY(try20);
				j1 = strtol(buf, &STR2, 10);
				j = STR2 - buf + 1;
				WITH->UU.U73.i1 = j1;
				WITH->blnk = false;
				touched = true;
				RECOVER(try20);
				if (P_escapecode == -20)
					_Escape(P_escapecode);
				ENDTRY(try20);
			}

			if (!tryconfig_(&V))
			{
				WITH->UU.U73.i1 = savei;
				WITH->blnk = saveb;
				touched = saveb2;
			}
			break;

		case 'H':
			strcpy(STR1, strltrim(buf));
			strcpy(buf, STR1);
			saveb2 = touched;
			saveb = WITH->blnk;
			savei = WITH->UU.U73.i1;
			if (*buf == '\0' ||
					!((buf[0] >= 'a' && buf[0] <= 'f') || (buf[0] >= 'A' && buf[0] <= 'F') ||
						isdigit(buf[0])))
			{
				if (V.kattr[V.p - 1].opt)
				{
					WITH->blnk = true;
				}
				else
				{
					WITH->blnk = false;
					WITH->UU.U73.i1 = V.kattr[V.p - 1].UU.U73.i1;
				}
				touched = true;
			}
			else
			{
				TRY(try21);
				WITH->UU.U73.i1 = strtol(buf, NULL, 16);
				WITH->blnk = false;
				touched = true;
				RECOVER(try21);
				if (P_escapecode == -20)
					_Escape(P_escapecode);
				ENDTRY(try21);
			}

			if (!tryconfig_(&V))
			{
				WITH->UU.U73.i1 = savei;
				WITH->blnk = saveb;
				touched = saveb2;
			}
			break;

		case 'C':
			strcpy(savebuf, WITH->UU.c);
			strcpy(STR1, strltrim(strrtrim(strcpy(STR3, buf))));
			strcpy(buf, STR1);
			if (*buf == '\0' && !V.kattr[V.p - 1].opt)
			{
				strcpy(buf, V.kattr[V.p - 1].UU.c);
			}
			else if (strlen(buf) > V.kattr[V.p - 1].prec)
			{
				buf[V.kattr[V.p - 1].prec] = '\0';
			}
			strcpy(WITH->UU.c, buf);

			if (tryconfig_(&V))
				touched = true;
			else
				strcpy(WITH->UU.c, savebuf);
			WITH->blnk = (*WITH->UU.c == '\0');
			break;

		case 'A':
			strcpy(savebuf, WITH->UU.sp);
			strcpy(STR1, strltrim(strrtrim(strcpy(STR3, buf))));
			strcpy(buf, STR1);
			if (*buf == '\0' && !V.kattr[V.p - 1].opt)
				strcpy(buf, V.kattr[V.p - 1].UU.sp);
			strchange(&WITH->UU.sp, buf);
			if (tryconfig_(&V))
				touched = true;
			else
				strchange(&WITH->UU.sp, savebuf);
			WITH->blnk = (*WITH->UU.sp == '\0');
			break;

		case 'B':
			saveb = WITH->blnk;
			saveb2 = WITH->UU.b;
			if (*buf == '\0')
				strcpy(buf, " ");
			if (buf[0] == '1' || buf[0] == 'y' || buf[0] == 'Y' || buf[0] == 't' || buf[0] == 'T')
			{
				WITH->UU.b = true;
				WITH->blnk = false;
			}
			else if (buf[0] == '0' || buf[0] == 'n' || buf[0] == 'N' || buf[0] == 'f' || buf[0] == 'F')
			{
				WITH->UU.b = false;
				WITH->blnk = false;
			}
			else if ((buf[0] == 'x' || buf[0] == 'X' || buf[0] == ' ') && V.kattr[V.p - 1].opt)
			{
				WITH->blnk = true;
			}
			else if (buf[0] == 'z' || buf[0] == 'Z' || buf[0] == ' ')
			{
				WITH->UU.b = !WITH->UU.b;
				WITH->blnk = false;
			}

			if (tryconfig_(&V))
			{
				touched = true;
			}
			else
			{
				WITH->blnk = saveb;
				WITH->UU.b = saveb2;
			}
			break;

		case 'V':
			savei = WITH->UU.nv;
			strcpy(STR1, strltrim(strrtrim(strcpy(STR3, buf))));
			strcpy(buf, STR1);
			l1 = V.kattr[V.p - 1].UU.U86.v;
			while (l1 != NULL && strcicmp(l1->s, buf) != 0)
				l1 = l1->next;

			if (l1 != NULL)
				WITH->UU.nv = (long)l1->value;

			if (tryconfig_(&V))
				touched = true;
			else
				WITH->UU.nv = savei;
			break;
	}
	return touched;
}


static void setgattr(log_grec *g, short num, char *value_)
{
	char value[256];
	log_grec *g1;
	short pg;
	log_tool *saveacttool;

	if (num < 1 || num > g->kind->numattrs)
		return;
	strcpy(value, value_);
	saveacttool = gg.acttool;
	gg.actgate = g;
	gg.actnode = NULL;
	gg.acttool = g->kind->simtype;
	gg.action = act_configchgate;
	if (setattr(g->attr, g->kind->attr, num, value))
	{
		stamp(&gg.gattrstamp);
		g1 = NULL;
		pg = gg.numpages + 1;
		while (pg > 1 && g1 != NULL)
		{
			pg--;
			g1 = gg.pages[pg - 1]->gbase;
			while (g1 != NULL && g1 != g)
				g1 = g1->next;
		}
		if (g1 != NULL)
			chpage(gg.pages[pg - 1]);
	}
	gg.acttool = saveacttool;
}


static void setnattr(log_nrec *n, short num, char *value_)
{
	char value[256];
	log_tool *saveacttool;

	strcpy(value, value_);
	if (num < 1 || num > n->simtype->nnumattrs)
		return;
	saveacttool = gg.acttool;
	gg.actgate = NULL;
	gg.actnode = n;
	gg.acttool = n->simtype;
	gg.action = act_configchnode;
	if (setattr(n->attr, n->simtype->nattr, num, value))
		stamp(&gg.nattrstamp);
	gg.acttool = saveacttool;
}


typedef short flipvec[8];
typedef flipvec fliparr[4];


const fliparr flips = {
	{ 4, 5, 6, 7, 0, 1, 2, 3 },
	{ 1, 2, 3, 0, 5, 6, 7, 4 },
	{ 4, 7, 6, 5, 0, 3, 2, 1 },
	{ 6, 5, 4, 7, 2, 1, 0, 3 }
};


static short doflip(short rot, short mode)
{
	return flips[mode][rot];
}


/// Flip a gate in the circuit.
static void flipgate(log_grec *g)
{
	short i;
	log_krec *WITH;

	if (g->kind->flag.U3.noflip)
	{
		configgate(g);
		return;
	}

	WITH = g->kind;

	if (g->kind->flag.U3.toggle)
		i = 0;
	else
		i = cureditmode;

	switch (i)
	{

		case 0:
		case 1:
		case 2:
		case 3:
			remcursor();
			clipon();
			eragate(g);
			disconnectgate(g);
			g->rot = doflip(g->rot, i);
			g->g = g->rot * log_kindoffset + (g->g & (log_kindoffset - 1));
			drawgatex(g);
			clipoff();
			if (!connectgate(g))
			{
				frygate(g);
				disposegate(&g);
			}
			restorecursor();
			refreshsoon();
			break;

		case 4:
			configgate(g);
			break;
	}
}


/// Flip a gate in menu area.
static void flipkind()
{
	short i, z, k;

	remcursor();
	i = (gg.t.x - kindgroupleft) / kindgroupspacing + 1;
	k = kindgroup[i - 1] & (log_kindoffset - 1);
	if (kindgroup[i - 1] != 0 && !kind[k - 1]->flag.U3.noflip)
	{
		clipoff();
		kdrawgatec(i, gg.color.backgr);
		z = kindgroup[i - 1] / log_kindoffset;
		if (kind[k - 1]->flag.U3.named)
		{
			z = doflip(z, 2);
		}
		else
		{
			switch (cureditmode)
			{

				case 1:
				case 2:
				case 3:
					z = doflip(z, cureditmode);
					break;

				case 4:
					z = doflip(z, 1);
					break;
			}
		}
		kindgroup[i - 1] = z * log_kindoffset + k;
		kdrawgatec(i, gg.color.kindgate);
	}
	restorecursor();
}


void settofrom(log_grec **g, char *name)
{
	if (!(*g)->kind->flag.U3.named)
		return;
	disconnectgate(*g);
	clipon();
	eragate(*g);
	clipoff();
	(*g)->sig = getsignal((*g)->sig, name);
	if (connectgate(*g))
	{
		clipon();
		drawgatex(*g);
		clipoff();
		return;
	}
	frygate(*g);
	disposegate(g);
	*g = NULL;
}

/// Enter a new name for nameable gate.
static void adjustsignal(log_grec *g)
{
	char n[256];
	uchar ch;
	int rightface;
	char STR1[256], STR2[256];

	rightface = (g->kind->flag.U3.nright != (g->rot == 0));
	gsignallabel(g->x, g->y, g, gg.color.backgr);
	*n = '\0';
	do
	{
		if (strlen(n) < 256-1)
			strcat(n, "_");
		remcursor();
		m_colormode((long)m_xor);
		m_color((long)gg.color.signal);
		if (rightface)
		{
			drawstr2((int)(g->x * gg.scale - gg.xoff - m_strwidth(logfont_lfont,
							strrtrim(strcpy(STR1, n))) + NAMEGAP),
					(int)(g->y * gg.scale - gg.yoff - 3),
					strrtrim(strcpy(STR2, n)));
		}
		else
		{
			drawstr2((int)(g->x * gg.scale - gg.xoff - NAMEGAP),
					(int)(g->y * gg.scale - gg.yoff - 3),
					strrtrim(strcpy(STR1, n)));
		}
		m_colormode((long)m_normal);
		if (!pollkbd2())
		{
			do
			{
				pass();
				pen();
			} while (!(pollkbd2() || gg.t.dn));
		}
		remcursor();
		m_colormode((long)m_xor);
		m_color((long)gg.color.signal);
		if (rightface)
		{
			drawstr2((int)(g->x * gg.scale - gg.xoff - m_strwidth(logfont_lfont,
							strrtrim(strcpy(STR1, n))) + NAMEGAP),
					(int)(g->y * gg.scale - gg.yoff - 3),
					strrtrim(strcpy(STR2, n)));
		}
		else
		{
			drawstr2((int)(g->x * gg.scale - gg.xoff - NAMEGAP),
					(int)(g->y * gg.scale - gg.yoff - 3),
					strrtrim(strcpy(STR1, n)));
		}
		m_colormode((long)m_normal);
		n[strlen(n) - 1] = '\0';
		if (pollkbd2())
		{
			ch = inkey2();
			if ((strlen(n) < 256-2 && ch > ' ' && ch != 250 && ch != 251)
					|| (ch == ' ' && *n != '\0'))
				sprintf(n + strlen(n), "%c", ch);

			if (ch == '\007' && *n != '\0')
				n[strlen(n) - 1] = '\0';
		}
	} while (!((ch < 32 && ((1L << ch) & 0x2008) != 0) || gg.t.dn));
	remcursor();
	m_color((long)gg.color.signal);
	if (rightface)
	{
		drawstr2((int)(g->x * gg.scale - gg.xoff - m_strwidth(logfont_lfont,
						strrtrim(strcpy(STR1, n))) + NAMEGAP),
				(int)(g->y * gg.scale - gg.yoff - 3), strrtrim(strcpy(STR2, n)));
	}
	else
	{
		drawstr2((int)(g->x * gg.scale - gg.xoff - NAMEGAP),
				(int)(g->y * gg.scale - gg.yoff - 3), strrtrim(strcpy(STR1, n)));
	}
	settofrom(&g, n);
}


static void touchgate(log_grec *g)
{
	short xx, yy;

	if (g->kind->flag.U3.named)
	{
		adjustsignal(g);
		return;
	}
	xx = gg.gridx - g->x;
	yy = gg.gridy - g->y;
	gg.actx = log_irotxx[g->rot] * xx + log_irotyx[g->rot] * yy;
	gg.acty = log_irotxy[g->rot] * xx + log_irotyy[g->rot] * yy;
	xx = gg.t.x - g->x * gg.scale + gg.xoff;
	yy = gg.t.y - g->y * gg.scale + gg.yoff;
	gg.actx2 = log_irotxx[g->rot] * xx + log_irotyx[g->rot] * yy;
	gg.acty2 = log_irotxy[g->rot] * xx + log_irotyy[g->rot] * yy;
	gg.actflag = false;
	calltoolgate(g, act_touchgate);

	if (gg.actflag)
		chpageplace(gg.pages[gg.curpage - 1], g->x - g->kind->bbmax, g->y - g->kind->bbmax,
				g->x + g->kind->bbmax, g->y + g->kind->bbmax);
	else if (!gg.invisible || g->kind->flag.U3.visible)
		flipgate(g);
}

static void findattrnum2(short numattrs, log_kattrrec *kattr, strlist_t *lbl, char *name, char *kinds, short *num)
{
	strlist_t *l1;

	if (strsubset(name, "0123456789"))
	{
		if (*name == '\0')
			*num = 0;
		else
			*num = strtol(name, NULL, 0);
	}
	else
	{
		l1 = lbl;
		while (l1 != NULL && l1->kind != '\001')
			l1 = l1->next;

		if (l1 != NULL)
			l1 = strlist_find((strlist_t *)l1->value, name);

		if (l1 != NULL)
			*num = (long)l1->value;
		else
			*num = 0;
	}

	if (*num < 1 || *num > numattrs ||
			(strposc(kinds, kattr[*num - 1].dtype, 1L) == 0 && *kinds != '\0'))
		*num = 0;
}


static void findattrnum(log_krec *k, char *name, char *kinds, short *num)
{
	findattrnum2(k->numattrs, k->attr, k->lbl, name, kinds, num);
}


static void findattrname2(short numattrs, log_kattrrec *kattr, strlist_t *lbl, short num, char *name)
{
	strlist_t *l1;

	if (num < 1 || num > numattrs)
	{
		*name = '\0';
		return;
	}
	l1 = lbl;

	while (l1 != NULL && l1->kind != '\001')
		l1 = l1->next;

	if (l1 != NULL)
	{
		l1 = (strlist_t *)l1->value;
		while (l1 != NULL && (long)l1->value != num)
			l1 = l1->next;
	}

	if (l1 != NULL)
		strcpy(name, l1->s);
	else
		sprintf(name, "%d", num);
}


static void findattrname(log_krec *k, short num, char *name)
{
	findattrname2(k->numattrs, k->attr, k->lbl, num, name);
}


static void findpinnum(log_krec *k, char *name_, short *num)
{
	char name[256];

	strcpy(name, name_);
	if (*name == '#')
		strcpy_overlap(name, name + 1);

	if (strsubset(name, "0123456789"))
	{
		if (*name == '\0')
		{
			*num = 0;
			return;
		}
		*num = strtol(name, NULL, 0);
		if (*num < 1 || *num > k->numpins)
			*num = 0;
		return;
	}
	*num = 1;
	while (*num <= k->numpins &&
			strlist_find(k->pinnames[*num - 1], name) == NULL)
		(*num)++;
	if (*num > k->numpins)
		*num = 0;
}


static void findpinname(log_krec *k, short num, char *name)
{
	if (num >= 1 && num <= k->numpins && k->pinnames[num - 1] != NULL)
	{
		strcpy(name, k->pinnames[num - 1]->s);
		return;
	}

	if (num == 0)
		*name = '\0';
	else
		sprintf(name, "%d", num);
}


static void findpointmarker(log_krec *k, short num, short *x, short *y)
{
	long i;

	i = k->numvects;
	while (i >= 1 && (k->vector[i - 1].vkind != 'p' || k->vector[i - 1].UU.U98.num != num))
		i--;

	if (i >= 1)
	{
		*x = k->vector[i - 1].x1;
		*y = k->vector[i - 1].y1;
	}
}


static void findboxmarker(log_krec *k, short num, short *x1, short *y1, short *x2, short *y2)
{
	long i;

	i = k->numvects;
	while (i >= 1 && (k->vector[i - 1].vkind != 'b' || k->vector[i - 1].UU.U98.num != num))
		i--;

	if (i < 1)
		return;

	k->x1 = k->vector[i - 1].x1;
	k->y1 = k->vector[i - 1].y1;
	k->x2 = k->vector[i - 1].UU.U99.x2;
	k->y2 = k->vector[i - 1].UU.U99.y2;
}


static void initcolors ();

/* Local variables for doshellescape: */
struct LOC_doshellescape {
	long oldc;
} ;

static void recov(struct LOC_doshellescape *LINK)
{
	initwindow2();

	if (m_maxcolor != LINK->oldc)
		initcolormap();

	initcolors();
	refrwindow();
}


static void doshellescape(char *arg)
{
	struct LOC_doshellescape V;
	long i, j;
	char STR1[256];

	m_graphics_off();
	m_alpha_on();
	V.oldc = m_maxcolor;
	TRY(try22);
	closedumpfiles();
	m_alpha_off();
	m_graphics_on();
	recov(&V);
	RECOVER(try22);
	i = P_escapecode;
	j = P_ioresult;
	recov(&V);
	if (i != -20)
	{
		beginerror();
		nc_puts(_ShowEscape(STR1, i, j, "During shell escape"));
		enderror();
	}
	ENDTRY(try22);
}

static void dumphistory()
{
	log_htrec *ht;
	log_hnrec *hn;
	short i, FORLIM;

	m_graphics_off();
	clearalpha();
	m_alpha_on();
	hn = gg.hnbase;
	i = 0;
	nc_printf("Time               ");
	while (hn != NULL)
	{
		nc_printf("%s%*c", gg.signaltab[hn->sig - 1].name, (int)(15 - strlen(gg.signaltab[hn->sig - 1].name)), ' ');
		hn = hn->next;
		i++;
	}
	nc_putchar('\n');
	ht = gg.htbase;
	while (ht != NULL)
	{
		nc_printf("% .5E   ", ht->time);
		FORLIM = hncount;
		for (i = 0; i < FORLIM; i++)
			nc_printf("%12.6f   ", na_srtor(ht->val[i]));
		nc_putchar('\n');
		ht = ht->next;
	}
	nc_printf("Press any key to continue.\n");
	waitforkey();
	gg.showpage = 0;
}


static void reshuffle()
{
	log_hnrec *hn;
	short i;

	i = 0;
	hn = gg.hnbase;
	while (hn != NULL)
	{
		i++;
		hn->num = i;
		hn = hn->next;
	}
	hncount = i;
	histreset = true;
}


void histdelsignals()
{
	log_hnrec *hn;
	log_tool *tool;

	while (gg.hnbase != NULL)
	{
		tool = gg.signaltab[gg.hnbase->sig - 1].np->simtype;
		disposeattrs(&gg.hnbase->attr, tool->hnumattrs, tool->hattr);
		hn = gg.hnbase->next;
		Free(gg.hnbase);
		gg.hnbase = hn;
	}
	histgridhn = NULL;
	if (histgridwhich == 2)
		histgridwhich = 0;
	reshuffle();
}


#define unit            (histdivision / 2)


static void alignsigname(short *y)
{
	if (histgattr[histalignsigs - 1].UU.b)
		*y = histdown - (histdown - *y + unit / 2) / unit * unit;
}

#undef unit


static short histaddsignalpos()
{
	log_hnrec *hn;
	short y;

	y = histdivision / 2;
	do
	{
		hn = gg.hnbase;
		while (hn != NULL && (hn->y + histvscale < y - histvscale ||
					hn->y - histvscale > y + histvscale))
			hn = hn->next;
		y += histdivision;
	} while (hn != NULL && y <= histdown);
	y -= histdivision;
	alignsigname(&y);
	return y;
}


void histaddsignal(log_hnrec **hn, short sig, short y)
{
	log_tool *tool;

	*hn = (log_hnrec *)Malloc(sizeof(log_hnrec));
	(*hn)->y = y;
	(*hn)->sig = sig;
	tool = gg.signaltab[sig - 1].np->simtype;
	newattrs(&(*hn)->attr, tool->hnumattrs, tool->hattr);
	if (gg.traceflag)
		fprintf(tracefile, "Histaddsignal: %ld, %d\n",
				(long)(*hn)->attr, tool->hnumattrs);
	(*hn)->next = gg.hnbase;
	gg.hnbase = *hn;
	reshuffle();
}


static void hnproc1()
{
	calltool(gg.acttool, act_confighist);
}


static void hnproc2()
{
	calltool(gg.acttool, act_configchhist);
}


static void hnproc3()
{
	calltool(gg.acttool, act_configrelhist);
}


static void haproc1()
{
	log_gattrrec *WITH;

	WITH = &histgattr[histcurcount - 1];
	if (WITH->UU.U73.i1 != htcount)
	{
		WITH->UU.U73.i1 = htcount;
		WITH->blnk = false;
		WITH->changed = true;
	}
	WITH = &histgattr[histcurtime - 1];
	if (WITH->UU.r != gg.time)
	{
		WITH->UU.r = gg.time;
		WITH->blnk = false;
		WITH->changed = true;
	}
	WITH = &histgattr[histcurtstep - 1];
	if (WITH->UU.r == gg.prevtimestep)
		return;
	WITH->UU.r = gg.prevtimestep;
	WITH->blnk = false;
	WITH->changed = true;
}


static void haproc2()
{
	log_gattrrec *WITH;

	WITH = &histgattr[gg.actx - 1];
	switch (gg.actx)
	{

		case histtrigsig:
			if (!strcmp(WITH->UU.c, "(none)"))
				histtrig = 0;
			else
				histtrig = getsignal(0, WITH->UU.c);
			break;

		case histstarttime:
			if (WITH->UU.r <= 0)
				WITH->blnk = true;

			if (!WITH->blnk && WITH->UU.r >= histgattr[histstoptime - 1].UU.r)
			{
				histgattr[histstoptime - 1].blnk = true;
				histgattr[histstoptime - 1].changed = true;
			}
			break;

		case histstoptime:
			if (WITH->UU.r < 0)
				WITH->blnk = true;

			if (!WITH->blnk && WITH->UU.r <= histgattr[histstarttime - 1].UU.r)
			{
				histgattr[histstarttime - 1].blnk = true;
				histgattr[histstarttime - 1].changed = true;
			}
			break;

		case histminstep:
		case histmaxstep:
			if (WITH->UU.r <= 0)
				WITH->blnk = true;
			break;

		case histsecdiv:
			if (WITH->UU.r <= 0)
				gg.actflag = false;
			break;

		case histmaxcount:
			if (WITH->UU.U73.i1 < 2)
				WITH->UU.U73.i1 = 2;
			break;

		case histcurcount:
		case histcurtime:
		case histcurtstep:
			gg.actflag = false;
			break;
	}
}


static void haproc3()
{
	/* nothing at all! */
}


#define quit1pos        2

#define configpos       (quit1pos + 44)
#define triggerpos      (configpos + 63)
#define resetpos        (triggerpos + 51)
#define onoffpos        (resetpos + 45)
#define gridpos         (onoffpos + 26)
#define gridwpos        (gridpos + 60)
#define fastpos         (gridwpos + 54)
#define dumppos         (fastpos + 40)
#define nextpos         (dumppos + 40)

#define quit2pos        484

#define otherpos        (quit2pos - 50)

#define axiswidth       50

#define xaxispos        (quit1pos + 44)
#define yaxispos        (xaxispos + axiswidth)
#define exprpos         (yaxispos + axiswidth + 20)
#define timepos         (exprpos + 65)
#define plotpos         (timepos + 80)
#define next2pos        (plotpos + 50)

#define nummenus        2


#define tolerance       15


/* Local variables for historycommand: */
struct LOC_historycommand {
	int rflag, oldtrigger, oldreset, oldonoff, oldfast;
	short oldgridmode, oldgridwhich, gridx1, gridy1, gridx2, gridy2, valuey,
		  whichmenu, vmenu1, vmenu2;
	int xactive, yactive;
	char xexpr[256], yexpr[256];
	double gridval, gridval2, gridtime;
	char gridstr[81];
	long stamp;
} ;

static double valueattime(log_hnrec *hn, double time, struct LOC_historycommand *LINK)
{
	log_htrec *ht, *ht0;
	double v1, v2, alpha;

	ht = gg.htbase;
	ht0 = NULL;
	while (ht != NULL && ht->time < time)
	{
		ht0 = ht;
		ht = ht->next;
	}

	if (ht == NULL || ht0 == NULL)
	{
		return 0.0;
	}
	else
	{
		v1 = na_srtor(ht0->val[hn->num - 1]);
		v2 = na_srtor(ht->val[hn->num - 1]);
		alpha = (time - ht0->time) / (ht->time - ht0->time);
		return (v1 + (v2 - v1) * alpha);
	}
}

static log_hnrec *findtrace(short x, short y, struct LOC_historycommand *LINK)
{
	log_htrec *ht, *ht0;
	double v1, v2, value, alpha, time;
	log_hnrec *hn, *besthn;
	short yy, besty;

	time = (double)(x - histleft) / histdivision * histgattr[histsecdiv - 1].UU.r + histgattr[histfirsttime - 1].UU.r;
	ht = gg.htbase;
	ht0 = NULL;
	while (ht != NULL && ht->time < time)
	{
		ht0 = ht;
		ht = ht->next;
	}

	if (ht == NULL || ht0 == NULL)
	{
		return NULL;
	}
	else
	{
		alpha = (time - ht0->time) / (ht->time - ht0->time);
		hn = gg.hnbase;
		besty = 9999;
		while (hn != NULL)
		{
			v1 = na_srtor(ht0->val[hn->num - 1]);
			v2 = na_srtor(ht->val[hn->num - 1]);
			value = v1 + (v2 - v1) * alpha;
			historyvaluey(hn, value, &v1, &yy);
			if (abs(yy - y) < besty)
			{
				besty = abs(yy - y);
				besthn = hn;
			}
			hn = hn->next;
		}

		if (besty < tolerance)
			return besthn;
		else
			return NULL;
	}
}

#undef tolerance

static void refrtrigger(struct LOC_historycommand *LINK)
{
	if (gg.histactive == LINK->oldtrigger || LINK->whichmenu != 0)
		return;
	remcursor();
	clipoff();
	if (gg.histactive)
		m_color((long)gg.color.selword);
	else
		m_color((long)gg.color.menuword);
	drawstr2((int)triggerpos, LINK->vmenu1, "Trigger");
	LINK->oldtrigger = gg.histactive;
}

static void refrreset(struct LOC_historycommand *LINK)
{
	if (gg.resetflag == LINK->oldreset || LINK->whichmenu != 0)
		return;
	remcursor();
	clipoff();
	if (gg.resetflag)
		m_color((long)gg.color.selword);
	else
		m_color((long)gg.color.menuword);
	drawstr2((int)resetpos, LINK->vmenu1, "Reset");
	LINK->oldreset = gg.resetflag;
}

static void refronoff(struct LOC_historycommand *LINK)
{
	if (gg.pwrflag == LINK->oldonoff || LINK->whichmenu != 0)
		return;
	remcursor();
	clipoff();
	if (gg.pwrflag)
	{
		m_color((long)gg.color.selword);
		drawstr2((int)onoffpos, LINK->vmenu1, "ON ");
	}
	else
	{
		m_color((long)gg.color.menuword);
		drawstr2((int)onoffpos, LINK->vmenu1, "OFF");
	}
	LINK->oldonoff = gg.pwrflag;
}

static void refrgridmode(struct LOC_historycommand *LINK)
{
	if (histgridmode != LINK->oldgridmode && LINK->whichmenu == 0)
	{
		remcursor();
		clipoff();
		m_color((long)gg.color.selword);
		switch (histgridmode)
		{

			case 0:
				drawstr2((int)gridpos, LINK->vmenu1, "Delta:   ");
				break;

			case 1:
				drawstr2((int)gridpos, LINK->vmenu1, "Absolute:");
				break;

			case 2:
				drawstr2((int)gridpos, LINK->vmenu1, "Value:   ");
				break;

			case 3:
				drawstr2((int)gridpos, LINK->vmenu1, "Slope:   ");
				break;
		}
		LINK->oldgridmode = histgridmode;
	}

	if (histgridwhich == LINK->oldgridwhich)
		return;
	remcursor();
	clipoff();
	switch (histgridwhich)
	{

		case 0:
			m_color((long)gg.color.selword);
			drawstr2((int)gridwpos, LINK->vmenu1, "Time  ");
			break;

		case 1:
			m_color((long)gg.color.selword);
			drawstr2((int)gridwpos, LINK->vmenu1, "Freq  ");
			break;

		case 2:
			m_color((long)gg.color.menuword);
			drawstr2((int)gridwpos, LINK->vmenu1, "Signal");
			break;
	}
	LINK->oldgridwhich = histgridwhich;
}

static void refrfast(struct LOC_historycommand *LINK)
{
	int fast;

	fast = (gg.fastspeed == gg.fastmax);
	if (fast == LINK->oldfast || LINK->whichmenu != 0)
		return;
	remcursor();
	clipoff();
	if (fast)
		m_color((long)gg.color.selword);
	else
		m_color((long)gg.color.menuword);
	drawstr2((int)fastpos, LINK->vmenu1, "Fast");
	LINK->oldfast = fast;
}

static void drawsigname(log_hnrec *hn, short opt, struct LOC_historycommand *LINK)
{
	remcursor();
	clipoff();
	if (opt == 3)
		m_colormode((long)m_xor);
	else
		m_colormode((long)m_normal);

	if (opt == 2)
		m_color((long)gg.color.backgr);
	else if (hn == histgridhn && opt != 1)
		m_color((long)gg.color.selword);
	else
		m_color((long)gg.color.signal);

	drawstr2(0, hn->y - 4, gg.signaltab[hn->sig - 1].name);
	m_colormode((long)m_normal);
}

/* Local variables for drawhistory: */
struct LOC_drawhistory {
	struct LOC_historycommand *LINK;
	short x, y, oldx;
	log_htrec *ht;
	short *ar;
} ;

static void drawtrace(log_hnrec *hn, short i, struct LOC_drawhistory *LINK)
{
	historypointy(hn, LINK->ht, &LINK->y);
	if ((unsigned long)gg.acty > 15)
	{
		LINK->ar[i - 1] = log_maxshint;
		return;
	}
	m_color(gg.acty);
	if (LINK->ar[i - 1] != log_maxshint)
	{
		m_move((long)LINK->oldx, (long)LINK->ar[i - 1]);
		m_draw((long)LINK->x, (long)LINK->y);
	}
	LINK->ar[i - 1] = LINK->y;
}

static void drawhistory(struct LOC_historycommand *LINK)
{
	struct LOC_drawhistory V;
	short i, gridi;
	log_hnrec *hn;
	log_htrec *ht1;
	short FORLIM;

	V.LINK = LINK;
	hn = gg.hnbase;
	while (hn != NULL)
	{
		if (histgridhn != hn)
			drawsigname(hn, 0, LINK);
		hn = hn->next;
	}

	if (histgridhn != NULL)
		drawsigname(histgridhn, 0, LINK);

	if (gg.htbase == NULL)
		return;

	V.ar = (short *)Malloc(hncount * 2);
	FORLIM = hncount;
	for (i = 0; i < FORLIM; i++)
		V.ar[i] = log_maxshint;

	m_clip((long)histleft, 0L, (long)across, (long)histdown);
	V.ht = gg.htbase;
	ht1 = gg.htbase;

	while (ht1 != NULL && ht1->time < histgattr[histfirsttime - 1].UU.r)
	{
		V.ht = ht1;
		ht1 = ht1->next;
	}

	do
	{
		historypointx(V.ht, &V.x);
		hn = gg.hnbase;
		i = 0;
		while (hn != NULL)
		{
			i++;
			if (histgridhn != hn)
				drawtrace(hn, i, &V);
			else
				gridi = i;
			hn = hn->next;
		}
		if (histgridhn != NULL)
			drawtrace(histgridhn, gridi, &V);
		V.oldx = V.x;
		V.ht = V.ht->next;
	} while (V.ht != NULL && V.x <= across);
	m_noclip();
	Free(V.ar);
}

static void setgridwhich(log_hnrec *hn, struct LOC_historycommand *LINK)
{
	if (histgridhn != NULL)
		drawsigname(histgridhn, 1, LINK);
	histgridhn = hn;
	if (hn != NULL)
		histgridwhich = 2;
	else if (histgridwhich == 2)
		histgridwhich = 0;
	if (histgridhn != NULL)
		drawsigname(histgridhn, 0, LINK);
}

static void nextgridwhich(struct LOC_historycommand *LINK)
{
	log_hnrec *nexthn;

	nexthn = histgridhn;
	if (nexthn != NULL)
		drawsigname(nexthn, 1, LINK);
	switch (histgridwhich)
	{

		case 0:
			histgridwhich = 1;
			break;

		case 1:
			nexthn = gg.hnbase;
			if (nexthn != NULL)
				histgridwhich = 2;
			else
				histgridwhich = 0;
			break;

		case 2:
			nexthn = nexthn->next;
			if (nexthn == NULL)
				histgridwhich = 0;
			break;
	}
	histgridhn = nexthn;
	if (histgridhn != NULL)
		drawsigname(histgridhn, 0, LINK);
}

static void nextgridmode(struct LOC_historycommand *LINK)
{
	if (histgridmode == 0)
		histgridmode = 3;
	else
		histgridmode--;
}

static void confighistmode(struct LOC_historycommand *LINK)
{
	int flag;
	log_hnrec *hn;
	void (*TEMP) ();
	void (*TEMP5) ();
	void (*TEMP6) ();

	flag = histgattr[histalignsigs - 1].UU.b;
	TEMP = haproc1;
	TEMP5 = haproc2;
	TEMP6 = haproc3;
	editattrs(histgattr, histnumattrs, histkattr, histlbl, "Scope", TEMP, TEMP5,
			TEMP6, &LINK->stamp);
	if (flag)
		return;
	hn = gg.hnbase;
	while (hn != NULL)
	{
		alignsigname(&hn->y);
		hn = hn->next;
	}
}

static void passetc(struct LOC_historycommand *LINK)
{
	m_noclip();
	refrtrigger(LINK);
	refrreset(LINK);
	refronoff(LINK);
	refrfast(LINK);
	pass();
	pen();
}

static void fastmode(struct LOC_historycommand *LINK)
{
	long t0;

	t0 = timers_sysclock();
	do {
		gg.fastspeed = gg.fastmax;
		passetc(LINK);
	} while (gg.t.near_ && timers_sysclock() <= t0 + 50);
	gg.fastspeed = gg.fastmax;
}

static void dumpcmd(struct LOC_historycommand *LINK)
{
	char fn[256];
	log_htrec *ht;
	log_hnrec *hn;
	short i;
	long j;
	char STR2[256], STR3[256];
	short FORLIM;

	if (gg.htbase == NULL)
		return;

	m_color((long)gg.color.selword);
	drawstr2((int)dumppos, LINK->vmenu1, "Dump");
	beginbottom();
	m_alpha_on();
	nc_printf("Dump file name");

	if (*dumpfname != '\0')
		nc_printf(" [%s]", dumpfname);
	nc_printf(": ");
	readlnpass(fn, 0);
	endbottom();
	beginbottom();
	m_alpha_on();

	if (*fn != '\0')
		setdumpname(fn);
	else
		nc_printf("Dump file is %s\n", dumpfname);

	if (*dumpfname != '\0') {
		TRY(try23);

		/* Procedure to write the scope data to the dump file.

		   GG.HNBASE points to a linked list of History Names.  HN is a temp. variable.
		   HNCOUNT is a count of the number of things in this list.
		   The name of a trace is given by:  gg.signaltab^[hn^.sig].name^

		   GG.HTBASE points to a list of History Timesteps.  HT is a temp. variable.
		   For each entry, the time is stored in:  ht^.time
		   and the value of the I'th history trace is:  na_srtor(ht^.val[i])
		   */
		dumpmessage("");
		sprintf(STR2, "%s (", dumpfname);
		dumpmessage(STR2);   /*file name in header*/
		dumpmessage("(DATA:");
		dumpmessage("(TITLE: TIME )");
		dumpmessage("(POINTS:");
		ht = gg.htbase;

		/* list of time values */
		while (ht != NULL)
		{
			sprintf(STR2, "%g", ht->time);
			dumpmessage(STR2);
			ht = ht->next;
		}
		dumpmessage("))");
		hn = gg.hnbase;
		FORLIM = hncount;

		/* for each trace... */
		for (i = 0; i < FORLIM; i++)
		{
			dumpmessage("(DATA:");   /*write its name*/
			sprintf(STR3, "(TITLE: \\\\%s\\)", gg.signaltab[hn->sig - 1].name);
			dumpmessage(STR3);
			dumpmessage("(POINTS:");
			ht = gg.htbase;

			/* and list of values */
			while (ht != NULL)
			{
				sprintf(STR3, "%g", na_srtor(ht->val[i]));
				dumpmessage(STR3);
				ht = ht->next;
			}
			dumpmessage("))");
			hn = hn->next;
		}
		dumpmessage(")");
		/* End of Dump routine.*/

		closedump();
		RECOVER(try23);
		if (P_escapecode == -20)
			_Escape(P_escapecode);
		i = P_ioresult;
		j = P_escapecode;
		beginerror();
		if (Debugging || debugprint)
			nc_printf("%ld/%d/%ld   ", j, i, EXCP_LINE);
		nc_printf("Unable to write dump file \"%s\"\n", dumpfname);
		enderror();
		ENDTRY(try23);
	}
	endbottom();
	LINK->rflag = true;
}

static void refraxis(long pos, char *name, char *expr, int flag, struct LOC_historycommand *LINK)
{
	if (flag)
		m_color((long)gg.color.selword);
	else
		m_color((long)gg.color.menuword);

	drawstr2((int)pos, LINK->vmenu1, name);
	uerase((int)pos, LINK->vmenu2 - 2, (int)(pos + axiswidth - 3),
			LINK->vmenu2 + 12);
	m_clip(pos, LINK->vmenu2 - 2L, pos + axiswidth - 3, LINK->vmenu2 + 12L);
	m_color((long)gg.color.menuword);
	drawstr2((int)pos, LINK->vmenu2, expr);
	m_noclip();
}

static void refraxes(struct LOC_historycommand *LINK)
{
	refraxis(xaxispos, "X-axis", LINK->xexpr, LINK->xactive, LINK);
	refraxis(yaxispos, "Y-axis", LINK->yexpr, LINK->yactive, LINK);
}

static void setaxis(char *expr, struct LOC_historycommand *LINK)
{
	if (LINK->xactive)
	{
		LINK->xactive = false;
		strcpy(LINK->xexpr, expr);
		if (*LINK->yexpr == '\0')
			LINK->yactive = true;
		refraxes(LINK);
		return;
	}

	if (!LINK->yactive)
		return;

	LINK->yactive = false;
	strcpy(LINK->yexpr, expr);
	if (*LINK->xexpr == '\0')
		LINK->xactive = true;
	refraxes(LINK);
}

static void getexpr(struct LOC_historycommand *LINK)
{
	char expr[256];

	if (!(LINK->xactive || LINK->yactive))
		return;
	beginbottom();
	m_alpha_on();
	nc_printf("Expression to plot: ");
	readlnpass(expr, 0);
	endbottom();
	strcompress(expr, " ", true);
	if (*expr != '\0')
		setaxis(expr, LINK);
}

static void plotcmd(struct LOC_historycommand *LINK)
{
	char STR1[256];

	if (*LINK->xexpr == '\0' || *LINK->yexpr == '\0')
		return;
	sprintf(STR1, "PLOT SCOPE %s %s", LINK->xexpr, LINK->yexpr);
	assertfunc(STR1);
	dofunction();
	clearfunc();
	LINK->rflag = true;
}

static void measurecmd(struct LOC_historycommand *LINK)
{
	char STR1[81];

	do
	{
		while ((gg.t.x < histleft || gg.t.y > histdown) && gg.t.depressed)
			pen();
		LINK->gridx2 = gg.t.x;
		LINK->gridy2 = gg.t.y;
		m_colormode((long)m_xor);
		m_color((long)gg.color.scroll);
		*LINK->gridstr = '\0';
		if (histgridwhich == 2)
		{
			gg.actx = histgridhn->sig;
			gg.actgattr = histgridhn->attr;
		}
		*gg.actstr = '\0';
		gg.acty = histgridmode;
		switch (histgridwhich * 100 + histgridmode)
		{

			case 0:
				m_drawline((long)LINK->gridx1, 0L, (long)LINK->gridx1, (long)histdown);
				m_drawline((long)LINK->gridx2, 0L, (long)LINK->gridx2, (long)histdown);
				gg.actval = (double)abs(LINK->gridx2 - LINK->gridx1) / histdivision;
				realunit(LINK->gridstr, gg.actval * histgattr[histsecdiv - 1].UU.r, 4, "s", true);
				break;

			case 1:
			case 101:
			case 2:
			case 102:
				m_drawline((long)LINK->gridx2, 0L, (long)LINK->gridx2, (long)histdown);
				gg.actval = (double)(LINK->gridx2 - histleft) / histdivision;
				realunit(LINK->gridstr, gg.actval * histgattr[histsecdiv - 1].UU.r +
						histgattr[histfirsttime - 1].UU.r, 4, "s", true);
				break;

			case 3:
			case 103:
				m_drawline((long)LINK->gridx1, (long)LINK->gridy1, (long)LINK->gridx2,
						(long)LINK->gridy2);
				break;

			case 100:
				m_drawline((long)LINK->gridx1, 0L, (long)LINK->gridx1, (long)histdown);
				m_drawline((long)LINK->gridx2, 0L, (long)LINK->gridx2, (long)histdown);
				gg.actval = (double)abs(LINK->gridx2 - LINK->gridx1) / histdivision;
				if (gg.actval != 0)
					realunit(LINK->gridstr,
							1 / (gg.actval * histgattr[histsecdiv - 1].UU.r), 4, "Hz", true);
				break;

			case 200:
				m_drawline((long)histleft, (long)LINK->gridy1, (long)across, (long)LINK->gridy1);
				m_drawline((long)histleft, (long)LINK->gridy2, (long)across, (long)LINK->gridy2);
				gg.actval2 = (double)LINK->gridy1 / histdivision;
				gg.actval3 = (double)LINK->gridy2 / histdivision;
				gg.actval = fabs(gg.actval3 - gg.actval2);
				calltoolnode(gg.signaltab[gg.actx - 1].np, act_histstr);

				if (*gg.actstr == '\0')
					realunit(LINK->gridstr, gg.actval, 4, "", true);
				else
					strcpy(LINK->gridstr, gg.actstr);
				break;

			case 201:
				m_drawline((long)histleft, (long)LINK->gridy2, (long)across, (long)LINK->gridy2);
				gg.actval = (double)(histgridhn->y - LINK->gridy2) / histdivision;
				calltoolnode(gg.signaltab[gg.actx - 1].np, act_histstr);

				if (*gg.actstr == '\0')
					realunit(LINK->gridstr, gg.actval, 4, "", true);
				else
					strcpy(LINK->gridstr, gg.actstr);
				break;

			case 202:
				LINK->gridtime =
					(double)(LINK->gridx2 - histleft) / histdivision *
					histgattr[histsecdiv - 1].UU.r + histgattr[histfirsttime - 1].UU.r;
				LINK->gridval2 = valueattime(histgridhn, LINK->gridtime, LINK);
				historyvaluey(histgridhn, LINK->gridval2, &LINK->gridval, &LINK->valuey);
				m_drawline((long)LINK->gridx2, 0L, (long)LINK->gridx2, (long)histdown);
				m_drawline((long)histleft, (long)LINK->valuey, (long)across, (long)LINK->valuey);
				gg.actval = LINK->gridval;
				gg.actval2 = LINK->gridval2;
				gg.actval3 = LINK->gridtime;
				*gg.actstr = '\0';
				gg.acty = histgridmode;
				gg.actx = histgridhn->sig;
				gg.actgattr = histgridhn->attr;
				calltoolnode(gg.signaltab[gg.actx - 1].np, act_histstr);

				if (*gg.actstr == '\0')
					strcpy(gg.actstr, realunit(STR1, gg.actval, 4, "", true));
				sprintf(LINK->gridstr, "%s  ->  %s", realunit(STR1, LINK->gridtime, 4, "s", true), gg.actstr);
				break;

			case 203:
				m_drawline((long)LINK->gridx1, (long)LINK->gridy1, (long)LINK->gridx2, (long)LINK->gridy2);
				if (LINK->gridx1 != LINK->gridx2)
				{
					gg.actval2 = (LINK->gridx2 - LINK->gridx1) * histgattr[histsecdiv - 1].UU.r;
					gg.actval = (LINK->gridy1 - LINK->gridy2) / gg.actval2;
					gg.actval2 /= histdivision;
					gg.actval3 = (double)LINK->gridy1 / histdivision;
					calltoolnode(gg.signaltab[gg.actx - 1].np, act_histstr);

					if (*gg.actstr == '\0')
						realunit(LINK->gridstr, gg.actval, 4, "", true);
					else
						strcpy(LINK->gridstr, gg.actstr);
					strcat(LINK->gridstr, "/s");
				}
				break;
		}

		m_color((long)gg.color.selword);
		m_colormode((long)m_normal);
		clipoff();
		drawstr2((int)gridpos, LINK->vmenu2, LINK->gridstr);

		do
		{
			passetc(LINK);
		} while (!gg.t.moving);

		m_colormode((long)m_xor);
		m_color((long)gg.color.selword);
		clipoff();
		drawstr2((int)gridpos, LINK->vmenu2, LINK->gridstr);
		m_color((long)gg.color.scroll);

		switch (histgridwhich * 100 + histgridmode)
		{

			case 0:
			case 100:
				m_drawline((long)LINK->gridx1, 0L, (long)LINK->gridx1, (long)histdown);
				m_drawline((long)LINK->gridx2, 0L, (long)LINK->gridx2, (long)histdown);
				break;

			case 1:
			case 101:
			case 2:
			case 102:
				m_drawline((long)LINK->gridx2, 0L, (long)LINK->gridx2, (long)histdown);
				break;

			case 3:
			case 103:
			case 203:
				m_drawline((long)LINK->gridx1, (long)LINK->gridy1, (long)LINK->gridx2,
						(long)LINK->gridy2);
				break;

			case 200:
				m_drawline((long)histleft, (long)LINK->gridy1, (long)across,
						(long)LINK->gridy1);
				m_drawline((long)histleft, (long)LINK->gridy2, (long)across,
						(long)LINK->gridy2);
				break;

			case 201:
				m_drawline((long)histleft, (long)LINK->gridy2, (long)across,
						(long)LINK->gridy2);
				break;

			case 202:
				m_drawline((long)LINK->gridx2, 0L, (long)LINK->gridx2, (long)histdown);
				m_drawline((long)histleft, (long)LINK->valuey, (long)across,
						(long)LINK->valuey);
				break;
		}
		m_colormode((long)m_normal);
	} while (gg.t.depressed);
}

/// Enter History (Scope) display mode.
static void historycommand()
{
	struct LOC_historycommand V;
	short y, i;
	long t;
	log_hnrec *hn, *hn0, *hn1, *hn2, *hn1a, *hn2a, *movinghn;
	uchar ch;
	char name[256];
	log_tool *tool;
	int flag, rmflag, keyflag;
	short scrollpos, scrollmin;
	double newhistfirsttime;
	int gridflag;
	void (*TEMP) ();
	void (*TEMP5) ();
	void (*TEMP6) ();
	char STR2[256];
	char STR3[256];
	char TEMP7;

	remcursor();
	clipoff();
	V.vmenu1 = histdown + 5;
	V.vmenu2 = V.vmenu1 + 11;
	gridflag = false;
	*V.xexpr = '\0';
	*V.yexpr = '\0';
	V.xactive = false;
	V.yactive = true;
	V.whichmenu = 0;
	V.rflag = true;
	rmflag = true;
	hn = gg.hnbase;
	while (hn != NULL)
	{
		if (hn->y > histdown)
			hn->y = histaddsignalpos();
		hn = hn->next;
	}
	do
	{
		if (V.rflag)
		{
			clearwindow();
			movinghn = NULL;
			pass();
			gg.showpage = log_page_history;
			rmflag = true;
		}

		if (rmflag)
		{
			uerase(0, histdown + 1, across, down);
			m_color((long)gg.color.menuword);
			drawstr2(quit1pos, V.vmenu1, "QUIT");
			drawstr2(quit2pos, V.vmenu1, "QUIT");
			switch (V.whichmenu)
			{

				case 0:
					drawstr2((int)configpos, V.vmenu1, "Configure");
					V.oldfast = (gg.fastspeed != gg.fastmax);
					refrfast(&V);
					m_color((long)gg.color.menuword);
					drawstr2((int)dumppos, V.vmenu1, "Dump");
					drawstr2((int)otherpos, V.vmenu1, "PLOT");
					V.oldtrigger = !gg.histactive;
					refrtrigger(&V);
					V.oldreset = !gg.resetflag;
					refrreset(&V);
					V.oldonoff = !gg.pwrflag;
					refronoff(&V);
					V.oldgridmode = histgridmode - 1;
					V.oldgridwhich = histgridwhich - 1;
					refrgridmode(&V);
					break;

				case 1:
					drawstr2((int)exprpos, V.vmenu1, "Expression");
					drawstr2((int)timepos, V.vmenu1, "Time");
					drawstr2((int)plotpos, V.vmenu1, "Plot");
					drawstr2((int)otherpos, V.vmenu1, "SCOPE");
					refraxes(&V);
					break;
			}
		}
		if (V.rflag)
		{
			drawhistdivisions(histleft, across);
			drawhistory(&V);
		}

		V.rflag = false;
		rmflag = false;
		refrgridmode(&V);
		ch = '\0';
		do
		{
			if (histgattr[histsweeptype - 1].UU.nv == histsweep_contin && gg.histactive)
			{
				passetc(&V);
				flag = (gg.time < histgattr[histfirsttime - 1].UU.r + histdivsacross * histgattr[histsecdiv - 1].UU.r);
				if (!flag && histonscreen) {
					histgattr[histfirsttime - 1].UU.r = gg.time - histdivsacross * histgattr[histsecdiv - 1].UU.r * 0.1;
					ch = ' ';
				}
				histonscreen = flag;
			} else {
				passetc(&V);
				histonscreen = false;
			}
			if (pollkbd2())
				ch = inkey2();
		} while (!(gg.t.dn || ch != '\0'));

		if (gg.t.dn)
		{
			remcursor();
			clipoff();
			V.gridx1 = gg.t.x;
			V.gridy1 = gg.t.y;
			if (gg.t.y > histdown)
			{
				if (gg.t.x < configpos - 3)
					ch = '\003';
				else if (gg.t.x >= quit2pos - 3)
					ch = '\003';
				else if (gg.t.x >= otherpos - 3) {
					V.whichmenu = (V.whichmenu + 1) & (nummenus - 1);
					rmflag = true;
				}
				else
				{
					switch (V.whichmenu)
					{

						case 0:
							if (gg.t.x < triggerpos - 3)
							{
								confighistmode(&V);
								V.rflag = true;
							}
							else if (gg.t.x < resetpos - 3)
							{
								if (gg.histactive)
								{
									gg.histactive = false;
								}
								else
								{
									gg.histactive = true;
									histreset = true;
								}
							}
							else if (gg.t.x < onoffpos - 3)
							{
								gg.resetflag = true;
							}
							else if (gg.t.x < gridpos - 3)
							{
								gg.pwrflag = !gg.pwrflag;
							}
							else if (gg.t.x < gridwpos - 3)
							{
								nextgridmode(&V);
							}
							else if (gg.t.x < fastpos - 3)
							{
								nextgridwhich(&V);
							}
							else if (gg.t.x < dumppos - 3)
							{
								fastmode(&V);
							}
							else if (gg.t.x < nextpos - 3)
							{
								dumpcmd(&V);
							}
							break;

						case 1:
							if (gg.t.x < yaxispos - 3)
							{
								V.xactive = !V.xactive;
								V.yactive = false;
								refraxes(&V);
							}
							else if (gg.t.x < exprpos - 3)
							{
								V.xactive = false;
								V.yactive = !V.yactive;
								refraxes(&V);
							}
							else if (gg.t.x < timepos - 3)
							{
								getexpr(&V);
							}
							else if (gg.t.x < plotpos - 3)
							{
								setaxis("Time", &V);
							}
							else if (gg.t.x < next2pos - 3)
							{
								plotcmd(&V);
							}
							break;
					}
				}
			}
			else if (gg.t.x < histleft - 3)
			{
				hn = gg.hnbase;
				hn0 = NULL;
				hn1 = NULL;
				hn2 = NULL;
				hn1a = NULL;
				hn2a = NULL;
				while (hn != NULL)
				{
					if (hn->y > gg.t.y - histvscale && hn->y < gg.t.y + histvscale)
					{
						if (histgridhn == hn)
						{
							hn2 = hn;
							hn2a = hn0;
						}
						else
						{
							hn1 = hn;
							hn1a = hn0;
						}
					}
					hn0 = hn;
					hn = hn->next;
				}

				if (hn2 != NULL)
				{
					hn = hn2;
					hn0 = hn2a;
				}
				else
				{
					hn = hn1;
					hn0 = hn1a;
				}

				if (hn != NULL)
				{
					switch (V.whichmenu)
					{

						case 0:
							if (justtap())
							{
								gg.actnode = gg.signaltab[hn->sig - 1].np;
								gg.acty = hn->sig;
								gg.acttool = gg.actnode->simtype;
								TEMP = hnproc1;
								TEMP5 = hnproc2;
								TEMP6 = hnproc3;
								editattrs(hn->attr, gg.acttool->hnumattrs, gg.acttool->hattr,
										gg.acttool->hlbl, gg.signaltab[hn->sig - 1].name,
										TEMP, TEMP5, TEMP6, &V.stamp);
								V.rflag = true;
							}
							else
							{
								movinghn = hn;
								drawsigname(hn, 2, &V);
								do
								{
									hn->y = gg.t.y;
									flag = (gg.t.y > 0 && gg.t.y < histdown + 3);
									alignsigname(&hn->y);
									if (gg.t.y >= histdown)
										gg.t.y = histdown;
									if (flag)
										drawsigname(hn, 3, &V);
									do
									{
										passetc(&V);
									} while (gg.t.depressed && gg.t.y == gg.t0.y);
									if (flag)
										drawsigname(hn, 3, &V);
								} while (gg.t.depressed);

								if (!flag)
								{
									if (gg.traceflag)
										fprintf(tracefile, "Delete trace %ld\n", (long)hn);
									if (hn == histgridhn) {
										histgridwhich = 0;
										histgridhn = NULL;
									}
									tool = gg.signaltab[hn->sig - 1].np->simtype;
									disposeattrs(&hn->attr, tool->hnumattrs, tool->hattr);
									if (hn0 == NULL)
										gg.hnbase = hn->next;
									else
										hn0->next = hn->next;
									Free(hn);
									reshuffle();
								}
								V.rflag = true;
							}
							break;

						case 1:
							setaxis(gg.signaltab[hn->sig - 1].name, &V);
							break;
					}
				}
			}
			else
			{
				switch (V.whichmenu)
				{

					case 0:
						if (justtap())
						{
							hn = findtrace(V.gridx1, V.gridy1, &V);
							if (hn != NULL)
								setgridwhich(hn, &V);
						}
						else
						{
							measurecmd(&V);
						}
						break;

					case 1:
						hn = findtrace((int)gg.t.x, (int)gg.t.y, &V);
						if (hn != NULL)
							setaxis(gg.signaltab[hn->sig - 1].name, &V);
						break;
				}
			}
		} else if (ch == 251 || ch == 250 || ch == ' ')
		{
			V.rflag = true;
		}
		else if (ch == '<')
		{
			do
			{
				histgattr[histsecdiv - 1].UU.r *= 2;
				t = timers_sysclock() + knobwait;
				do
				{
				} while (!(timers_sysclock() > t || pollkbd2()));

				if (testkey2() == '<')
					ch = inkey2();
				else
					V.rflag = true;
			} while (!V.rflag);
		}
		else if (ch == '>')
		{
			do
			{
				histgattr[histsecdiv - 1].UU.r /= 2;
				t = timers_sysclock() + knobwait;
				do
				{
				} while (!(timers_sysclock() > t || pollkbd2()));
				if (testkey2() == '>')
					ch = inkey2();
				else
					V.rflag = true;
			} while (!V.rflag);
		}
		else if (ch < 32 && ((1L << ch) & 0x10000100L) != 0)
		{
			scrollpos = 0;
			gg.actval = -histgattr[histfirsttime - 1].UU.r / histgattr[histsecdiv - 1].UU.r;

			if (gg.actval < -histdivsacross)
				scrollmin = -across * 2;
			else
				scrollmin = (long)floor(gg.actval * histdivision + 0.5);

			do
			{
				if (ch == '\b')
					scrollpos -= hscrollspd * 5;
				if (ch == '\034')
					scrollpos += hscrollspd * 5;
				if (scrollpos < scrollmin)
					scrollpos = scrollmin;
				m_colormode((long)m_xor);
				m_color((long)gg.color.scroll);
				if (scrollpos > 0)
					m_drawline(scrollpos + histleft - 1L, 0L, scrollpos + histleft - 1L, (long)histdown);
				m_drawline(scrollpos + across + 1L, 0L, scrollpos + across + 1L, (long)histdown);
				m_colormode((long)m_normal);
				t = timers_sysclock() + knobwait;
				do
				{
				} while (!(timers_sysclock() > t || pollkbd2()));
				TEMP7 = testkey2();

				if ((uchar)TEMP7 < 32 && ((1L << TEMP7) & 0x10000100L) != 0)
					ch = inkey2();
				else
					ch = ' ';

				m_colormode((long)m_xor);
				m_color((long)gg.color.scroll);
				if (scrollpos > 0)
					m_drawline(scrollpos + histleft - 1L, 0L, scrollpos + histleft - 1L, (long)histdown);
				m_drawline(scrollpos + across + 1L, 0L, scrollpos + across + 1L, (long)histdown);
				m_colormode((long)m_normal);
			} while (ch < 32 && ((1L << ch) & 0x10000100L) != 0 && abs(scrollpos) <= across - histleft - 10);

			if (scrollpos == scrollmin)
				newhistfirsttime = 0.0;
			else
				newhistfirsttime = histgattr[histfirsttime - 1].UU.r +
					(double)scrollpos / histdivision * histgattr[histsecdiv - 1].UU.r;

			if (histgattr[histfirsttime - 1].UU.r != newhistfirsttime)
			{
				histgattr[histfirsttime - 1].UU.r = newhistfirsttime;
				histgattr[histfirsttime - 1].changed = true;
				V.rflag = true;
			}
			histonscreen = false;
		}
		else if (ch == '!')
		{
			doshellescape("");
			V.rflag = true;
		}
		else if (ch > ' ')
		{
			switch (V.whichmenu)
			{

				case 0:
					y = histaddsignalpos();
					*name = '\0';
					keyflag = true;
					do
					{
						strcat(name, "_");
						remcursor();
						clipoff();
						m_colormode((long)m_xor);
						m_color((long)gg.color.signal);
						drawstr2(0, y - 4, name);
						m_colormode((long)m_normal);
						while (!gg.t.dn && !keyflag)
						{
							passetc(&V);
							if (pollkbd2())
							{
								ch = inkey2();
								keyflag = true;
							}
						}
						m_colormode((long)m_xor);
						m_color((long)gg.color.signal);
						clipoff();
						drawstr2(0, y - 4, name);
						m_colormode((long)m_normal);
						name[strlen(name) - 1] = '\0';
						if (keyflag)
						{
							keyflag = false;
							if (ch > ' ' || (ch == ' ' && *name != '\0'))
								sprintf(name + strlen(name), "%c", ch);
							if (ch == '\007' && *name != '\0')
								name[strlen(name) - 1] = '\0';
						}
					} while (!((ch < 32 && ((1L << ch) & 0x2008) != 0) || gg.t.dn));
					sprintf(STR2, "%c", ch);

					if (!strcmp(STR2, EXEC))
						*name = '\0';
					else
						strcpy(name, strrtrim(strcpy(STR3, name)));

					remcursor();
					clipoff();
					i = getsignal(0, name);
					if (i != 0)
					{
						histaddsignal(&hn, i, y);
						V.rflag = true;
					}
					break;

				case 1:
					if (V.xactive || V.yactive)
					{
						ungetkey2(realkey);
						getexpr(&V);
					}
					break;
			}
		}
	} while (strcmp((sprintf(STR2, "%c", ch), STR2), EXEC));
}

#undef quit1pos
#undef configpos
#undef triggerpos
#undef resetpos
#undef onoffpos
#undef gridpos
#undef gridwpos
#undef fastpos
#undef dumppos
#undef nextpos
#undef quit2pos
#undef otherpos
#undef axiswidth
#undef xaxispos
#undef yaxispos
#undef exprpos
#undef timepos
#undef plotpos
#undef next2pos
#undef nummenus

static void savecommand()
{
	short pgnum;
	char filename[256];
	short FORLIM;
	char STR1[256];
	char STR2[256];

	garbagecoll();
	if (*gg.funcarg == '\0')
	{
		FORLIM = gg.numpages;
		for (pgnum = 1; pgnum <= FORLIM; pgnum++)
		{
			if (!isPageChanged(gg.pages[pgnum - 1]))
			{
				sprintf(STR2, "No changes to page %d", pgnum);
				message(STR2);
			}
			else if (curfilename[pgnum - 1] == NULL)
			{
				sprintf(STR2, "Page %d: no file name!", pgnum);
				message(STR2);
			}
			else
			{
				sprintf(STR1, "Saving page %d", pgnum);
				message(STR1);
				safesavepage(pgnum, curfilename[pgnum - 1]);
			}
		}
	}
	else if (!strcmp(gg.funcarg, "*"))
	{
		beginbottom();
		nc_printf("File name to save: ");
		if (curfilename[gg.curpage - 1] != NULL)
			strcpy(filename, curfilename[gg.curpage - 1]);
		else
			*filename = '\0';
		readlnpass(filename, 3);
		endbottom();
		if (*filename != '\0')
			safesavepage((int)gg.curpage, filename);
	}
	else
	{
		safesavepage((int)gg.curpage, gg.funcarg);
	}
	clearfunc();
}


static void namecommand()
{
	char filename[256];
	char STR2[256], STR3[256];

	if (*gg.funcarg == '\0') {
		beginbottom();
		nc_printf("File name for page %ld", gg.curpage);
		if (curfilename[gg.curpage - 1] != NULL)
			nc_printf(" [%s]", curfilename[gg.curpage - 1]);
		nc_printf(": ");
		readlnpass(filename, 0);
		endbottom();
	} else if (!strcmp(gg.funcarg, "?"))
		*filename = '\0';
	else
		strcpy(filename, gg.funcarg);
	if (*filename != '\0') {
		if (!strcmp(filename, "\"\"") || *strcpy(STR3, strltrim(filename)) == '\0')
			Free(curfilename[gg.curpage - 1]);
		else {
			newci_fixfname(filename, "lgf", "");
			if (curfilename[gg.curpage - 1] == NULL)
				curfilename[gg.curpage - 1] = (char *)Malloc(256);
			strcpy(curfilename[gg.curpage - 1], filename);
		}
	}
	if (curfilename[gg.curpage - 1] == NULL) {
		sprintf(STR2, "No file name for page %ld", gg.curpage);
		message(STR2);
	} else {
		sprintf(STR2, "Current file name is %s", curfilename[gg.curpage - 1]);
		message(STR2);
	}
	clearfunc();
}

#define catsize         21
#define maxdirmax       45
#define maxdircol       8

/// Load circuit page from disk. Previous contents of current page are lost.
static void loadcommand()
{
	short i, j, x, y, x1, y1, dirmax, dircol;
	char filename[256];
	char ch;
	char name[fidleng + 1];
	char (*dirs[maxdircol + 1])[fidleng + 1];
	char STR3[256];

	dircol = P_imin2((txacross - 14L) / 15, (long)maxdircol);
	dirmax = -1;
	for (i = 0; i <= dircol; i++)
		dirs[i] = NULL;
	if (!strcmp(gg.funcarg, "*"))
	{
		beginbottom();
		nc_printf("Name of file to load: ");
		readlnpass(filename, 0);
		endbottom();
	}
	else if (*gg.funcarg == '\0')
	{
		clearshowalpha();
		nc_gotoXY(5, txdown - 1);
		nc_printf("Press ENTER alone to abort command.\n");
		nc_gotoXY(2, txdown - 2);
		nc_printf("Name of file to load: ");
		*filename = '\0';
		ch = '\015';
		do
		{
			do
			{
				x = (gg.t.ax - 2) / 15;
				y = gg.t.ay - 2;
				if (x < 0 || (unsigned)y > dirmax || x > dircol || dirs[x] == NULL || *dirs[x][y] == '\0')
				{
					x = -1;
				}
				else
				{
					nc_gotoXY(x * 15 + 2, y + 2);
					nc_printf("%c%s%c", chrinverse, dirs[x][y], chrplain);
				}

				if (gg.t.near_ && gg.t.inalpha)
					nc_gotoXY(gg.t.ax, gg.t.ay);
				else
					nc_gotoXY((int)(strlen(filename) + 24), txdown - 2);

				nc_cursor_on();
				do
				{
					pass();
					pen();
					x1 = (gg.t.ax - 2) / 15;
					y1 = gg.t.ay - 2;
					/* x1 := (gg.t.x-22) div 90;
y1 := (gg.t.y-46) div 15;
if gg.t.x < 22 then x1 := 0;
if gg.t.y < 46 then y1 := 0; */
					if (y1 < 0 || x1 < 0 || y1 > dirmax || x1 > dircol || dirs[x1] == NULL || *dirs[x1][y1] == '\0')
						x1 = -1;
				} while (!(pollkbd2() || gg.t.dn || x != x1 || y != y1) && gg.t.near_);
				nc_cursor_off();
				remcursor();
				if (x != -1)
				{
					nc_gotoXY(x * 15 + 2, y + 2);
					nc_puts(dirs[x][y]);
				}
			} while (!(pollkbd2() || gg.t.dn));

			if (gg.t.dn && x != -1)
			{
				strcpy(filename, dirs[x][y]);
				nc_gotoXY(24, txdown - 2);
				nc_printf("%s\t", filename);
			}

			if (gg.t.dn)
				ch = '\015';
			else
				ch = inkey2();

			if (ch == '\007' && *filename != '\0')
			{
				nc_gotoXY((int)(strlen(filename) + 23), txdown - 2);
				nc_putchar(' ');
				filename[strlen(filename) - 1] = '\0';
			}
			else if (ch >= ' ' && ch <= '~' && strlen(filename) < 255)
			{
				nc_gotoXY((int)(strlen(filename) + 24), txdown - 2);
				nc_putchar(ch);
				strcat(filename, " ");
				filename[strlen(filename) - 1] = ch;
			}
		} while ((uchar)ch >= 32 || ((1L << ch) & 0x2008) == 0);

		for (i = 0; i <= dircol; i++)
		{
			if (dirs[i] != NULL)
				Free(dirs[i]);
		}

		sprintf(STR3, "%c", ch);
		if (!strcmp(STR3, EXEC))
			*filename = '\0';
		clearalpha();

		if (m_autoraise)
		{
		}
		else
		{
			m_graphics_on();
		}
	}
	else
	{
		strcpy(filename, gg.funcarg);
	}

	if (*filename != '\0')
	{
		beginbottom();
		TRY(try27);
		nc_printf("Loading file %s\n", filename);
		m_alpha_on();
		loadpage(filename, name);
		endbottom();
		RECOVER(try27);
		i = P_escapecode;
		j = P_ioresult;
		endbottom();
		beginerror();
		if (Debugging || debugprint || gg.traceflag)
			nc_printf("%d/%d/%ld  ", i, j, EXCP_LINE);
		if (i == -10 && (unsigned)j < 32 && ((1L << j) & 0x600) != 0)
			nc_printf("Can't find file \"%s\"\n", filename);
		else if (i == -20)
			nc_printf("STOP key pressed while loading file.\n");
		else if (i > 0)
			nc_printf("Format error in file \"%s\" (%s)\n", filename, name);
		else
			nc_printf("Unable to load file \"%s\"\n", filename);
		enderror();
		ENDTRY(try27);
	}
	clearfunc();
	refrwindow();
}

#undef catsize
#undef maxdirmax
#undef maxdircol

static void readcommand()
{
	char filename[256], reason[256];
	long i, j;

	if (*gg.funcarg == '\0' || !strcmp(gg.funcarg, "*"))
	{
		beginbottom();
		nc_printf("Name of file to read: ");
		readlnpass(filename, 0);
		endbottom();
	}
	else
	{
		strcpy(filename, gg.funcarg);
	}

	if (*filename != '\0')
	{
		beginbottom();
		nc_printf("Reading file %s\n", filename);
		deleverything();
		garbagecoll();
		TRY(try28);
		m_alpha_on();
		readpage(filename, reason);
		endbottom();
		RECOVER(try28);
		i = P_escapecode;
		j = P_ioresult;
		endbottom();
		beginerror();
		if (Debugging || debugprint || gg.traceflag)
			nc_printf("%ld/%ld/%ld  ", i, j, EXCP_LINE);

		if (i == -10 && (unsigned long)j < 32 && ((1L << j) & 0x600) != 0)
			nc_printf("Can't find file \"%s\"\n", filename);
		else if (i == -20)
			nc_printf("STOP key pressed while reading file.\n");
		else if (i > 0)
			nc_printf("Format error in file \"%s\" (%s)\n", filename, reason);
		else
			nc_printf("Unable to read file \"%s\"\n", filename);
		enderror();
		ENDTRY(try28);
	}
	clearfunc();
}


static void singlestep()
{
	gg.singlestepcount = getint(gg.funcarg, 1L);
	steppingoff = gg.pwrflag;
	gg.pwrflag = true;
	gg.refrflag = true;
}


static void getgatecmd()
{
	short i;
	char name[9];

	do
	{
		getword(gg.funcarg, name);
		if (*name != '\0')
			i = readlibrary(name);
	} while (*name != '\0');
	clearfunc();
}


static void makegatecmd()
{
	short i;
	char name[9];

	do
	{
		getword(gg.funcarg, name);
		if (*name != '\0') {
			i = readlibrary(name);
			if (i != 0)
				uaddgate(0, 0, i);
		}
	} while (*name != '\0');
	clearfunc();
}


#define maxtools        23


/* Local variables for toolcommand: */
struct LOC_toolcommand {
	log_tool *ta[maxtools + 1];
	int refrflag, wasselected;
} ;

static void describe(long i, struct LOC_toolcommand *LINK)
{
	if (LINK->ta[i] == NULL)
		return;
	remcursor();
	nc_gotoXY(4, (int)i);
	if (LINK->ta[i]->ready)
		nc_putchar(chryellow);
	nc_printf("%16s   ", LINK->ta[i]->name);
	nc_puts(LINK->ta[i]->comment);
	nc_printf("%c\n", chrgreen);
}

static long which(struct LOC_toolcommand *LINK)
{
	if (gg.t.near_ && gg.t.ay >= 1 && gg.t.ay <= maxtools)
		return (gg.t.ay);
	else
		return -1;
}

static void selecttool(char *name, struct LOC_toolcommand *LINK)
{
	log_tool *tp;

	tp = findtool(name);
	if (!tp->ready)
	{
		return;
	}

	calltool(tp, act_select);
	if (!tp->keep)
		closetool(tp);
	LINK->refrflag = true;
	LINK->wasselected = true;
}

static void toolcommand()
{
	struct LOC_toolcommand V;
	log_tool *tp;
	long i, j, numtools;
	int exitflag;
	char ch;

	V.wasselected = false;
	if (*gg.funcarg != '\0')
	{
		selecttool(gg.funcarg, &V);
	}
	else
	{
		do
		{
			for (i = 0; i <= maxtools; i++)
				V.ta[i] = NULL;
			numtools = 0;
			tp = gg.toolbase;
			while (tp != NULL)
			{
				if (*tp->comment != '\0' && numtools < maxtools)
				{
					numtools++;
					V.ta[numtools + 1] = tp;
				}
				tp = tp->next;
			}
			clearshowalpha();
			nc_printf("LOG tools available:\n");
			for (i = 0; i <= maxtools; i++)
				describe(i, &V);
			exitflag = false;
			V.refrflag = false;
			do
			{
				do
				{
					j = which(&V);
					if (j >= 0)
					{
						nc_putchar(chrinverse);
						describe(j, &V);
						nc_putchar(chrplain);
					}

					if (gg.t.near_ && gg.t.inalpha)
						nc_gotoXY(gg.t.ax, gg.t.ay);
					else
						noblink();

					do
					{
						pass();
						pen();
					} while (!(which(&V) != j || gg.t.dn || pollkbd2()));

					if (j >= 0)
						describe(j, &V);
				}
				while (!(gg.t.dn || pollkbd2()));
				nc_gotoXY(0, txdown);

				if (pollkbd2())
					ch = toupper(inkey2());
				else
					ch = '\0';

				if (gg.t.dn)
				{
					switch (j)
					{

						case -1:
							ch = '\003';
							break;

						default:
							ch = (char)(j + 64);
							if (V.ta[ch - 64] == NULL)
								ch = '\003';
							break;
					}
				}

				if (ch >= 'A' && ch <= 'X')
				{
					tp = V.ta[ch - 64];
					if (tp != NULL)
					{
						selecttool(tp->name, &V);
						exitflag = true;
					}
				}
				else if (ch == '\003' || ch == '\015' || ch == ' ')
				{
					exitflag = true;
				}
			} while (!(V.refrflag || exitflag));
		} while (!exitflag);
	}
	clearalpha();
	if (V.wasselected)
	{
		m_graphics_on();
		refresh();
	}
	m_graphics_on();
}

#undef maxtools


#define whichmin        (-2)


static log_tool *counttool(long i)
{
	log_tool *tp;

	tp = gg.toolbase;
	while (tp->nostatus)
		tp = tp->next;
	while (i > 1)
	{
		tp = tp->next;
		if (!tp->nostatus)
			i--;
	}
	return tp;
}

static char *onoff(char *Result, int b)
{
	if (b)
		return strcpy(Result, "ON ");
	else
		return strcpy(Result, "OFF");
}

static void status_log()
{
	timerec time;
	daterec date;
	char STR2[9];
	char STR3[81];
	char * months[12] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

	if (gg.actflag)
	{
		sysdate(&date);
		nc_printf("   Date %2d-%s-%02d   Time\n",
				date.day, months[date.month - 1], date.year);
		nc_printf("   Available memory:   \n\n");
		nc_printf("   Simulation time:    \n");
		nc_printf("   Simulation timestep:\n");
		nc_printf("   Simulation status:  \n\n\n");
		nc_printf("   Invisible mode %s", onoff(STR2, gg.invisible));
		nc_printf("     Invisible labels %s", onoff(STR2, gg.textinvisible));
		nc_printf("   Glow mode %s\n", onoff(STR2, gg.glowmode));
		nc_printf("   Simulation %s", onoff(STR2, gg.pwrflag));
		nc_printf("         Snap-to-grid %s", onoff(STR2, snapflag));
		nc_printf("       Markers %s\n", onoff(STR2, gg.markers));
		nc_printf("   Quiet mode %s", onoff(STR2, gg.quiet));
		nc_printf("         Verbose mode %s", onoff(STR2, gg.verbose));
		nc_printf("       Trace mode %s\n", onoff(STR2, gg.traceflag));
		nc_gotoXY(0, txdown - 3);
		nc_printf("Use the arrow keys to view status pages.  Press the space bar to quit.\n");
		status_oldtime = -1.0;
		status_oldtstep = -1.0;
		status_oldmem = -1;
	}

	if (labs(timers_sysclock() - gg.actx) > 50)
	{
		nc_gotoXY(25, 4);
		systime(&time);
		nc_printf("%2d:%02d:%02ld", time.hour, time.minute, time.centisecond / 100L);
		gg.actx = timers_sysclock();
	}

	if (gg.time != status_oldtime)
	{
		nc_gotoXY(23, 7);
		nc_printf("%s\t", realunit(STR3, gg.time, 4, "sec", false));
		status_oldtime = gg.time;
	}

	if (gg.prevtimestep != status_oldtstep)
	{
		nc_gotoXY(23, 8);
		nc_printf("%s\t", realunit(STR3, gg.prevtimestep, 4, "sec", false));
		status_oldtstep = gg.prevtimestep;
	}

	if (gg.simstate == gg.oldsimstate && gg.simstatetool == gg.oldsimstatetool)
		return;
	nc_gotoXY(23, 9);

	switch (gg.simstate)
	{

		case simst_off:
			nc_printf("Off\t");
			break;

		case simst_notactive:
			nc_printf("Not active\t");
			break;

		case simst_notready:
			nc_printf("Halted (%s)\t", gg.simstatetool->shortname);
			break;

		case simst_running:
			if (gg.simstatetool == NULL)
				nc_printf("Running (scope-mode interpolation)\t");
			else
				nc_printf("Running (%s)\t", gg.simstatetool->shortname);
			break;

		default:
			break;
	}
}


static void status_mem()
{
	long i, j;
	log_grec *g;
	log_hwrec *hw;
	log_vwrec *vw;
	log_lrec *l;
	log_brec *b;
	log_nrec *n;
	long FORLIM;

	if (!gg.actflag)
		return;

	nc_printf("          gates  wires labels  boxes  changed  filename\n");
	for (i = 0; i <= 8; i++)
	{
		if (i + 1 == gg.curpage)
			nc_putchar(chryellow);
		nc_printf("Page %ld -- ", i + 1);
		if (i + 1 <= gg.numpages)
		{
			g = gg.pages[i]->gbase;
			j = 0;
			while (g != NULL)
			{
				j++;
				g = g->next;
			}
			nc_printf("%4ld", j);
			hw = gg.pages[i]->hwbase;
			j = 0;
			while (hw != NULL)
			{
				j++;
				hw = hw->next;
			}
			vw = gg.pages[i]->vwbase;
			while (vw != NULL)
			{
				j++;
				vw = vw->next;
			}
			nc_printf("%7ld", j);
			l = gg.pages[i]->lbase;
			j = 0;
			while (l != NULL)
			{
				j++;
				l = l->next;
			}
			nc_printf("%7ld", j);
			b = gg.pages[i]->bbase;
			j = 0;
			while (b != NULL)
			{
				j++;
				b = b->next;
			}
			nc_printf("%7ld     ", j);

			if (isPageChanged(gg.pages[i]))
				nc_printf("Yes    ");
			else
				nc_printf("No     ");

			if (curfilename[i] != NULL)
				nc_puts(curfilename[i]);
		}
		else
		{
			nc_printf("Empty.");
		}
		nc_printf("%c\n", chrgreen);
	}
	nc_printf("\nPaste  -- ");
	if (copybuf.valid)
	{
		g = copybuf.gcopy;
		j = 0;
		while (g != NULL)
		{
			j++;
			g = g->next;
		}
		nc_printf("%4ld", j);
		hw = copybuf.hwcopy;
		j = 0;
		while (hw != NULL)
		{
			j++;
			hw = hw->next;
		}
		vw = copybuf.vwcopy;
		while (vw != NULL)
		{
			j++;
			vw = vw->next;
		}
		nc_printf("%7ld", j);
		l = copybuf.lcopy;
		j = 0;
		while (l != NULL)
		{
			j++;
			l = l->next;
		}
		nc_printf("%7ld", j);
		b = copybuf.bcopy;
		j = 0;
		while (b != NULL)
		{
			j++;
			b = b->next;
		}
		nc_printf("%7ld\n", j);
	}
	else
	{
		nc_printf("Empty.\n");
	}
	nc_putchar('\n');
	n = gg.nbase;
	j = 0;
	while (n != NULL)
	{
		j++;
		n = n->next;
	}
	nc_printf("Total nodes:         %12ld\n", j);
	j = 0;
	FORLIM = gg.lastsignal;
	for (i = 0; i < FORLIM; i++)
	{
		if (gg.signaltab[i].np != NULL)
			j++;
	}
	nc_printf("Total signal names:  %12ld\n", j);
}

static void status_macro(int v)
{
	macrorec *mp;
	char ch;
	long i, j;
	int half;
	int k;
	static int n;
	int n_displayed;
	int n_macros;
	int n_to_print;

	if (!gg.actflag)
		return;
	for (k=0,mp = macrobase; mp != NULL; mp = mp->next, k++)
		;
	n_macros = k;
	n_displayed = ((txacross / 40)+1) * (txdown -4);

	half = k / 2;
	if (k & 1)
		half++;
	if (v > 0)
		n += (txdown -4);
	else if (v < 0)
		n -= (txdown -4);
	if (n > half)
		n -= (txdown -4);
	if (n < 0)
		n = 0;
	mp = macrobase;
	i = 2;
	n_to_print = P_min(half-n,txdown-4);
	do
	{
		j = 4;
		for (k=0; k < n && mp != NULL; k++)
			mp = mp->next;
		do
		{
			if (mp != NULL)
			{
				nc_gotoXY((int)i, (int)j);
				ch = mp->key;
				if (ch == '\007')
					nc_printf("bs ");
				else if (ch == '\t')
					nc_printf("tab");
				else if (ch == '\015')
					nc_printf("cr ");
				else if (ch == ' ')
					nc_printf("sp ");
				else if ((ch & 255) >= 168 && (ch & 255) <= 193)
					nc_printf("^%c ", (char)((uchar)ch - 'g'));
				else if (ch <= '\037' || (ch & (~127)) != 0)
					nc_printf("%3d", ch);
				else
					nc_printf("%c  ", ch);
				nc_printf("  %s", mp->name);
				mp = mp->next;
			}
			j++;
		} while (j < txdown && j < n_to_print+4 && mp != NULL);

		if (i < 40)
		{
			for (k = 0, mp = macrobase; mp != NULL && k < half ; k++)
				mp = mp->next;
		}
		i += 40;
	} while (i <= txacross && mp != NULL);

	for (k=0; mp != NULL; k++, mp = mp->next)
		;
	nc_gotoXY((int)2, (int)txdown);
	nc_printf("[%d of %d]",((n+1)/(n_displayed/2))+1,(n_macros/n_displayed)+1);

}

static void statusdisplay(char *name_)
{
	char name[256];
	char ch;
	short which, whichmax;
	long savex, savey;
	int refrflag, exitflag;
	log_tool *tp;
	char STR2[256], STR3[256];
	int n = 0;

	strcpy(name, name_);
	clearshowalpha();
	whichmax = 0;
	tp = gg.toolbase;
	while (tp != NULL)
	{
		if (!tp->nostatus)
			whichmax++;
		tp = tp->next;
	}
	strupper(name, strcpy(STR2, strltrim(strrtrim(strcpy(STR3, name)))));

	if (!strcmp(name, "LOG"))
		which = -2;
	else if (!strcmp(name, "MEMORY"))
		which = -1;
	else if (!strcmp(name, "MACRO") || !strcmp(name, "KEYBOARD"))
		which = 0;
	else
	{
		tp = gg.toolbase;
		which = 1;
		while (tp != NULL && strcmp(tp->name, name))
		{
			if (!tp->nostatus)
				which++;
			tp = tp->next;
		}
		if (tp == NULL)
			which = whichmin;
	}

	exitflag = false;
	do
	{
		clearshowalpha();
		nc_printf("                           LOG V%s    Status Display\n\n",
				log_version);
		nc_printf("Status:  ");
		switch (which)
		{

			case -2:
				nc_printf("LOG");
				break;

			case -1:
				nc_printf("Memory");
				break;

			case 0:
				nc_printf("Keyboard assignments");
				break;

			default:
				tp = counttool((long)which);
				nc_puts(tp->name);
				if (*tp->comment != '\0')
					nc_printf(" (%s)", tp->comment);
				break;
		}
		savex = 0;
		savey = 0;
		refrflag = false;
		gg.actflag = true;
		nc_gotoXY(0, 4);
		do
		{
			gg.actx = savex;
			gg.acty = savey;
			switch (which)
			{

				case -2:
					status_log();
					break;

				case -1:
					status_mem();
					break;

				case 0:
					status_macro(n);
					n = 0;
					break;

				default:
					if (tp->ready)
						calltool(tp, act_status);
					if (gg.actflag) {
						if (tp->ready)
							nc_printf("   (No status for this tool)\n");
						else
							nc_printf("   (Not yet activated)\n");
					}
					break;
			}
			savex = gg.actx;
			savey = gg.acty;
			pen();
			pass();
			ch = '\0';
			if (gg.t.dn)
				ch = '\003';
			if (pollkbd2())
				ch = inkey2();
			switch (ch)
			{
				case 31:
					n = -1;
					refrflag = true;
					break;

				case 10:
					n = 1;
					refrflag = true;
					break;

				case '+':
				case '\034':
					if (which < whichmax)
					{
						which++;
						refrflag = true;
					}
					break;

				case '-':
				case '\b':
					if (which > whichmin)
					{
						which--;
						refrflag = true;
					}
					break;

				case ' ':
				case '\003':
				case '\015':
					exitflag = true;
					break;
			}
			gg.actflag = false;
			nc_gotoXY(9, 2);
		} while (!(refrflag || exitflag));
	} while (!exitflag);
	if (m_autoraise)
	{
		m_graphics_on();  /* fix from WES */
	}
	clearalpha();
}

#undef whichmin


#define cols            4


/** @brief Process a pop-up menu selection.

  Menu items are named in MENU.
  Use result to look up in table S to find a char.
  */
static void popupmenu(short num)
{
	short i, j, x, y, xx, x0, y0, x1, y1, x00, y00;
	logmenurec *menu;
	char STR1[256];

	menu = popupmenus[num - 1];
	popup_grid = (cursortype == grid);
	remcursor();
	clipoff();
	m_color((long)gg.color.backgr);
	m_fillrect(kindgroupleft + 1L, (long)baseline, kindgroupright - 1L,
			(long)down);
	m_color((long)gg.color.popupbox);
	m_drawrect(kindgroupleft + 3L, (long)baseline, kindgroupright - 3L,
			down - 2L);
	m_drawrect(kindgroupleft + 5L, baseline + 2L, kindgroupright - 5L,
			down - 4L);
	xx = (kindgroupsize * kindgroupspacing - 10) / (cols + 1);
	x = across / 2 - xx * (cols - 1) / 2;
	x0 = x;
	for (i = 1; i <= 4; i++)
	{
		y = line1 + 2;
		for (j = i * 2 - 2; j < i * 2; j++)
		{
			if (*menu[j].name != '\0') {
				if (menu[j].bool_ != NULL && *menu[j].bool_)
					m_color((long)gg.color.selword);
				else if (menu[j].name[0] == '*')
					m_color((long)gg.color.redword);
				else if (menu[j].name[0] == '#')
					m_color((long)gg.color.blueword);
				else
					m_color((long)gg.color.popupword);

				if (menu[j].name[0] == '#' || menu[j].name[0] == '*')
				{
					sprintf(STR1, "%.*s",
							(int)(strlen(menu[j].name) - 1L), menu[j].name + 1);
					m_centerstr((long)x0, (long)y, logfont_lfont, STR1);
				}
				else
				{
					m_centerstr((long)x0, (long)y, logfont_lfont, menu[j].name);
				}
			}
			y += line2 - line1 - 4;
		}
		x0 += xx;
	}
	do
	{
		x0 = (gg.t.x - x + xx / 2) / xx;
		y0 = (gg.t.y - baseline) / 26;
		if (gg.t.x + xx / 2 < x || gg.incircuit || x0 > 3 || y0 > 1)
		{
			x0 = -1;
			y0 = 0;
		}
		else
		{
			x00 = x + x0 * xx - xx / 2;
			y00 = line1 + y0 * (line2 - line1 - 4) - 1;
			m_color((long)gg.color.popupsel);
			m_drawrect((long)x00, (long)y00, x00 + xx - 1L, y00 + 15L);
		}
		do
		{
			pass();
			pollkbd2();
			pen();
			x1 = (gg.t.x - x + xx / 2) / xx;
			y1 = (gg.t.y - baseline) / 26;
			if (gg.t.x + xx / 2 < x || gg.incircuit || x1 > 3 || y1 > 1)
			{
				x1 = -1;
				y1 = 0;
			}
		} while (gg.t.depressed && x0 == x1 && y0 == y1);
		remcursor();
		clipoff();
		if (x0 != -1) {
			m_color((long)gg.color.backgr);
			m_drawrect((long)x00, (long)y00, x00 + xx - 1L, y00 + 15L);
		}
	} while (gg.t.depressed);
	/*  m_putcpicture(kindgroupleft+1, baseline, menupicture);  */
	refrmenu();
	if (x0 == -1)
		return;
	i = x0 * 2 + 1;
	if (y0 != 0)
		i++;
	if (*menu[i - 1].name == '\0')
		return;
	immedscroll = true;
	assertfunc(menu[i - 1].cmd);
	immedscroll = false;
	scroll();
}

#undef cols


static short reducecolor8(short c)
{
	short Result;

	if (!eightcolors)
		return c;

	switch (c)
	{

		case log_mred:
			Result = log_red;
			break;

		case log_orange:
			Result = log_red;
			break;

		case log_black:
			Result = log_black8;
			break;

		case log_pink:
			Result = log_red;
			break;

		case log_dyellow:
			Result = log_yellow;
			break;

		case log_xgreen:
			Result = log_green;
			break;

		case log_dcyan:
			Result = log_black8;
			break;

		case log_dred:
			Result = log_black8;
			break;

		case log_lgray:
			Result = log_white;
			break;

		case log_cred:
			Result = log_cred8;
			break;

		default:
			Result = c;
			break;
	}
	return Result;
}



static void setcmap(short n, short r, short g, short b)
{
	n = reducecolor8(n);
	rcolormap[n] = r;
	gcolormap[n] = g;
	bcolormap[n] = b;
	m_vsetcolors((long)n, 1L, &rcolormap[n], &gcolormap[n], &bcolormap[n]);
}


static void getcolor(char *name_, short *val, short default_color)
{
	char name[256];
	strlist_t *cp;
	char STR2[256];
	char STR3[256];

	strcpy(name, name_);
	strupper(name, strcpy(STR2, strltrim(strrtrim(strcpy(STR3, name)))));
	cp = strlist_find(colorbase, name);
	if (cp != NULL)
		*val = (long)cp->value;
	else
		*val = reducecolor8(default_color);
}

static void initcolors()
{
	getcolor("BACKGR", &gg.color.backgr, log_gray);   /* Color of background */
	getcolor("CURSOR", &gg.color.cursor, log_white);   /* Color of cursors */
	recolor_log_cursors(gg.color.cursor, false);
	getcolor("GATE", &gg.color.gate, log_cyan);   /* Color of gate in circuit */
	getcolor("DIMGATE", &gg.color.dimgate, log_dcyan);
	/* Color of dimmed gate */
	getcolor("KINDGATE", &gg.color.kindgate, log_green);
	/* Color of gate in menu area */
	getcolor("CATGATE", &gg.color.catgate, log_cyan);
	/* Color of gate in catalog */
	getcolor("GATEPIN", &gg.color.gatepin, log_red);
	/* Color of pins on gates */
	getcolor("PINNUM", &gg.color.pinnum, log_yellow);
	/* Color of pin numbers on gates */
	getcolor("WIRE", gg.color.wire, log_green);
	getcolor("VLSIGREEN", &gg.color.wire[log_wcol_green - log_wcol_normal],
			log_green);
	getcolor("VLSIYELLOW", &gg.color.wire[log_wcol_yellow - log_wcol_normal],
			log_yellow);
	getcolor("VLSIRED", &gg.color.wire[log_wcol_red - log_wcol_normal], log_red);
	getcolor("VLSIBLUE", &gg.color.wire[log_wcol_blue - log_wcol_normal],
			log_cyan);
	getcolor("XWIRE", &gg.color.xwire, log_xgreen);
	/* Color of wires while moving */
	getcolor("SOLDER", &gg.color.solder, log_green);
	/* Color of solder point in circuit */
	getcolor("CONFLICT", &gg.color.conflict, log_cred);
	/* Color of wire during node conflict */
	getcolor("MENUWORD", &gg.color.menuword, log_green);
	/* Color of words in menu area */
	getcolor("SELWORD", &gg.color.selword, log_yellow);
	/* Color of highlighted words */
	getcolor("MESSAGE", &gg.color.message, log_yellow);
	/* Color of messages */
	getcolor("BASELINE", &gg.color.baseline, log_cyan);
	/* Color of baseline (default) */
	getcolor("PAGE1", &gg.color.page1, log_green);   /* Color of word "PAGE" */
	getcolor("PAGE2", &gg.color.page2, log_yellow);   /* Color of page number */
	getcolor("PAGE3", &gg.color.page3, log_green);   /* Color of word "OF" */
	getcolor("PAGE4", &gg.color.page4, log_yellow);
	/* Color of number of pages */
	getcolor("SELECT", &gg.color.select, log_white);
	/* Color of selected areas */
	getcolor("SCROLL", &gg.color.scroll, log_white);
	/* Color of scroll feedback */
	getcolor("SIGNAL", &gg.color.signal, log_pink);
	/* Color of TO/FROM node name */
	getcolor("LABELTEXT", &gg.color.labeltext, log_dyellow);
	/* Color of text labels */
	getcolor("DASHBOX", &gg.color.dashbox, log_dyellow);
	/* Color of dashed boxes */
	getcolor("MARKER", &gg.color.marker, log_pink);
	/* Color of printing markers */
	getcolor("CHART", &gg.color.chart, log_yellow);
	/* Color of node history trace */
	getcolor("DIVISION", &gg.color.division, log_white);
	/* Color of scope divisions */
	getcolor("POPUPBOX", &gg.color.popupbox, log_white);
	/* Color of pop-up menu outline */
	getcolor("POPUPWORD", &gg.color.popupword, log_green);
	/* Color of text in pop-up menus */
	getcolor("POPUPSEL", &gg.color.popupsel, log_pink);
	/* Color of selection in pop-up menus */
	getcolor("REDWORD", &gg.color.redword, log_pink);
	/* Color of special menu items */
	getcolor("BLUEWORD", &gg.color.blueword, log_cyan);
	/* Color of experimental menu items */
	getcolor("VLSICUT", &gg.color.vlsicut, log_white);
	/* Color of VLSI contact cuts */
	getcolor("GATEGREEN", &gategreen, log_green);
	getcolor("GATEYELLOW", &gateyellow, log_yellow);
	getcolor("GATERED", &gatered, log_red);
	getcolor("GATEWHITE", &gatewhite, log_white);
	getcolor("GATEBLACK", &gateblack, log_black);
	getcolor("GATEORANGE", &gateorange, log_orange);
	getcolor("DEFINEBOX", &defineboxcolor, log_cyan);
	/* Color of DEFINE command's box */
	getcolor("DEFINEBACK", &definebackcolor, log_black);
	/* Color of background of DEFINE box */
	getcolor("DEFINETEXT", &definetextcolor, log_green);
	/* Color of text in DEFINE box */
	getcolor("CATALOGBOX", &catboxcolor, log_dyellow);
	/* Color of Catalog grouping box */
	calltools(act_color);
	stamp(&gg.colorstamp);
}


static void echomacro(char key)
{
	macrorec *mp;
	char name[256];
	char STR1[256];

	if (isupper(key))
		sprintf(name, "Shift-%c key", key);
	else if (key > ' ' && (key & (~127)) == 0)
		sprintf(name, "%c key", key);
	else if (key == ' ')
		strcpy(name, "Space bar");
	else if (key == '\007')
		strcpy(name, "Backspace key");
	else if (key == '\t')
		strcpy(name, "Tab key");
	else if (key == '\015')
		strcpy(name, "Enter key");
	else
		sprintf(name, "Key #%d", key);
	mp = macrobase;
	while (mp != NULL && mp->key != key)
		mp = mp->next;

	if (mp == NULL)
	{
		sprintf(STR1, "%s is not defined.", name);
		message(STR1);
	} else
	{
		sprintf(STR1, "%s is defined as \"%s\"", name, mp->name);
		message(STR1);
	}
}


static void definemacro(char key, char *name)
{
	macrorec *current, *last, *new;

	last = NULL;
	current = macrobase;
	while (current != NULL && current->key < key)
	{
		last = current;
		current = current->next;
	}

	new = (macrorec *)Malloc(sizeof(macrorec));
	new->key = key;
	strcpy(new->name, name);

	if (current != NULL && current->key == key)
	{
		new->next = current->next;
		Free(current);
	}
	else
	{
		new->next = current;
	}

	if (last == NULL)
		macrobase = new;
	else
		last->next = new;
}


static void definemenu(short num, short part, char *name, char *cmd, short bool_)
{
	logmenurec *menu;

	menu = popupmenus[num - 1];
	strcpy(menu[part - 1].name, name);
	strcpy(menu[part - 1].cmd, cmd);
	switch (bool_)
	{

		case 1:
			menu[part - 1].bool_ = &gg.invisible;
			break;

		case 2:
			menu[part - 1].bool_ = &gg.textinvisible;
			break;

		case 3:
			menu[part - 1].bool_ = &gg.markers;
			break;

		case 4:
			menu[part - 1].bool_ = &vlsi;
			break;

		case 5:
			menu[part - 1].bool_ = &popup_grid;
			break;

		case 6:
			menu[part - 1].bool_ = &gg.probemode;
			break;

		case 7:
			menu[part - 1].bool_ = &gg.glowmode;
			break;

		case 8:
			menu[part - 1].bool_ = &snapflag;
			break;

		case 9:
			menu[part - 1].bool_ = &gg.pwrflag;
			break;

		default:
			menu[part - 1].bool_ = NULL;
			break;
	}
}

typedef char cnamearray[16][8];

const cnamearray cnames =
{
	"GRAY", "RED", "GREEN", "YELLOW", "MRED", "ORANGE", "CYAN", "WHITE",
	"BLACK", "PINK", "DYELLOW", "XGREEN", "DCYAN", "DRED", "LGRAY", "CRED"
};


static long getcolorname(char *w_)
{
	long Result;
	char w[256];
	short i;
	char STR1[256];
	char STR2[256];

	strcpy(w, w_);
	strcpy(STR1, strltrim(strrtrim(strupper(STR2, w))));
	strcpy(w, STR1);
	i = 15;
	while (i >= 0 && strcmp(cnames[i], w))
		i--;
	if (i >= 0)
		return i;
	TRY(try29);
	i = getint(w, -1L);
	if ((unsigned)i > 15)
		Result = -1;
	else
		Result = i;
	RECOVER(try29);
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	else
		return -1;
	ENDTRY(try29);
	return Result;
}

void docnffunction()
{
	char arg[256];
	long i, j, k;
	char ch;
	strlist_t *cp;
	long clrarr[3];
	int flag;
	librstrrec *lsp, **lspp;
	char STR2[256];

	if (*gg.func == '\0')
		return;
	if (!strcmp(gg.func, "COLOR"))
	{
		getword(gg.funcarg, arg);
		k = getcolorname(arg);
		if (k >= 0)
		{
			flag = true;
			for (i = 1; i <= 3; i++)
			{
				j = getint(gg.funcarg, -1L);
				if ((unsigned long)j > 255)
					flag = false;
				else
					clrarr[i - 1] = j;
			}
			if (flag)
			{
				k = reducecolor8((int)k);
				rcolormap[k] = clrarr[0];
				gcolormap[k] = clrarr[1];
				bcolormap[k] = clrarr[2];
				m_vsetcolors(k, 1L, &rcolormap[k], &gcolormap[k], &bcolormap[k]);
			}
			clearfunc();
			return;
		}
		k = getcolorname(gg.funcarg);
		strlist_remove(&colorbase, arg);
		if (k >= 0)
		{
			cp = strlist_add(&colorbase, arg);
			cp->value = (na_long)k;
		}
		else if (gg.initdone)
		{
			beginerror();
			nc_printf("No such color as \"%s\"\n", gg.funcarg);
			enderror();
		}
		if (gg.initdone)
		{
			initcolors();
			if (gg.showpage > 0)
				refrwindow();
		}
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "MACRO"))
	{
		strwordx(gg.funcarg, arg);
		if (*arg != '\0')
		{
			if (strlen(arg) > 1)
			{
				if (strcicmp(arg, "sp") == 0)
					ch = ' ';
				else if (strcicmp(arg, "bs") == 0)
					ch = '\007';
				else if (strcicmp(arg, "tab") == 0)
					ch = '\t';
				else if (strcicmp(arg, "cr") == 0)
					ch = '\015';
				else if (strlen(arg) == 2 && arg[0] == '^')
					ch = (arg[1] & 31) + 167;
				else
					ch = (char)getint(arg, 32L);
			}
			else
			{
				ch = arg[0];
			}

			if (*gg.funcarg == '\0')
				echomacro(ch);
			else
				definemacro(ch, gg.funcarg);
		}
		else if (gg.initdone)
		{
			statusdisplay("MACRO");
		}
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "POPUP")) {
		i = getint(gg.funcarg, 0L);
		j = getint(gg.funcarg, 0L);
		if (i >= 1 && i <= 4 && j >= 1 && j <= 8) {
			k = getint(gg.funcarg, 0L);
			getword(gg.funcarg, arg);
			if (*arg != '\0')
				definemenu((int)i, (int)j, gg.funcarg, arg, (int)k);
			else
				definemenu((int)i, (int)j, "", "", 0);
		}
		clearfunc();
		return;
	}
	if (!strcmp(gg.func, "GROUP")) {
		i = getint(gg.funcarg, 0L);
		if ((unsigned long)i <= 8) {
			if (*gg.funcarg == '\0')
				curlistgroup = i;
			else {
				getword(gg.funcarg, arg);
				strchange(&librgroupnames[i], arg);
			}
		}
		clearfunc();
		return;
	}
	if (strcmp(gg.func, "LIBRSTR"))
		return;
	/*   else if gg.func = 'VLSI' then
		 begin
		 getbool(gg.funcarg^, vlsi);
		 clearfunc;
		 end;  */
	getword(gg.funcarg, arg);
	if (*arg != '\0' && *gg.funcarg != '\0') {
		strcpy(arg, strreverse(STR2, arg));
		lspp = &librstrs;
		while (*lspp != NULL && strcmp((*lspp)->name, arg)) {
			if (strcmp(arg, (*lspp)->name) < 0)
				lspp = &(*lspp)->left;
			else
				lspp = &(*lspp)->right;
		}
		if (*lspp == NULL) {
			lsp = (librstrrec *)Malloc(sizeof(librstrrec));
			*lspp = lsp;
			lsp->left = NULL;
			lsp->right = NULL;
			strcpy(lsp->name, arg);
			lsp->str = NULL;
		} else
			lsp = *lspp;
		strchange(&lsp->str, gg.funcarg);
	}
	clearfunc();
}

void dofunction()
{
	log_tool *tp;
	char cmd[17];
	short i;
	int oflag;
	char STR2[256];
	short FORLIM;
	unsigned char TEMP;

	TRY(try30);
	remcursor();
	doimmedfunction();
	if (*gg.func != '\0')
	{
		strcpy(cmd, gg.func);
		tp = gg.toolbase;
		while (tp != NULL && !strcmp(gg.func, cmd))
		{
			if (tp->ready)
				calltool(tp, act_func);
			tp = tp->next;
		}
	}
	docnffunction();
	if (*gg.func != '\0')
	{
		if (!strcmp(gg.func, "DEL"))
		{
			delcommand();
		}
		else if (!strcmp(gg.func, "COPY"))
		{
			copycommand();
		}
		else if (!strcmp(gg.func, "CLEAR"))
		{
			clearfunc();
			deleverything();
			histdelsignals();
		}
		else if (!strcmp(gg.func, "PASTE"))
		{
			pastecommand(0L);
		}
		else if (!strcmp(gg.func, "EXTRACT"))
		{
			extract();
		}
		else if (!strcmp(gg.func, "MOVE"))
		{
			movecommand(strcmp(gg.funcarg, "*") != 0);
		}
		else if (!strcmp(gg.func, "OPENH"))
		{
			openhoriz();
		}
		else if (!strcmp(gg.func, "OPENV"))
		{
			openvert();
		}
		else if (!strcmp(gg.func, "CLOSEH"))
		{
			closehoriz();
		}
		else if (!strcmp(gg.func, "CLOSEV"))
		{
			closevert();
		}
		else if (!strcmp(gg.func, "CENTER"))
		{
			centercommand();
		}
		else if (!strcmp(gg.func, "YARDSTICK"))
		{
			yardstickcommand();
		}
		else if (!strcmp(gg.func, "DEFINE"))
		{
			gatedefinitioncommand();
		}
		else if (!strcmp(gg.func, "TOOL"))
		{
			toolcommand();
			clearfunc();
		}
		else if (!strcmp(gg.func, "LABEL"))
		{
			clearfunc();
			if (!gg.textinvisible)
			{
				if (*gg.funcarg == '\0')
					editlabel(NULL);
				else
					addlabel(&gg.nearlabel, gg.funcarg);
			}
		}
		else if (!strcmp(gg.func, "BOX"))
		{
			if (!gg.textinvisible)
				addbox();
			else
				clearfunc();
		}
		else if (!strcmp(gg.func, "VLSI"))
		{
			clearfunc();
			if (gg.pages[gg.curpage - 1]->gbase == NULL &&
					gg.pages[gg.curpage - 1]->hwbase == NULL &&
					gg.pages[gg.curpage - 1]->vwbase == NULL)
				setvlsimode(!vlsi);
		}
		else if (!strcmp(gg.func, "FAST"))
		{
			clearfunc();
			message("FAST mode enabled.");
			gg.fastspeed = gg.fastmax;
		}
		else if (!strcmp(gg.func, "SCOPE"))
		{
			clearfunc();
			historycommand();
		}
		else if (!strcmp(gg.func, "DUMPSCOPE"))
		{
			clearfunc();
			dumphistory();
		}
		else if (!strcmp(gg.func, "STEP"))
		{
			singlestep();
			clearfunc();
		}
		else if (!strcmp(gg.func, "TAPMODE"))
		{
			clearfunc();
			if (cureditmode == maxeditmodes)
				cureditmode = 1;
			else
				cureditmode++;
			refreditmode();
		}
		else if (!strcmp(gg.func, "ROT"))
		{
			clearfunc();
			cureditmode = 1;
			refreditmode();
		}
		else if (!strcmp(gg.func, "MIRX"))
		{
			clearfunc();
			cureditmode = 2;
			refreditmode();
		}
		else if (!strcmp(gg.func, "MIRY"))
		{
			clearfunc();
			cureditmode = 3;
			refreditmode();
		} else if (!strcmp(gg.func, "CNFG"))
		{
			clearfunc();
			cureditmode = 4;
			refreditmode();
		}
		else if (!strcmp(gg.func, "CAT"))
		{
			gatecatalog(false);
		}
		else if (!strcmp(gg.func, "LIBRARY"))
		{
			gatecatalog(true);
		}
		else if (!strcmp(gg.func, "MAKE"))
		{
			makegatecmd();
		}
		else if (!strcmp(gg.func, "GET"))
		{
			getgatecmd();
		}
		else if (!strcmp(gg.func, "NAME"))
		{
			namecommand();
		}
		else if (!strcmp(gg.func, "SAVE"))
		{
			savecommand();
		}
		else if (!strcmp(gg.func, "LOAD"))
		{
			loadcommand();
		}
		else if (!strcmp(gg.func, "READ"))
		{
			readcommand();
		}
		else if (!strcmp(gg.func, "MESSAGE"))
		{
			message(gg.funcarg);
			clearfunc();
		}
		else if (!strcmp(gg.func, "VMESSAGE"))
		{
			vmessage(gg.funcarg);
			clearfunc();
		}
		else if (!strcmp(gg.func, "DUMPCONFLICTS"))
		{
			dumpconflicts();
			clearfunc();
		}
		else if (!strcmp(gg.func, "TRAIN"))
		{
			clearfunc();
		}
		else if (!strcmp(gg.func, "STATUS"))
		{
			statusdisplay(gg.funcarg);
			clearfunc();
		}
		else if (!strcmp(gg.func, "MARKER"))
		{
			oflag = gg.markers;
			getbool(gg.funcarg, &gg.markers);
			if (gg.markers != oflag)
			{
				if (gg.markers)
				{
					gg.markerx1 = (gg.xoff + 30) / gg.scale;
					gg.markery1 = (gg.yoff + 30) / gg.scale;
					gg.markerx2 = (gg.xoff + across - 30) / gg.scale;
					gg.markery2 = (gg.yoff + baseline - 30) / gg.scale;
					drawmarkersc(gg.color.marker);
				} else
				{
					drawmarkersc(gg.color.backgr);
				}
				stamp(&gg.markerstamp);
			}
			vmessageflag("Markers are ", gg.markers);
			clearfunc();
		}
		else if (!strcmp(gg.func, "TRAIN+"))
		{
			if (training)
			{
				help_show(helpptr, "+Q");
			}
			clearfunc();
		}
		else if (!strcmp(gg.func, "TRAIN-"))
		{
			if (training)
			{
				help_show(helpptr, "-Q");
			}
			clearfunc();
		}
		else if (!strcmp(gg.func, "REFRESH"))
		{
			clearfunc();
			refrwindow();
		}
		else if (!strcmp(gg.func, "RESIZE"))
		{
			clearfunc();
			resize_window();
		}
		else if (!strcmp(gg.func, "REFR"))
			scroll();
		else if (!strcmp(gg.func, "SHELL"))
		{
			doshellescape(gg.funcarg);
			clearfunc();
		}
		else if (*gg.func == '!')
		{
			sprintf(STR2, "%s %s", gg.func, gg.funcarg);
			doshellescape(STR2);
			clearfunc();
		}
		else if (!strcmp(gg.func, "EXIT"))
		{
			if (strcmp(gg.funcarg, "*") && !anychanged())
			{
				P_escapecode = 20;
				goto _Ltry30;
			}
			clearshowalpha();
			nc_putchar('\n');
			if (anychanged())
			{
				nc_printf("The following page(s) have not been saved: ");
				FORLIM = gg.numpages;
				for (i = 1; i <= FORLIM; i++)
				{
					if (isPageChanged(gg.pages[i - 1]))
						nc_printf("%d ", i);
				}
				nc_putchar('\n');
			}
			m_alpha_on();
			m_graphics_off();
			nc_gotoXY(31, 10);
			nc_printf("%c%cExit from program? %c", chryellow, chrplain, chrgreen);
			do
				TEMP = m_inkey();
			while ((TEMP == 250)||(TEMP == 251));
			if (TEMP == 'y' || TEMP == 'Y')
			{
				P_escapecode = 20;
				goto _Ltry30;
			}
			m_graphics_on();
			clearalpha();
			clearfunc();
		}
		else if (!strcmp(gg.func, "ABORT"))
		{
			remcursor();
			if (cursortype == grid)
				cursortype = normal;
			gg.probemode = false;
			clearfunc();
		}
		else if (!strcmp(gg.func, "X"))
		{
			clearfunc();
		}
	}
	RECOVER2(try30,_Ltry30);
	if (excpline == -1)
		excpline = EXCP_LINE;
	if (P_escapecode != 5)
		_Escape(P_escapecode);
	remcursor();
	gg.startpoint = false;
	log_setmode("");
	/* refreshsoon; */
	if (gg.t.dn && !gg.incircuit)
	{
		gg.t.dn = false;
		gg.t.depressed = false;
	}
	cursortype = normal;
	ENDTRY(try30);

	/*debugging*/
}



static void initmacros()
{
	definemacro(171, "EXIT *");   /*control-D*/
	definemacro(':', "DO");
	definemacro(' ', "REFRESH");
	definemacro('!', "SHELL");
	definemacro('1', "PAGE 1");
	definemacro('2', "PAGE 2");
	definemacro('3', "PAGE 3");
	definemacro('4', "PAGE 4");
	definemacro('5', "PAGE 5");
	definemacro('6', "PAGE 6");
	definemacro('7', "PAGE 7");
	definemacro('8', "PAGE 8");
	definemacro('9', "PAGE 9");
	definemacro('*', "PASTE");
	definemacro(',', "ALTPOSN");
	definemacro('.', "PROBE");
	definemacro('/', "COPY");
	definemacro('0', "RESET");
	definemacro('<', "ZOOMDN");
	definemacro('>', "ZOOMUP");
	definemacro('?', "HELP");
	definemacro('+', "PAGE +");
	definemacro('-', "PAGE -");
	definemacro('A', "AUTOWINDOW");
	definemacro('b', "BOX");
	definemacro('c', "CNFG");
	definemacro('C', "CAT");
	definemacro('d', "DEL");
	definemacro('D', "DEFINE");
	definemacro('e', "EXAMINE");
	definemacro('E', "EXAMINE");
	definemacro('f', "FAST");
	definemacro('g', "GLOW");
	definemacro('G', "GRID");
	definemacro('h', "HOME");
	definemacro('i', "INVISIBLE");
	definemacro('I', "INVLABEL");
	definemacro('k', "SHOWCONFLICTS");
	definemacro('l', "LABEL");
	definemacro('L', "LOAD");
	definemacro('m', "MOVE");
	definemacro('M', "TAPMODE");
	definemacro('n', "INVPINNUM");
	definemacro('o', "ONOFF");
	definemacro('r', "ROT");
	definemacro('R', "RESET");
	definemacro('q', "QUIET");
	definemacro('Q', "EXIT *");
	definemacro('s', "SCOPE");
	definemacro('S', "SAVE *");
	definemacro('t', "STEP");
	definemacro('T', "TOOL");
	definemacro('v', "VERBOSE");
	definemacro('x', "EXAMINE");
	definemacro('y', "YARDSTICK");
	definemacro('Z', "EXIT *");
	definemacro(250, "REFRESH");
	definemacro(251, "RESIZE");
}


static void newmenu(logmenurec **menu)
{
	*menu = (logmenurec *)Malloc(sizeof(logmenurec) * 8);
	for (short i = 0; i <= 7; i++)
		*(*menu)[i].name = '\0';
}


static void initmenus()
{
	for (short i = 0; i <= 3; i++)
		newmenu(&popupmenus[i]);
	definemenu(1, 1, "Box", "BOX", 0);
	definemenu(1, 2, "Label", "LABEL", 0);
	definemenu(1, 3, "Invisible", "INVISIBLE", 1);
	definemenu(1, 4, "Inv lbls", "INVLABEL", 2);
	definemenu(1, 5, "Zoom up", "ZOOMUP", 0);
	definemenu(1, 6, "Zoom dn", "ZOOMDN", 0);
	definemenu(1, 7, "Markers", "MARKER", 3);
	definemenu(1, 8, "Yardstick", "YARDSTICK", 0);
	definemenu(2, 1, "Delete", "DEL", 0);
	definemenu(2, 2, "Copy", "COPY", 0);
	definemenu(2, 3, "Move", "MOVE", 0);
	definemenu(2, 4, "Paste", "PASTE", 0);
	definemenu(2, 5, "#Open H", "OPENH", 0);
	definemenu(2, 6, "#Open V", "OPENV", 0);
	definemenu(2, 7, "#Close H", "CLOSEH", 0);
	definemenu(2, 8, "#Close V", "CLOSEV", 0);
	definemenu(3, 1, "Grid", "GRID", 5);
	definemenu(3, 2, "Probe", "PROBE", 6);
	definemenu(3, 3, "Glow", "GLOW", 7);
	definemenu(3, 5, "Alt posn", "ALTPOSN", 0);
	definemenu(3, 6, "Home", "HOME", 0);
	definemenu(3, 8, "Refresh", "REFRESH", 0);
	definemenu(4, 1, "Load page", "LOAD", 0);
	definemenu(4, 2, "Save page", "SAVE *", 0);
	definemenu(4, 4, "Status", "STATUS", 0);
	definemenu(4, 5, "Scope", "SCOPE", 0);
	definemenu(4, 6, "Reset", "RESET", 0);
	definemenu(4, 7, "Simulation", "ONOFF", 9);
	definemenu(4, 8, "*Exit", "EXIT", 0);
}



static void inithooks()
{
	gg.hook.hidecursor = hidecursor_hook;
	gg.hook.unhidecursor = unhidecursor_hook;
	gg.hook.hidecursorrect = hidecursorrect_hook;
	gg.hook.nodeconflict = reportnodeconflict;
	gg.hook.gateconflict = reportgateconflict;
	gg.hook.clearconflicts = clearconflicts;
	gg.hook.setdimgate = setdimgate;
	gg.hook.xform = xformcoords;
	gg.hook.getsig = getsigname;
	gg.hook.report = report;
	gg.hook.message = message;
	gg.hook.vmessage = vmessage;
	gg.hook.vmessageflag = vmessageflag;
	gg.hook.trace = tracemessage;
	gg.hook.dump = dumpmessage;
	gg.hook.setdump = setdumpname;
	gg.hook.closefiles = closedumpfiles;
	gg.hook.clearalpha = clearalpha;
	gg.hook.drawstr = drawstr2;
	gg.hook.rightstr = rightstr2;
	gg.hook.centerstr = centerstr2;
	gg.hook.clip = clipon;
	gg.hook.remcursor = remcursor;
	gg.hook.clearwindow = clearwindow;
	gg.hook.setmode = log_setmode;
	gg.hook.getword = getword;
	gg.hook.getint = pgetint;
	gg.hook.getreal = pgetreal;
	gg.hook.getbool = getbool;
	gg.hook.pen = pen;
	gg.hook.pass = pass;
	gg.hook.trykbd = trykbd;
	gg.hook.scroll = scroll;
	gg.hook.trykbdscroll = trykbdscroll;
	gg.hook.inkey = inkey3;
	gg.hook.udrawgate = udrawgatec;
	gg.hook.gdrawgate = drawgatec;
	gg.hook.drawgate = drawgatex;
	gg.hook.erasegate = eragate;
	gg.hook.drawnode = drawnodec;
	gg.hook.refresh = refresh;
	gg.hook.refreshsoon = refreshsoon;
	gg.hook.beginbottom = beginbottom;
	gg.hook.endbottom = endbottom;
	gg.hook.beginerror = beginerror;
	gg.hook.enderror = enderror;
	gg.hook.redraw = refrwindow1;
	gg.hook.redrawwindow = refrwindow;
	gg.hook.change = chpage;
	gg.hook.newgate = newgate;
	gg.hook.copygate = copygate;
	gg.hook.disposegate = disposegate;
	gg.hook.newnode = newnode;
	gg.hook.copynode = copynode;
	gg.hook.disposenode = disposenode;
	gg.hook.switchnode = switchnode;
	gg.hook.dumpnodes = dumpnodes;
	gg.hook.readlnpass = readlnpass;
	gg.hook.clearfunc = clearfunc;
	gg.hook.refrfunc = refrfunc;
	gg.hook.assertfunc = assertfunc;
	gg.hook.findobject = closergate;
	gg.hook.findwire = closerwire;
	gg.hook.deletehw = delhwire;
	gg.hook.deletevw = delvwire;
	gg.hook.disconnectgate = disconnectgate;
	gg.hook.deletegate = delgate;
	gg.hook.addhw = addhwire2;
	gg.hook.addvw = addvwire2;
	gg.hook.connectgate = pconnectgate;
	gg.hook.waitnear = waitnear;
	gg.hook.addgate = addgate1;
	gg.hook.uaddgate = uaddgate;
	gg.hook.getgate = getgate;
	gg.hook.realunit = prealunit;
	gg.hook.realunit2 = prealunit2;
	gg.hook.addlabel = addlabel;
	gg.hook.getcolor = getcolor;
	gg.hook.setcmap = setcmap;
	gg.hook.parselabel = parselabel;
	gg.hook.editattrs = editattrsx;
	gg.hook.newattrs = newattrs;
	gg.hook.disposeattrs = disposeattrs;
	gg.hook.copyattrs = copyattrs;
	gg.hook.initwindow = initwindow;
	gg.hook.clearshowalpha = clearshowalpha;
	gg.hook.setupregion = setupregion;
	gg.hook.stamp = stamp;
	gg.hook.noclip = clipoff;
	gg.hook2 = (log_hook2_t *)Malloc(sizeof(log_hook2_t));
	gg.hook2->getnodeval = getnodeval;
	gg.hook2->solderat = solderat;
	gg.hook2->unsolderwires = unsolderwires;
	gg.hook2->settofrom = settofrom;
	gg.hook2->findattrnum = findattrnum;
	gg.hook2->findattrname = findattrname;
	gg.hook2->findpinnum = findpinnum;
	gg.hook2->findpinname = findpinname;
	gg.hook2->setgattr = setgattr;
	gg.hook2->setnattr = setnattr;
	gg.hook2->findpointmarker = findpointmarker;
	gg.hook2->findboxmarker = findboxmarker;
	gg.hook2->send_general = send_general;
	gg.hook2->send_gennode = send_gennode;
	gg.hook2->send_genkind = send_genkind;
	gg.hook2->send_gengate = send_gengate;
	gg.hook2->send_all = send_all;
	gg.hook2->send_allnodes = send_allnodes;
	gg.hook2->send_allkinds = send_allkinds;
	gg.hook2->send_allgates = send_allgates;
	gg.hook2->addlabelat = addlabelat;
	gg.hook2->addboxat = addboxat;
	gg.hook2->plainxform = plainxformcoords;
	gg.hook2->findattrnum2 = findattrnum2;
	gg.hook2->findattrname2 = findattrname2;
	gg.hook2->showpinname = showpinname;
}


#define rtn             "INITIALIZE"

/** @brief Initialize all variables.
  Read basic gates from library.
  Clear catalog window.
  Initialize global data structures.
  */
static void initialize()
{
	struct LOC_initialize V;
	newci_parserec swtab[10];
	short i, k, curgate;
	short gptr[maxgatesfiles], ggroup[maxgatesfiles], gsize[maxgatesfiles];
	char gname[maxgatesfiles][9];
	strlist_t *l1;
	char s[256];
	int flag;
	char cnfname[256];
	char STR2[256], STR3[256];
	short FORLIM;
	log_sigrec *WITH1;

	strcpy(swtab[0].switch_, "cC");   /* CNF file name */
	swtab[0].kind = 'M';
	strcpy(swtab[1].switch_, "vV");   /* Vanilla LOG (no CNF) */
	swtab[1].kind = 'B';
	strcpy(swtab[2].switch_, "zZ");   /* Trace mode initially on */
	swtab[2].kind = 'S';
	strcpy(swtab[3].switch_, "dD");   /* Dump file name */
	swtab[3].kind = 'S';
	strcpy(swtab[4].switch_, "tT");   /* Trace file name */
	swtab[4].kind = 'S';
	strcpy(swtab[5].switch_, "hH");   /* Home directory name */
	swtab[5].kind = 'S';
	strcpy(swtab[6].switch_, "rR");   /* Tool to run */
	swtab[6].kind = 'M';
	strcpy(swtab[7].switch_, "xX");   /* XDisplay name */
	swtab[7].kind = 'M';

	newci_parseswitch(swtab, 8, V.cmdbuf);

	/* handle other arguments */
	if (*V.cmdbuf != '\0')
	{
		if (*V.cmdbuf != '\0')
			nc_printf("Unrecognized option:  -%s\n", V.cmdbuf);
		nc_printf("\nUsage:  LOG [ -v ] [ -c cnffile ] [ -x X_display_name ] [ file ]\n");
		_Escape(0);
	}

	/* set display name id exists */
	if (swtab[7].used > 0)
	{
		m_set_display_name(swtab[7].UU.s);
	}

	gg.curstamp = 0;
	initwindow2();
	initcolormap();
	libptr = 0;
	helpptr = 0;
	cursortype = normal;
	oldcursortype = normal;
	setscale(0);
	curlistgroup = 0;
	messagepos = topmessagepos;
	resetmessages();
	gg.verbose = true;
	tempverbose = false;
	gg.glowmode = false;
	gg.probemode = false;
	gg.textinvisible = false;
	gg.pnuminvisible = true;
	gg.invisible = false;
	gg.pwrflag = true;
	gg.busyflag = true;
	snapflag = false;
	glowsolder = true;
	gg.quiet = false;
	gg.dotsvisible = true;
	gg.showconflicts = false;
	pushedbackkey = '\0';
	conflictenbl = true;
	conflictstop = false;
	conflictdelay = 1;
	anyconflicts = false;
	gg.actstr = (char *)Malloc(256);
	*gg.actstr = '\0';
	gg.maxsignal = defmaxsignal;
	strlist_init(&histlbl);
	strlist_append(&histlbl, "Scope mode");
	strlist_append(&histlbl, "");
	strlist_append(&histlbl, "VContinuous,On Reset,Triggered,Manual:Type of trigger:");
	strlist_append(&histlbl, "Triggered;C(none):Trigger signal:");
	strlist_append(&histlbl, "On Reset;OUs:Time to start trigger:");
	strlist_append(&histlbl, "On Reset;OUs:Time to stop trigger:");
	strlist_append(&histlbl, "VTrigger off,Simulation off:Action when memory full:");
	strlist_append(&histlbl, "");
	strlist_append(&histlbl, "Us:Minimum timestep:");
	strlist_append(&histlbl, "Us:Maximum timestep:");
	strlist_append(&histlbl, "I100000:Timestep memory limit:");
	strlist_append(&histlbl, "I:Timestep memory size:");
	strlist_append(&histlbl, "");
	strlist_append(&histlbl, "BY:Align signal names?");
	strlist_append(&histlbl, "");
	strlist_append(&histlbl, "Us,0:Time at left edge:");
	strlist_append(&histlbl, "Us:Current time:");
	strlist_append(&histlbl, "");
	strlist_append(&histlbl, "Us,100us:Seconds per division:");
	strlist_append(&histlbl, "Us:Current timestep:");
	parselabel(&histlbl, &histnumattrs, &histkattr);
	newattrs(&histgattr, histnumattrs, histkattr);
	gg.singlestepcount = 0;
	steppingoff = false;
	vlsi = false;
	gg.dumpopen = false;
	dumpfname = (char *)Malloc(256);
	*dumpfname = '\0';
	gg.traceopen = false;
	gg.tracefile = &tracefile;
	tracefname = (char *)Malloc(256);
	*tracefname = '\0';

	if (swtab[3].used > 0)
		strcpy(dumpfname, swtab[3].UU.s);
	if (swtab[4].used > 0)
		strcpy(tracefname, swtab[4].UU.s);
	else if (swtab[2].used > 0)
		strcpy(tracefname, swtab[2].UU.s);
	if (*tracefname != '\0')
		newci_fixfname(tracefname, "text", "");
	gg.traceflag = (swtab[2].used > 0);
	if (gg.traceflag)
		tracemessage("Trace mode ON");
	gg.homedirname = (char *)Malloc(256);

	if (swtab[5].used > 0)
	{
		if (*swtab[5].UU.s == '\0')
			sethomedirname("");
		else
			sethomedirname(swtab[5].UU.s);
	}
	else
	{
		sethomedirname("~/log");
	}
	cursorhide = false;
	initmenus();
	gg.funcarg = (char *)Malloc(256);
	*gg.funcarg = '\0';
	gg.toolbase = NULL;
	inithooks();
	gg.simstate = simst_null;
	gg.simstatetool = NULL;
	for (i = 0; i < maxgatesfiles; i++)
		gatesname[i] = NULL;
	for (i = 0; i <= 8; i++)
		librgroupnames[i] = NULL;
	loghelpname = (char *)Malloc(256);
	strcpy(loghelpname, "loghelp");
	lognewsname = (char *)Malloc(256);
	strcpy(lognewsname, "lognews");
	gg.fastmin = deffastmin;
	gg.fastmax = deffastmax;
	gg.fastrate = deffastrate;
	gg.fastspeed = gg.fastmin;
	stamp(&gg.labelstamp);
	stamp(&gg.boxstamp);
	stamp(&gg.msgstamp);
	stamp(&gg.refrstamp);
	stamp(&gg.nattrstamp);
	stamp(&gg.gattrstamp);
	stamp(&gg.sigstamp);
	stamp(&gg.loadstamp);
	stamp(&gg.colorstamp);
	stamp(&gg.resetstamp);
	stamp(&gg.markerstamp);
	strlist_init(&V.loadgates);
	strlist_init(&V.logmenu);
	strlist_init(&gatefilenames);
	cnfbase = NULL;
	V.cnflast = NULL;
	colorbase = NULL;
	macrobase = NULL;
	initmacros();
	librstrs = NULL;
	commandlist = NULL;
	thingstodo = NULL;
	nexttodo = NULL;
	messages = NULL;
	*cnfname = '\0';
	flag = false;
	if (swtab[0].used > 0)
		strcpy(cnfname, swtab[0].UU.s);
	if (*cnfname == '\0' && swtab[1].used == 0)
		strcpy(cnfname, "*");
	if (*cnfname != '\0')
		readcnf(cnfname, &V);
	initcolors();
	gg.color.curbaseline = gg.color.baseline;
	gg.baselinecolor = gg.color.baseline;
	newci_fixfname(loghelpname, "help", "");
	newci_fixfname(lognewsname, "text", "");
	displaynews = (*lognewsname != '\0');
	if (P_argc > 1)
	{
		sprintf(STR2, "LOAD %s", P_argv[1]);
		strlist_append(&thingstodo, STR2);
		FORLIM = P_argc;
		for (i = 2; i < FORLIM; i++)
		{
			sprintf(STR3, "PAGE %d", i);
			strlist_append(&thingstodo, STR3);
			sprintf(STR3, "LOAD %s", P_argv[i]);
			strlist_append(&thingstodo, STR3);
		}
		displaynews = false;
	}
	if (swtab[6].used > 0 && *swtab[6].UU.s != '\0')
	{
		sprintf(STR3, "TOOL %s", swtab[6].UU.s);
		strlist_append(&thingstodo, STR3);
		displaynews = false;
		justonecommand = true;
	}
	else
	{
		justonecommand = false;
	}
	if (gatefilenames == NULL)
		addgatesfile("log", &V);
	idxsize = 0;
	for (k = 0; k < maxgatesfiles; k++)
	{
		*gname[k] = '\0';
		if (gatefilenames != NULL) {
			TRY(try35);
			gatesname[k] = (char *)Malloc(256);
			strcpy(gatesname[k], gatefilenames->s);
			gatefilenames = gatefilenames->next;
			libf1[k] = (filerecfilerec *)Malloc(sizeof(filerecfilerec));
			libf1[k]->f = NULL;
			newci_fixfname(gatesname[k], "gate", "");
			flag = false;
			TRY(try36);
			if (libf1[k]->f != NULL)
				libf1[k]->f = freopen(gatesname[k], "r", libf1[k]->f);
			else
				libf1[k]->f = fopen(gatesname[k], "r");

			if (libf1[k]->f == NULL)
			{
				P_escapecode = -10;
				P_ioresult = FileNotFound;
				goto _Ltry36;
			}
			RESETBUF(libf1[k]->f, filerec);
			flag = true;
			RECOVER2(try36,_Ltry36);
			if (P_escapecode == -20)
				goto _Ltry35;
			TRY(try37);
			if (libf1[k]->f != NULL)
			{
				sprintf(STR2, "%s/%s", GetChipmunkPath("LOGLIB", LOGLIB),
						gatesname[k]);
				libf1[k]->f = freopen(STR2, "r", libf1[k]->f);
			} else
			{
				sprintf(STR2, "%s/%s", GetChipmunkPath("LOGLIB", LOGLIB),
						gatesname[k]);
				libf1[k]->f = fopen(STR2, "r");
			}
			if (libf1[k]->f == NULL)
			{
				P_escapecode = -10;
				P_ioresult = FileNotFound;
				goto _Ltry37;
			}
			RESETBUF(libf1[k]->f, filerec);
			sprintf(STR2, "%s/%s", GetChipmunkPath("LOGLIB", LOGLIB),
					gatesname[k]);
			strcpy(gatesname[k], STR2);
			flag = true;
			RECOVER2(try37,_Ltry37);
			if (P_escapecode == -20)
				goto _Ltry35;
			ENDTRY(try37);
			ENDTRY(try36);
			if (flag)
			{
				if (libf1[k]->f != NULL)
					rewind(libf1[k]->f);
				else
					libf1[k]->f = tmpfile();
				if (libf1[k]->f == NULL)
				{
					P_escapecode = -10;
					P_ioresult = FileNotFound;
					goto _Ltry35;
				}
				SETUPBUF(libf1[k]->f, filerec);
				fseek(libf1[k]->f, 0L, 0);
				SETUPBUF(libf1[k]->f, filerec);
				GET(libf1[k]->f, filerec);
				gsize[k] = getshortsw((char *)(&GETFBUF(libf1[k]->f, filerec).b[4]));
				idxsize += gsize[k];
				libfstart[k] = GETFBUF(libf1[k]->f, filerec).b[6] + 1;
				gptr[k] = 0;
			}
			else
			{
				Free(libf1[k]);
				gatesname[k] = NULL;
			}
			RECOVER2(try35,_Ltry35);
			if (P_escapecode == -20)
				_Escape(P_escapecode);
			beginfatal();
			nc_printf("Can't open gates file \"%s\"\n", gatesname[k]);
			endfatal();
			ENDTRY(try35);
		}
	}
	if (idxsize == 0)
	{
		beginfatal();
		nc_printf("No gates files!\n");
		endfatal();
	}
	index_ = (char(*)[9])Malloc(idxsize * 10L);
	indexfile = (uchar *)Malloc(idxsize);
	indexoffset = (short *)Malloc(idxsize * sizeof(short));
	indexgroup = (uchar *)Malloc(idxsize);
	loadedgates = (uchar *)Malloc((idxsize + 9L) / 8);
	FORLIM = idxsize;
	for (i = 0; i < FORLIM; i++)
		P_clrbits_B(loadedgates, i, 0, 3);
	curgate = 1;
	do
	{
		k = 0;
		strcpy(s, "\177");
		for (i = 0; i < maxgatesfiles; i++)
		{
			if (gatesname[i] != NULL && *gname[i] == '\0' && gptr[i] < gsize[i])
			{
				TRY(try38);
				gptr[i]++;
				if ((gptr[i] & 31) == 0)
					GET(libf1[i]->f, filerec);
				strcpy(gname[i], "        ");
				strmove(8, GETFBUF(libf1[i]->f, filerec).ix[gptr[i] & 31], 1,
						gname[i], 1);
				ggroup[i] = 0;
				for (V.j = 1; V.j <= 8; V.j++)
				{
					if ((gname[i][V.j - 1] & (~127)) != 0)
					{
						ggroup[i] = V.j;
						gname[i][V.j - 1] = (char)((uchar)gname[i][V.j - 1] - 128);
					}
				}
				RECOVER(try38);
				if (P_escapecode == -20)
					_Escape(P_escapecode);
				beginfatal();
				nc_printf("Can't read gates file \"%s\"\n", gatesname[i]);
				endfatal();
				ENDTRY(try38);
			}
			if (strcmp(gname[i], s) < 0 && *gname[i] != '\0')
			{
				strcpy(s, gname[i]);
				k = i + 1;
			}
		}
		indexfile[curgate - 1] = k;
		indexoffset[curgate - 1] = gptr[k - 1];
		indexgroup[curgate - 1] = ggroup[k - 1];
		strcpy(index_[curgate - 1], strrtrim(strcpy(STR3, s)));
		for (i = 1; i <= maxgatesfiles; i++)
		{
			if (!strcmp(gname[i - 1], s))
			{
				*gname[i - 1] = '\0';
				if (i != k)
					idxsize--;
			}
		}
		curgate++;
	} while (curgate <= idxsize);
	if (m_across >= 1023)
	{
		maxkinds = P_imin2(280L, (long)maxmaxkinds);
		catwidth = 20;
	}
	else
	{
		maxkinds = P_imin2(70L, (long)maxmaxkinds);
		catwidth = 10;
	}
	catboxes = NULL;
	FORLIM = maxkinds;
	for (i = 0; i < FORLIM; i++)
		kind[i] = NULL;
	m_graphics_on();
	clearalpha();

	for(int i = 0; i < log_maxpages; ++i)
		gg.pages[i] = pageAlloc();
	gg.numpages = 1;
	gg.curpage = 1;
	gg.showpage = 1;
	realcurpage = 1;
	setvlsimode(vlsi);
	XRebindKeysym(m_display, XStringToKeysym("BackSpace"), NULL, 0, (unsigned char * )"\007", 1);
	gg.refrflag = true;
	gg.markers = false;
	*gg.func = '\0';
	*gg.genfunc = '\0';
	gg.xoff = origin;
	gg.yoff = origin;
	xoffp[0] = origin;
	yoffp[0] = origin;
	xoff0 = 0;
	yoff0 = 0;
	gatecount[0] = 0;
	gg.nbase = NULL;
	stamp(&gg.pages[0]->pagestamp);
	curfilename[0] = NULL;
	initbuf(&copybuf);
	nodeconflictbase = NULL;
	gateconflictbase = NULL;
	gg.movinghw = NULL;
	gg.movingvw = NULL;
	gg.signaltab = (log_sigrec *)Malloc(gg.maxsignal * sizeof(log_sigrec));
	FORLIM = gg.maxsignal;
	for (i = 0; i < FORLIM; i++)
	{
		WITH1 = &gg.signaltab[i];
		WITH1->name = NULL;
		WITH1->np = NULL;
	}
	gg.lastsignal = 0;
	gg.hnbase = NULL;
	hncount = 0;
	hnocount = 0;
	gg.htbase = NULL;
	gg.htlast = NULL;
	htcount = 0;
	histtrig = 0;
	gg.histactive = false;
	histonscreen = false;
	histgridmode = 2;
	histgridwhich = 0;
	histgridhn = NULL;
	gg.probenode = NULL;
	gg.probegate = NULL;
	gg.probepin = 0;
	gg.probesimtype = NULL;
	gg.probekind = NULL;
	probeflag = false;
	briefprobe = false;
	firsttraining = true;
	training = false;
	cureditmode = 1;
	*modename = '\0';
	modeflag = false;
	modetime = 0;
	modeprobekind = NULL;
	gg.time = 0.0;
	gg.prevtimestep = 0.0;
	reportnowait = false;
	gg.startpoint = false;
	simtype_ignore = findtool("0");
	if (!simtype_ignore->ready)
		report(10, rtn);
	simtype_common = findtool("1");
	if (!simtype_common->ready)
		report(11, rtn);
	i = 1;
	l1 = V.loadgates;
	while (l1 != NULL && i != 0)
	{
		if (*l1->s == '\0')
			getgategroup((strlist_t *)l1->value);
		else
			i = readlibrary(l1->s);
		l1 = l1->next;
	}
	strlist_empty(&V.loadgates);
	l1 = V.logmenu;
	FORLIM = kindgroupsize;
	for (i = 0; i < FORLIM; i++)
	{
		kindgroup[i] = 0;
		kindsig[i] = 0;
		kindattr[i] = NULL;
		if (l1 != NULL)
		{
			kindgroup[i] = readlibrary(l1->s);
			l1 = l1->next;
		}
	}
	strlist_empty(&V.logmenu);
	refrwindow();
	clearalpha();

	/*shift-right*/
}

#undef rtn

static void cleanlog()
{
	for(int i = 0; i < log_maxpages; ++i)
		pageFree(&gg.pages[i]);
	free(lognewsname);
	free(loghelpname);
}


static void shownews()
{
	FILE *f;
	char s[256];
	char *TEMP;

	f = NULL;
	TRY(try39);
	if (f != NULL)
		f = freopen(lognewsname, "r", f);
	else
		f = fopen(lognewsname, "r");

	if (f == NULL)
	{
		P_escapecode = -10;
		P_ioresult = FileNotFound;
		goto _Ltry39;
	}
	clearalpha();
	while (fgets(s, 256, f) != NULL)
	{
		TEMP = (char *)strchr(s, '\n');
		if (TEMP != NULL)
			*TEMP = '\0';
		nc_puts(s);
	}
	m_alpha_on();
	noblink();
	RECOVER2(try39,_Ltry39);
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	ENDTRY(try39);
	if (f != NULL)
		fclose(f);
}


/** @brief Initialize.
  Then Process pen: Add gates. Add wires and solder them.
  Refresh window when necessary.
  Process keyboard and menu areas.
  */
int main(int argc, char * argv[])
{
	char str1[81];
	long temp1;
	long temp2;
	char tempch;

	nc_text_in_window = 1;
	PASCAL_MAIN(argc, argv);
	srand(time(NULL));
	tracefile = NULL;
	dumpfile = NULL;
	gg.initdone = false;
	gg.traceopen = false;
	gg.dumpopen = false;
	excpline = -1;

	for (long i = 0; i < maxgatesfiles; i++)
		libf1[i] = NULL;

	TRY(try40);
	initialize();
	nc_printf("\210\f Starting\201\210 LOG\f\200");
	do
	{
		gg.initdone = true;
		gg.fastspeed = gg.fastmin;
		TRY(try41);
		do
		{
			excpline = -1;
			tempverbose = false;
			suppressdots = false;
			if (justonecommand && thingstodo == NULL)
			{
				P_escapecode = 20;
				goto _Ltry41;
			}
			if (gg.showpage <= 0)
				refrwindow();
			m_graphics_on();
			nc_cursor_off();
			while (messages != NULL)
			{
				switch ((long)messages->value)
				{

					case 0:
						message(messages->s);
						break;

					case 1:
						vmessage(messages->s);
						break;
				}
				strlist_delete(&messages, &messages);
			}
			if (gg.startpoint)
				crosshair(gg.posx, gg.posy);
			else
				nocrosshair();
			gg.stillnear = true;
			if (displaynews)
				shownews();
			if (*gg.func == '\0')
			{
				do
				{
					if (refrtimer == 0 && !gg.startpoint)
					{
						refresh();
						gg.refrflag = true;
					}
					if (refrtimer > 0)
					{
						if (gg.t.moving)
							refrtimer = refrtimer0;
						else
							refrtimer--;
					}
					pass();
					trykbdscroll();
					pen();
					if (!gg.stillnear)
					{
						gg.startpoint = false;
						gg.stillnear = true;
						nocrosshair();
					}

				} while (!(gg.t.dn || *gg.func != '\0'));
			}
			if (displaynews)
			{
				clearalpha();
				displaynews = false;
			}
			nocrosshair();
			clipoff();
			if (*gg.func != '\0')
			{
				commandfound = false;
				dofunction();
				if (!commandfound)
				{
					beginerror();
					nc_printf("Can't understand function %s\n", gg.func);
					enderror();
					clearfunc();
				}
			}
			else if (gg.incircuit)
			{
				pass();
				clipoff();
				gg.oldx = gg.posx;
				gg.oldy = gg.posy;
				ospointflag = gg.startpoint;
				gg.posx = gg.gridx;
				gg.posy = gg.gridy;
				gg.startpoint = true;
				if (!ospointflag && !justtap() && !gg.invisible && !gg.showconflicts)
				{
					if (gg.probemode)
					{
						yardstick();
					}
					else
					{
						moveobject();
						gg.movinghw = NULL;
						gg.movingvw = NULL;
					}
				}
				else
				{
					closergate(gg.posx, gg.posy);
					if (gg.nearlabel != NULL && !gg.invisible && !gg.textinvisible)
					{
						editlabel(gg.nearlabel);
						gg.startpoint = false;
					}
					else if (!ospointflag && gg.neargate != NULL)
					{
						if (gg.probemode)
						{
							configgate(gg.neargate);
							gg.startpoint = false;
						}
						else
						{
							touchgate(gg.neargate);
							gg.startpoint = false;
						}
					}
					else if (!ospointflag)
					{
						closerwire(gg.posx, gg.posy);
						if (gg.probemode)
						{
							if (gg.nearhw != NULL)
								confignode(gg.nearhw->node, "(Node)");
							else if (gg.nearvw != NULL)
								confignode(gg.nearvw->node, "(Node)");
							gg.startpoint = false;
						}
						else if (gg.nearhw != NULL && gg.nearvw != NULL &&
								gg.nearhw->x1 != gg.posx &&
								gg.nearhw->x2 != gg.posx &&
								gg.nearvw->y1 != gg.posy &&
								gg.nearvw->y2 != gg.posy)
						{
							if (findsolder(gg.posx, gg.posy) != NULL)
							{
								unsoldernear();
								pen();   /* Restore cursor */
								gg.startpoint = false;
							}
							else if (!vlsi || gg.nearhw->wcolr == log_wcol_blue ||
									gg.nearvw->wcolr == log_wcol_blue)
							{
								soldernear();
								gg.startpoint = false;
							}
						}
					}
					else
					{
						if (hvline(gg.oldx, gg.oldy, &gg.posx, &gg.posy))
						{
							if (gg.posx != gg.oldx)
								addhwire(gg.posx, gg.oldx, gg.posy, curwcolor);
							if (gg.posy != gg.oldy)
								addvwire(gg.posx, gg.oldy, gg.posy, curwcolor);
						}
					}
					if (gg.invisible || gg.probemode || gg.showconflicts)
						gg.startpoint = false;
				}
			}
			else if (inbox((int)(menux1 - 4), line1 - 5, 34, 20))
				popupmenu(1);
			else if (inbox((int)(menux1 - 4), line2 - 5, 34, 20))
				assertfunc("HELP");
			else if (inbox((int)(menux2 - 4), line1 - 5, 34, 20))
				popupmenu(2);
			else if (inbox((int)(menux2 - 4), line2 - 5, 34, 20))
				assertfunc("CAT");
			else if (inbox((int)(across + menux3 - 4), line1 - 5, 34, 20))
				popupmenu(3);
			else if (inbox((int)(across + menux4 - 4), line1 - 5, 34, 20))
				popupmenu(4);
			else if (inbox((int)(across + menux4 - 4), line2 - 5, 34, 20))
				assertfunc("TAPMODE");
			else
			{
				if (kindgroupleft <= gg.t.x && gg.t.x < kindgroupright &&
						gg.t.y < down) {
					temp1 = (gg.t.x - kindgroupleft) / kindgroupspacing + 1;
					if (gg.probemode)
						configkind((int)temp1);
					else if (justtap())
						flipkind();
					else if (!gg.invisible && !gg.probemode && !gg.showconflicts)
					{
						if (addgate(kindgroup[temp1 - 1], kindsig[temp1 - 1],
									kindattr[temp1 - 1]))
							nextkindsig((int)temp1);
					}
				}
				gg.startpoint = false;
			}
		} while (!pigsfly);
		RECOVER2(try41,_Ltry41);
		gg.curpage = realcurpage;
		*gg.func = '\0';
		if (excpline == -1)
			excpline = EXCP_LINE;
		if (P_escapecode == 20)
		{
			P_escapecode = -20;
			goto _Ltry40;
		}

		if (P_escapecode != -20)
			goto _Ltry40;
		clearshowalpha();
		nc_printf( "\n\n%c%cPress CLR I/O again to exit, any other key to continue.%c\n\n",
				chryellow, chrplain, chrgreen);
		nk_keybufclear();
		waitforkey();
		m_graphics_on();
		clearalpha();
		ENDTRY(try41);
		refrwindow();
	} while (!pigsfly);
	RECOVER2(try40,_Ltry40);
	TRY(try42);
	temp1 = P_ioresult;
	temp2 = P_escapecode;
	if (excpline == -1)
		excpline = EXCP_LINE;
	for (long i = 1; i <= maxgatesfiles; i++)
	{
		if (libf1[i - 1] != NULL)
		{
			if (libf1[i - 1]->f != NULL)
				fclose(libf1[i - 1]->f);
			libf1[i - 1]->f = NULL;
		}
	}
	if (temp2 == 0)
	{
		showalpha();
	}
	else if (temp2 == -20 && timers_sysclock() < watchdog + watchdogtime)
	{
		clearshowalpha();
		nc_printf("%c%c\n\n", chrgreen, chrplain);
	}
	else
	{
		clearshowalpha();
		nc_printf("\n\n%c%cUnexpected runtime error%c%c\n\n",
				chryellow, chrunderl, chrplain, chrgreen);
		nc_printf("    Escapecode =  %ld\n", temp2);
		if (temp2 == -10)
			nc_printf("    Ioresult   =  %ld = %s\n",
					temp1, ioresult_message(str1, temp1));
		if (excpline != -1)
			nc_printf("    Line num   =  %ld\n", excpline);
		nc_printf("\n Probable cause:  ");
		switch (-temp2)
		{

			case 2:
				nc_printf("Not enough memory\n");
				break;

			case 4:
			case 5:
			case 6:
			case 7:
			case 15:
			case 16:
			case 17:
			case 18:
			case 19:
				nc_printf("Arithmetic overflow\n");
				break;

			case 20:
				nc_printf("Unexpected user termination\n");
				break;

			case 10:
				nc_printf("I/O problem\n");
				break;

			case 26:
				nc_printf("Low-level I/O\n");
				break;

			default:
				nc_putchar('\n');
				break;
		}
		nc_putchar('\n');
		nk_keybufclear();
		if (gg.initdone)
		{
			nc_printf("Want debugging printout? ");
			do
			{
				tempch = m_inkey();
			} while (tempch != 'n' && tempch != 'N' && tempch != 'y' &&
					tempch != 'Y');
			nc_printf("%c\n", tempch);
			if (tempch == 'y' || tempch == 'Y')
			{
				TRY(try43);
				gg.traceflag = true;
				tracemessage("");
				fprintf(tracefile, "ESCAPECODE %ld\n", temp2);
				fprintf(tracefile, "IORESULT   %ld\n", temp1);
				fprintf(tracefile, "LINE       %ld\n\n", excpline);
				fprintf(tracefile, "Current page: %ld\n", gg.curpage);
				dumpnodes();
				RECOVER(try43);
				nc_printf("Debugging printout failed.\n");
				ENDTRY(try43);
			}
			nc_putchar('\n');
		}
		long for_limit = gg.numpages;
		for (temp1 = 1; temp1 <= for_limit; temp1++)
		{
			nc_printf("Try to save page %ld? ", temp1);
			do
			{
				tempch = m_inkey();
			} while (tempch != 'n' && tempch != 'N' && tempch != 'y' &&
					tempch != 'Y');
			nc_printf("%c\n", tempch);
			if (tempch == 'y' || tempch == 'Y')
				emergencysave((int)temp1);
		}
	}
	RECOVER(try42);
	nc_printf("Error %d in outer TRY-RECOVER\n", P_escapecode);
	ENDTRY(try42);
	ENDTRY(try40);
	TRY(try44);
	closetools();
	RECOVER(try44);
	ENDTRY(try44);
	closedumpfiles();
	showalpha();
	cleanlog();
	if (dumpfile != NULL)
		fclose(dumpfile);
	if (tracefile != NULL)
		fclose(tracefile);
	exit(0);
}


